! ===========================================================================================================
! integer function qc07_regional_test.f90
!
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20191205 VH     Purge obsolete use of functions INPOLYGON and inpoly tp [revent run-time errors.
! 20191126 VH     Changed input lat/lon rom double precision to normal float, debug changes.
! 20190910 VH     Restored 'lost' Ylatmax =46 to prevent array overflow at CLS
! 20190829 JN     Test 7, Red Sea temperature lower limit changed from 21.7C to 21.0C.
! 20190521 VH     Adapted the new "inpoly" function, including calling lat/lon in the correct order, and add
!                 the DOGRID option to write out a grid of lon, lat, result for beoth Med and Red Seas.
! 20190521 JN     Added function "inpoly" correcponding to matlab func -"inpoly.m"
! 20190416 VH     Added debug lines to identify problems with INPOLYGON.
! 20170712 all    Release of qcplus v1.1 - many changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Included
!
! subroutine is_profile_regional(lat,lon)
! logical function inpolygon(xin, yin, xconv, yconv, nconv)
! ===========================================================================================================
! Notes
!
! This test applies to only certain regions of the world where conditions can be further
! qualified. in this case, specific ranges for observations from the Mediterranean and Red Seas
! further restrict what are considered sensible values. The Red Sea is defined by the region
! 10N,40E; 20N,50E; 30N,30E; 10N,40E and the Mediterranean Sea by the region 30N,6W;
! 30N,40E; 40N,35E; 42N,20E; 50N,15E; 40N,5E; 30N,6W.
! Action: Individual values that fail these ranges should be flagged as bad data and removed
! from the TESAC being distributed on the GTS. If both temperature and salinity values at the
! same depth both fail, then values for depth, temperature and salinity should be removed from
! the TESAC being distributed on the GTS.
!
! Red Sea                                                                                                        
! * Temperature in range 21.0 to 40.0°C
! * Salinity in range 2 to 41.0 PSU
!
! Mediterranean Sea
! * Temperature in range 10.0 to 40.0°C
! * Salinity in range 2 to 40.0 PSU
!
! Function return values:
!  0 = PASS - Success, all positions passed
!  1 = FAIL - Failure, some or all positions failed
! -1 = NOTD - cannot be determined
!

integer function qc07_regional_test(caller,info)  

use communication                       ! module
use data_arrays                         ! module  

implicit none
                                             
! ===========================================================================================================
! varialbe/array declarations

character(*), intent( in ) :: caller
integer,      intent( in ) :: info(3)

character(len=4), parameter :: sub = 'QC07'
character(len=21) ,parameter :: QCTEST='Regional range test'
integer :: result;

integer :: ifl, itype, beg, end;
integer :: ngood,nbad,nskip
real    :: lat, lon
! only T & S use the regional range/

real, parameter :: red_range_min(4) = (/-999.0, 21.0,  2.0, -999.0 /)
real, parameter :: red_range_max(4) = (/-999.0, 40.0, 41.0, -999.0 /)

real, parameter :: med_range_min(4) = (/-999.0, 10.0,  2.0, -999.0 /)
real, parameter :: med_range_max(4) = (/-999.0, 40.0, 40.0, -999.0 /)

real :: range_min, range_max

! ===========================================================================================================
! executable statements

ifl = regional_ncol; 

result = NOTD ! cannot be determined

itype=info(1); beg=info(2); end=info(3); ! run characteristics

if(DBGIO) call dbg_enter_type(sub,parmname(itype))

if(sea == "RED")then
    range_min = red_range_min(itype)
    range_max = red_range_max(itype)
elseif(sea == "MED")then
    range_min = med_range_min(itype)
    range_max = med_range_max(itype)
else
    write(msg,*)'Set up the error stuff for qc07_regional_test'; call wrtdbg(sub,msg);
    write(errmsg(1),'("FATAL ERROR 708 QC08 qc08_regional check")');
    write(errmsg(2),'("FATAL ERROR 708 QC08 regional sea must be RED or MED: ",a)')sea
    call error_exit(sub,708,2,errmsg); ! qc07 - regional test must be RED or MED
endif

do idat=beg,end											! idat=beg,end
    if(hasDATA(idat))then									! if hasDATA
	if(proftot(idat) > 0)then								! if ntot>0
	    ngood = 0; nbad = 0; nskip = 0;
	    do ilev=1,nlev
		if(skip_test(profflag(idat,ilev)))then;
		    if(DBLEV== 7) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip+1,"skp")
		    nskip = nskip + 1;
		    testflag(idat,ilev,ifl) = NOTF; go to 99;
		endif ! Skip this level
		if(profdata(idat,ilev) <range_min)then
		    testflag(idat,ilev,ifl) = BAD; nbad = nbad + 1;
		    if(DBLEV== 7) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip,"bad1")
		elseif(profdata(idat,ilev) > range_max)then
		    testflag(idat,ilev,ifl) = BAD; nbad = nbad + 1;
		    if(DBLEV== 7) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip,"bad2")
		else
		    testflag(idat,ilev,ifl) = GOOD; ngood = ngood + 1;
		    if(DBLEV== 7) call report_ngnbns(sub,idat,ilev,ifl,ngood,nbad,nskip,"gud")
		endif
		99 continue ! skip_test
	    enddo ! ilev=1,nlev

	endif 											! if ntot>0
	call evaluate_flags(sub,QCTEST,datname(idat),idat,ifl,beg,result); maxresult=max(maxresult,result);
    endif											! if hasDATA
enddo												! idat=beg,end

999 continue ! THE_END

! The result returned is the max flag of those evaluated
qc07_regional_test = maxresult;
if(DBGIO) call dbg_leave_type(sub,parmname(itype),result);

return
end function qc07_regional_test

! ===========================================================================================================
! subroutine is_profile_regional(lat,lon)
! ===========================================================================================================
! Notes
! ===========================================================================================================
!
! This subroutine tests to see if lat/lon is in one of the regonal areas MED or RED
!
! ===========================================================================================================

subroutine is_profile_regional(lat,lon)

use communication                       ! module
use data_arrays                         ! module  

implicit none
                                                   
! ===========================================================================================================
! varialbe/array declarations

real, intent( in ) :: lat
real, intent( in ) :: lon

character(len=14), parameter :: sub = 'is_prof_regnl'

integer n1,N_med,N_red,nmax_med,nmax_red
parameter (nmax_red=4,N_red=4)
parameter (nmax_med=7,N_med=7)
real  :: Xpoly_red(nmax_red),Ypoly_red(nmax_red)
real  :: Xpoly_med(nmax_med),Ypoly_med(nmax_med)
integer :: inpoly;  

double precision dlat, dlon

real :: Ylatmin, Ylatmid, Ylatmax

! What are the longitude ranges for each latitude between 10 & 50?
real west_lon, east_lon
real red_west_lon(17), red_east_lon(17)
real med_west_lon(16), med_east_lon(16)
logical :: is_red, is_med;
integer ilat, iloc
! ===========================================================================================================
! executable statements

if(DBGIO) call dbg_enter(sub)

data Xpoly_red/40,30,50,40/
data Ypoly_red/10,30,20,10 /

data Xpoly_med/-6,40,35,20,15,-5,-6/
data Ypoly_med/30,30,40,42,50,40,30/

!                   13   14   15   16   17   18   19   20   21   22   23   24   25   26   27   28   29/
data red_west_lon /41.0,40.0,39.0,39.0,37.5,37.0,37.0,37.0,36.5,35.0,35.0,34.5,34.0,33.5,33.0,32.0,32.0/
data red_east_lon /44.0,43.5,43.0,43.0,43.0,42.5,41.5,40.5,39.5,39.5,39.0,38.5,37.5,37.0,36.0,35.5,35.5/
!                   30   31   32   33   34   35   36   37   38   39   40   41   42   43   44   /45
data med_west_lon /15.0,15.0,11.0, 9.5, 9.5,-5.5,-5.5,-3.0,-1.0,-1.0, -.0, 0.0, 2.5, 2.5, 2.5,12.0/
data med_east_lon /34.5,35.0,36.0,36.0,36.0,36.0,36.5,29.0,28.0,27.0,30.0,29.0,20.0,19.5,17.5,15.5/

if(DBLEV== 7)then; write(msg,'("Test region of ",2f9.1)')lat,lon; call wrtdbg(sub,msg); endif

if(DBLEV== 7)then;
    write(msg1,'("REGNOR ",i5,f12.3,f12.3)')thiscyc,lon,lat;     ! Assume it is NOT regional
    write(msg2,'(" 0 ",i5,f12.3,f12.3," 777777")')thiscyc,lon,lat; ! for matlab, 777777 means regonal, 0 means regular
endif

if(DOGRID)then
    grxlon = grxlonlo - grinc;
    do while (grxlon < grxlonhi)
	grxlon = grxlon + grinc; dlon = DBLE(grxlon)
	grylat = grylatlo - grinc;
	do while (grylat < grylathi)
	    grylat = grylat + 1.00; dlat = DBLE(grylat);
	enddo
    enddo
    if(DOGRID .and. DBLEVEXIT)then; write(DBG,'("TEST EXIT after writing test grid")'); call exit(0); endif ! DBG
endif; ! DBGREG

Ylatmin = 13.0; Ylatmid = 30.0; Ylatmax = 46.0; ! red is 10N to 30N; med is 30N to 46N

if(lat < Ylatmin)then;
    if(DEBUG) write(DBG,'("DBG QC07 I lat south of latmin")');
    return; 
endif;
if(lat > Ylatmax)then;
    if(DEBUG) write(DBG,'("DBG QC07 I lat north of latmax")');
    return; 
endif;
ilat = ifix(lat);
if(lat >= 30.0)then;       ! Might be in the Med
    iloc = ilat - 29;
    west_lon = med_west_lon(iloc);
    east_lon = med_east_lon(iloc);
    if(lon < west_lon)then
	if(DEBUG) write(DBG,'("DBG QC07 I lon west of MED west_lon")');
	return;
    endif;
    if(lon > east_lon)then
	if(DEBUG) write(DBG,'("DBG QC07 I lon east of MED east_lon")');
        return;
    endif;
    write(msg,'("QC07 This profile is in the MED.")'); call wrttrc(sub,msg); if(DEBUG) write(DBG,'(a)')trim(msg);
    is_med = .true.;
    sea = "MED";
    do_regional_test = .true.;
else;   	! Might be in the Red
    iloc = ilat - 12;
    west_lon = red_west_lon(iloc);
    east_lon = red_east_lon(iloc);
    if(lon < west_lon)then
	if(DEBUG) write(DBG,'("DBG QC07 I lon west of RED west_lon")');
	return
    endif
    if(lon > east_lon)then
	if(DEBUG) write(DBG,'("DBG QC07 I lon east of RED east_lon")');
	return;
    endif;
    write(msg,'("QC07 This profile is in the RED.")'); call wrttrc(sub,msg); if(DEBUG) write(DBG,'(a)')trim(msg);
    is_red = .true.;
    sea = "RED";
    do_regional_test = .true.;
endif

! OLD if (INPOLYGON(lat, lon, Ypoly_red, Xpoly_red, N_red))then
! OLD if (inpoly(lon, lat, Xpoly_red, Ypoly_red, N_red) .eq. 1)then
! OLD     sea = "RED";
! OLD     do_regional_test = .true.;
! OLD     if(DEBUG)then; write(msg,'("This profile is RED")'); call wrtdbg(sub,msg); endif
! OLD     if(DBLEV== 7)then;
! OLD 	   write(msg1,'("REGRED",i5,f12.3,f12.3)')thiscyc,lon,lat;     ! It is RED regional
! OLD 	   write(msg2,'(" 1",i5,f12.3,f12.3," 777777")')thiscyc,lon,lat; ! for matlab, 777777 means regonal, 1 means RED sea
! OLD     endif
! OLD endif
! OLD 
! OLD !OLD if(INPOLYGON(lat, lon, Ypoly_med, Xpoly_med, N_med))then
! OLD if(inpoly(lon, lat, Xpoly_med, Ypoly_med, N_med) .eq. 1)then
! OLD     sea = "MED";
! OLD     do_regional_test = .true.;
! OLD     if(DEBUG)then; write(msg,'("This profile is MED")'); call wrtdbg(sub,msg); endif
! OLD     if(DBLEV== 7)then;                                                                        
! OLD 	   write(msg1,'("REGMED",i5,f12.3,f12.3)')thiscyc,lon,lat;     ! It is MED regional
! OLD 	   write(msg2,'("-1",i5,f12.3,f12.3," 777777")')thiscyc,lon,lat; ! for matlab, 777777 means regonal, -1 means MED sea
! OLD     endif
! OLD endif
! OLD 
! OLD if(DBLEV== 7)then;
! OLD     call wrtlog(sub,msg1);
! OLD     call wrtlog(sub,msg2);
! OLD endif

if(DEBUG)then; if(do_regional_test)then; call dbg_leave(sub,1); else; call dbg_leave(sub,0); endif; endif

return
end subroutine is_profile_regional

! ===========================================================================================================
! logical function inpolygon(xin, yin, xconv, yconv, nconv)
!
! ===========================================================================================================
! Notes

!  Revison date: Jan-05-04 (YHD - DATA statement does not work)
!
!
!  SUBROUTINE TO DETERMINE WHETHER OR NOT A POINT IS IN A POLYGON
!
!  INPUT PARAMETERS UNCHANGED IN THE ROUTINE
!  CONV CONTAINS THE COORDS OF THE POLYGON XCONV, YCONV
!  X AND Y ARE THE COORDINATES OF THE POINT TO BE CHECKED
!  NCONV IS THE NUMBER OF POINTS IN THE POLYGON
!
!  OUTPUT PARAMETERS
!  INPOLYGON IS UNSET ON ENTRY BUT IS RETURNED AS .TRUE. IF THE POINT
!  LIES INSIDE THE POLYGON, OTHERWISE RETURNED AS .FALSE.
!
! ===========================================================================================================
logical function inpolygon(xin, yin, xconv, yconv, nconv)

      REAL XCONV(NCONV), YCONV(NCONV),  XIN, YIN
      REAL Z(4)
      REAL X,Y
!      DATA Z/-32701,-32701,32701,32701/
! ===========================================================================================================
! executable statements
      X=XIN
      Y=YIN
      Z(1)=-32701.
      Z(2)=Z(1)
      Z(3)=32701.
      Z(4)=Z(3)

!      IF (Z(1).NE.-32701) GO TO 20
      IF ((Z(1)+32701).GT.0.001) GO TO 20
      DO 10 I = 1,NCONV
         Z(1)=MAX(Z(1),XCONV(I))
         Z(2)=MAX(Z(2),YCONV(I))
         Z(3)=MIN(Z(3),XCONV(I))
         Z(4)=MIN(Z(4),YCONV(I))
   10    CONTINUE
   20 INPOLYGON=.TRUE.
      IF(X .LT. Z(3) .OR. X .GT. Z(1)) INPOLYGON=.FALSE.
      IF(Y .LT. Z(4) .OR. Y .GT. Z(2)) INPOLYGON=.FALSE.
      IF(.NOT. INPOLYGON) RETURN
      J=0
      DO 90 I = 2,NCONV
         M=0
      IF ((YCONV(I-1)-Y)*(Y-YCONV(I))) 90,30,70
   30 IF (YCONV(I-1)-YCONV(I)) 60,40,50
   40 IF ((XCONV(I-1)-X)*(X-XCONV(I))) 90,100,100
   50    M=M-2
   60    M=M-1
   70    M=M+2
      IF ((Y-YCONV(I-1))*(XCONV(I)-XCONV(I-1))/(YCONV(I)-YCONV(I-1))+XCONV(I-1)-X)&
          90,100,80
   80    J=J+M
   90    CONTINUE
      INPOLYGON=.FALSE.
      IF(J/4*4 .NE. J) INPOLYGON=.TRUE.
     
  100 RETURN
      end function inpolygon

! ===========================================================================================================
integer function inpoly(x, y, xv, yv, nv)

use communication                       ! module
use data_arrays                         ! module  

! JN - 20190419 Note:  The existing function "inpolygon" hasn't been working correctly in all the scenarios.
!   	               So this is a newly made function which is based on the Matlab function - "inploy.m"
	
! 	Possible outcomes of the function: 
!       -  x,y is outside the polygon XV,YV       : inpoly=0
!       -  x,y is inside   "   "       "  "       : inpoly=1
!       -  x,y is near (but outside) the polygon (at an edge point)  : inpoly=2
!       -  something wrong, error occured     : inpoly=-1;
	
!       Points classified as "near" are actually outside
!       the polygon, but less than approx TOL away (This is sometimes
!       useful for identifying "peninsulas" when gridding data).
!
!       xv,yv are ordered lists of polygon vertices. The last point
!       in the list (identical to the first point) may be omitted.
!       There is no restriction on the shape of the polygon.
!
!       A tolerance for identifying "nearby" points is set as 1e-6.
      
          
    integer , intent(in) ::  nv                                                       
    real, intent(in) :: x,y,xv(nv),yv(nv);
    real :: dx(nv),dy(nv),atan2_dydx(nv),tmp,angs(nv-1);
    real :: lgt_val,dxv,dyv;
    real :: TOL,EPS,PI;

    integer :: ii,cnt_sum_dxdy;
    integer :: vertexpt, edgept,interiorpt;
    character(len=10) :: sub = 'inpoly';
    
    ! Constants used in calculations.               
    parameter (EPS=2.2204e-16,TOL=0.000001);
    parameter (PI = 3.141592653589793);

    do i=1,nv
	dx(i) = 0.0
	dy(i) = 0.0
    enddo 

    inpoly=-1;
    vertexpt=0;
    edgept=0;
    interiorpt=0;
    cnt_sum_dxdy=0;                              
                                                  
    if (nv==1)then         
    	write(errmsg(1),'("NON-FATAL ERROR 296 Can not form Polygon from a single edge point! ")')
    	write(errmsg(2),'("NON-FATAL ERROR 296 in profile: ",a)')trim(profile)
    	write(errmsg(3),'("NON-FATAL ERROR 296 for position: lat = ",f9.4," lon = ", f9.4)')y,x;
        call error_report(sub,296,3,errmsg); ! bad result from inpoly  
    	return;  
    endif;  
 
    !If the given Polygon is not closed, make sure it is closed.
    if((xv(nv) /= xv(1)) .or. (yv(nv) /= yv(1)))then    	  
        write(errmsg(1),'("NON-FATAL ERROR 297 Polygon is NOT closed! ")'); 
        write(errmsg(2),'("NON-FATAL ERROR 297 in profile: ",a)')trim(profile)  
        write(errmsg(3),'("NON-FATAL ERROR 297 for position: lat = ",f9.4," lon = ", f9.4)')y,x;
        call error_report(sub,297,3,errmsg); ! bad result from inpoly
        return;  
    endif; 
    do ii=1,nv
        dxv=xv(ii)-xv(ii-1);                                      
        dyv=yv(ii)-yv(ii-1); 
    	if(dxv+dyv .ne. 0)then             
    	   cnt_sum_dxdy=cnt_sum_dxdy+1;
    	endif    
    enddo  
   if (cnt_sum_dxdy .eq. 0)then    	  
        write(errmsg(1),'("NON-FATAL ERROR 298 Vertices are not valid to form a Polygon! ")')
        write(errmsg(2),'("NON-FATAL ERROR 298 in profile: ",a)')trim(profile)  
        write(errmsg(3),'("NON-FATAL ERROR 298 for position: lat = ",f9.4," lon = ", f9.4)')y,x;
        call error_report(sub,298,3,errmsg); ! bad result from inpoly
        return;  
   endif	   
    do ii=1,nv  
        ! finds the distance of the point from each vertex 
    	dx(ii)=x-xv(ii);
    	dy(ii)=y-yv(ii);      
    	! finds the atan2 between dy,dx
        atan2_dydx(ii)=atan2(dy(ii),dx(ii));                    
        if ((abs(dx(ii)) .lt. TOL) .and. (abs(dy(ii)) .lt. TOL)) then
            vertexpt=1;    
        endif                                                  
        if(ii<2)then;cycle;endif;  ! need ii>1 to do the rest
        dxv=xv(ii)-xv(ii-1);                                      
        dyv=yv(ii)-yv(ii-1); 

        tmp=atan2_dydx(ii) - atan2_dydx(ii-1);       
        angs(ii-1)= mod(tmp+3*PI+EPS, 2*PI)- PI-EPS; 
        lgt_val=(dx(ii-1)*dxv+dy(ii-1)*dyv)/(dxv*dxv+dyv*dyv);         
        distx=2*tol; disty=2*tol;
        if((lgt_val>0) .and. (lgt_val<1))then
        	distx=dx(ii-1)-lgt_val*dxv;
        	disty=dy(ii-1)-lgt_val*dyv;	
        endif
    	if(abs(distx)<tol .and. abs(disty)<tol)then
    	   edgept =1;
        endif
    enddo                 
                                           

   ! If the angles sum to 0, 4*pi, etc., we are outside - if they
   ! sum to  2*pi, 6*pi, etc. we are inside. This will result in a
   ! "checkerboard" inside/outside classification when polygon
   ! segments intersect.
                                                     
   interiorpt=mod(abs(nint(SUM(angs)/(2*PI))),2); 
	
   ! Classify points by priority                      
   inpoly=2*edgept;                
     
   if(vertexpt .ne. 0)then
      inpoly=2        
   endif

   if(interiorpt>0)then
       inpoly=1;
   endif
end function inpoly
! ===========================================================================================================
