! ===========================================================================================================
! module data_arrays.f90
!
! -----------------------------------------------------------------------------------------------------------
!
! Revision history
! date     who    what
! 20200325 VH     The check_date function now ok writh respect to BADGPSFIX. 
!                  in analyze_positions, check position block for monotonically increasing time for qc/ncprof,
!                 issue a FATAL ERROR if found. .
! 201911?5 VH     Moved MAX_MEASUREMENT from netcdf_arrays to to data_arrays.
!                 Logicals to track bad gps fixes in TIME_ASCENT_END, TIME_TRANS_START, measurement codes used
!                 to select JULD and positions. twonew functions: test_for_badgps to test if the year is too low (a
!                 uear before launch), or too high (two years after present year), and check_cdate_qc (returns the
!                 normal qc flagw (1,4,9) or a 'G' if the date has a bad gps fix year.
! 201909?? JN     Add logical hasCNDC_traj (modified 17 Jul 2019, merged later)
! 20190904 VH     Add logicals to identify bad GPS years in the position block (BADGPS)
! 20190829 JN     Small correctio to maxflag evaluation.
! 20190521 VH     AAdd DOGRID option to land_or_water, called by qc04 - write out a grid of lan/lon/result
! 20190416 VH     Revision of names & usage for variables used in interpolation and ice detection
! 20190212 JN     Fix the discrepancy between units and values in the QC/NCPROF files for the floats containing
!                 CONDUCTIVITY.
! 20190108 JN     MOdify WARNING 114 slightly.
! 20181017 VH     If pres has a qcflag of 4, do not change other data if qcflag is 9
!                 In evaluate_grade, do not count missing values.
! 20180920 CA     Revise how the resolution is determined. Change how the topo file is opened.
! 20180724 VH     INTERPOLATION/ICE DETECTION: Define arrays needed for interpolaiton & ice ancalyusis
! 20180619 JN     Change platform_type = SOLO_D to SOLO_D_MRV
!                 I167 (improve frozen profile warning) add logticals for greylised floats
!                 Do not include qc=8, interpolated values, in the nc file or grade
!                 I067 Do accept a pi_suppliedposnqc of 9, but issue a lognote
!                 added ALTOIR and SOLODIR from MRV.
!                 Add <param>_is_greylisted logicals, allow 9 for PI-supplied posn_qc
!                 format/string problems (vargo)
! 20180417 VH     General WARNING revisions.
! 20180307 VH     "no primary profile" changes merged into the new BR version
! 20180307 VH     Extensive changes to create BR file and no-primary-profile cases
! 20180129 VH     Numerous changes for biop.
! 20180111 JN     "no primary profile": Add two variables (primary_prof_found,nprf_start) changes (VH merged 20180307)
! 20171020 VH     Add ability to handle ALAMO floats (catchup on 20170305)
! 20170927 JN     Add pos_index_first703 pos_index_first703 declaration for traj changes
! 20170811 VH     Fix bad internal write when "Date int he future" is found, ittus CRITICAL_WARNING
! 20170720 VH     Keep track of zerolines independtntly for nprof1 & nprof2    
! 20170712 all    Release of qcplus v1.1 - many changes to all, some documented/tracked, some not
! 20160511 all    Initial release of qcplus v1.1.0
!
! ===========================================================================================================
! Notes:

! This module contains variable/array definitions used by qc * ncprof funtions, as well as any jointly-used
! variables or funcitns/subroutines that must be available everywhere.
! ===========================================================================================================

module data_arrays

use communication                                                ! module

implicit none

! ===========================================================================================================
! moedule varialbe/array declarations
! ===========================================================================================================

integer, parameter :: MAX_PARM_TABLE = 2000 ! Max. # of params in table

integer, parameter :: MAXCYC =  1999		! max nbr of CYCLES for a given float - must be multiple of 3
integer, parameter :: MAXCYC3 = 1998		! MAXCYC divisible by 3 (to define
integer, parameter :: MAXCYC4 = 2000		! MAXCYC divisible by 4 (to define

integer, parameter :: MAXPRF  = 4 		! max nbr of PROFILES (corresponds to N_PROF in the nc file)
integer, parameter :: MAXPRES = MAXPRF*2	! max nbr of PRES columns (regular and adjusted)
integer, parameter :: MAXTEMP = MAXPRF	! max nbr of TEMP columns (regular only
integer, parameter :: MAXPSAL = MAXPRF*2	! max nbr of PSAL columns (regular and adjusted)
integer, parameter :: MAXDOXY = 5		! max nbr of DOXY profiles(IRREGULAR!: DOXY1,DOXY2,DOXYA,DOXYB,DOXYC)
integer, parameter :: MAXNOQC = MAXPRF*1	! max nbr of other parameters, those that will NOT be qc'd

integer, parameter :: MAXDAT = MAXPRES+MAXTEMP+MAXPSAL+MAXDOXY;       ! Total number of colums that are to qc'd

! The next part is a total kluge, just used for handiness's sake

! ==============================================================================
! Adjust these when the MAXPRF is increased:

integer, parameter :: p1r= 1, p1a= 2, p2r= 3, p2a= 4, p3r= 5, p3a= 6, p4r= 7, p4a= 8; ! PRES indices
integer, parameter :: t1r= 9,         t2r=10,         t3r=11,         t4r=12;         ! TEMP indices
integer, parameter :: s1r=13, s1a=14, s2r=15, s2a=16, s3r=17, s3a=18, s4r=19, s4a=20; ! PSAL indices
integer, parameter :: do1=21, do2=22, doa=23, dob=24, doc=25;                         ! DOXY indices - IRREGULAR

integer  :: P(3) = (/ 1, 1, 8/)       ! data type p, 8 data cols (p,padj,p2,p2adj)
integer  :: T(3) = (/ 2, 9,12/)       ! data type t, 4 data cols (t,t2)
integer  :: S(3) = (/ 3,13,20/)       ! data type s, 8 data cols (s,sadj,s2,s2adj)
integer  :: O(3) = (/ 4,21,25/)       ! data type d, 5 data cols

!==================================================================================================
! Data PARAMETER informatio (umber and type of data parameters)
! These are the variables that go into the nc files - single-valued, to go into the N_PROF arrays

integer, parameter :: MAXPAR = 5	! max nbr of parameters in the qc file (pres,temp,psal,doxy,cndc)
integer, parameter :: MAXPAR_META = 10	!! max nbr of parameters in the qc metafile (includes bioArgo variables)
integer, parameter :: MAXNCPAR = 5	! max nbr of parameters in the nc file (pres,temp,cndc,psal,doxy
logical :: hasQCPAR(MAXPAR), hasPARPRF(MAXPAR,MAXPRF)

logical, dimension(MAXPAR,MAXPRF) :: hasFLAGS = reshape( (/ &
.true.,  .true., .true., .true.,  .false., &                         ! P1 T1 S1 D1 C1 no flags for CNDC at all
.true.,  .true., .true., .true.,  .false., &                         ! P2 T2 S2 D2 C2 no flags for CNDC at all
.true.,  .true., .true., .false., .false., &                         ! P3 T3 S3 D3 C3 no flags for d03, none CNDC at all
.true.,  .true., .true., .false., .false.  /),(/ MAXPAR,MAXPRF /)); ! P4 T4 S4 D4 C4 no flags for d04, none CNDC at all
! P4 T4 S4 D4 C4 no flags for d04, no flags for CNDC at all

integer, dimension(MAXPAR,MAXPRF) :: REGDAT = reshape( (/ &
p1r, t1r, s1r, do1, 0, &
p2r, t2r, s2r, do2, 0, &
p3r, t3r, s3r, 0,   0, &
p4r, t4r, s4r, 0,   0  /),(/ MAXPAR,MAXPRF /));

! This is the profile for REGULAR data (not adjusted) for the CORE data (NOT DOXY)
integer :: REGPRFCORE(MAXDAT) = (/ 1,0,2,0,3,0,4,0,1,2,3,4,1,0,2,0,3,0,4,0,0,0,0,0,0 /);
logical :: isPARADJ(MAXPAR) = (/ .false., .false., .false., .false., .false. /); ! is this parameret adjusted
! pres                 temp                 psal                 doxy             cndc

character(len=3), dimension(MAXPAR,MAXPRF) :: REGCHR = reshape( (/ &
"p1r","t1r","s1r","d1o","c1x", &
"p2r","t2r","s2r","d2o","c2x", &
"p3r","t3r","s3r","d3o","c3x", &
"p4r","t4r","s4r","d4o","c4x"  /),(/ MAXPAR,MAXPRF /));

logical :: hasNCPAR(MAXNCPAR), primary_prof_found =.true.;

character(len=4) ::   ncparname(MAXNCPAR) = (/"PRES","TEMP","CNDC","PSAL","DOXY"/); ! note the order is different from qc par
character(len=4) ::   ncparout(MAXNCPAR);

integer :: inca	! Counter for NCARRS
integer, parameter :: NCARRS = 5 ! total nbr of nc arrays to load (reg,reg_qc,adj,adj_qc,err)
integer :: nparam;	! How many data parameters do we actually have? (used sparingly)
! ipar, npar NOT USED GLOBALLY - use nparam for npar, use ipar only in ncprof_arrays.f90, define it there

character(len=1) ::   parmchar(MAXPAR) = (/"P",   "T",   "S",   "O",   "C"   /);
character(len=4) ::   parmname(MAXPAR) = (/"PRES","TEMP","PSAL","DOXY","CNDC"/)

! =================================================================================================================================
! Values and counters for the major arrays
integer :: iprf,nprf;			! for MAXPRF, number of profiles

! This is a wierd one - this give which pres is a variables "buddy" - which pres it is associated with
! they are grouped by which pres goes with each variable. example: temp1/reg is associated with pres1/reg
!
integer, parameter :: pbud(MAXDAT) = (/ &						! Dimensioned using MAXPRF=4
p1r, p1a, p2r, p2a, p3r, p3a, p4r, p4a, & ! pres is it's own buddy both reg & adj
p1r,      p2r,      p3r,      p4r,      & ! temp has only reg, not adj
p1r, p1a, p2r, p2a, p3r, p3a, p4r, p4a, & ! psal has both reg & adj
p1r, p2r, p2r, p2r, p2r  /);              ! doxy1 only with primary, others mulplple with nprf2 only (so far).

! Just slightly less weird, this is the salinity that corresponds to each temp - it's "buddy" also MAXPRF=4.

integer, parameter :: tbud(MAXPRF) = (/ s1r, s2r, s3r, s4r /);

! These are the "shorthans" data names - used in the flag title
character(len=6) :: datname(MAXDAT) = &						! Dimensioned using MAXPRF=4
(/  "P     ", "P ADJ ", "P2    ", "P2 ADJ",  "P3    ", "P3 ADJ",  "P4    ", "P4 ADJ",&
"T     ",           "T2    ",            "T3    ",          "T4    ",          &
"S     ", "S ADJ ", "S2    ", "S2 ADJ",  "S3    ", "S3 ADJ","S4    ", "S4 ADJ",&
"O     ", "O2    ", "O2A   ", "O2B   ", "O2C   "  /)
integer :: idat
integer :: res

integer, parameter :: N_HIST = 25             ! That is how many John Gilson wants
integer, parameter :: MAXTRAJ = 200000        ! max nbr of traj records
integer, parameter :: MAX_DRIFTperProf = 3000 ! max nbr of drift records from a single profile

integer, parameter :: NFLAGPRES = 3 ! number of pres flags
integer, parameter :: NCOLSPRES = MAXPRES + MAXPRES*NFLAGPRES ! nbr of PRES profile columns
integer, parameter :: NFLAGTEMP = 5 ! number of temp flags
integer, parameter :: NCOLSTEMP = MAXTEMP + MAXTEMP*NFLAGTEMP ! max nbr of TEMP profiles X 4 flags/profile
integer, parameter :: NFLAGPSAL = 6 ! number of psal flags
integer, parameter :: NCOLSPSAL = MAXPSAL + MAXPSAL*NFLAGPSAL ! max nbr of PSAL profiles X 4 flags/profile
integer, parameter :: NFLAGDOXY = 3 ! number of doxy flags
integer, parameter :: NCOLSDOXY = MAXDOXY + MAXDOXY+NFLAGDOXY ! one doxy sensor produces two phase values - not qc'd
integer, parameter :: MAXCOLS = 30 + NCOLSPRES+NCOLSTEMP+NCOLSPSAL+NCOLSDOXY+MAXNOQC ! just add 30, FIX LATER

character(len=40) :: flag_name;
character(len=25) :: pres_flg_name(NFLAGPRES) = (/ &
"GROSS CHECK              ", &
"PRESS INCREASING CHECK   ", &
"DEEPEST PRESS CHECK      " /)

character(len=25) :: rgnl_flg_name = "REGIONAL CHECK           ";

character(len=28) :: near_surf_name = "NEAR_SURFACE_MIXED AIR/WATER";

character(len=25) :: temp_flg_name(NFLAGTEMP) = (/ &
"GROSS CHECK              ", &
"SPIKE CHECK              ", &
"VERTICAL GRADIENT CHECK  ", &
"ANALYSIS CHECK           ", &
"CLIMATOLOGY CHECK        "  /)
character(len=25) :: psal_flg_name(NFLAGPSAL) = (/ &
"GROSS CHECK              ", &
"SPIKE CHECK              ", &
"VERTICAL GRADIENT CHECK  ", &
"ANALYSIS CHECK           ", &
"CLIMATOLOGY CHECK        ", &
"DENSITY GRADIENT CHECK   "  /)
character(len=25) :: doxy_flg_name(NFLAGDOXY) =(/ &
"GROSS CHECK              ", &
"SPIKE CHECK              ", &
"VERTICAL GRADIENT CHECK  "  /)

integer, parameter :: MAXLEV = 6500     ! max nbr of levels in one profile1
integer, parameter :: MAXHED = 10000    ! max nbr of header lines (shuld be same as MAXMCL)
integer, parameter :: MAXMET = 150	! max nbr of lines in the meta file

integer, parameter :: IND_NOT  = -1	! INDICATOR not evaluated
integer, parameter :: IND_OFF  =  0	! INDICATOR is off
integer, parameter :: IND_ON   =  1	! INDICATOR us on

integer, parameter :: NOQC      = 0 	! No QC was performed
integer, parameter :: GOOD      = 1 	! Good data All Argo real-time QC tests passed
integer, parameter :: PROB_GOOD = 2 	! Probably good
integer, parameter :: PROB_BAD  = 3 	! Probably bad data
integer, parameter :: BAD       = 4 	! Bad data
integer, parameter :: CHANGED   = 5 	! Value changed Value changed Value changed
! 6 & 7 are not used
integer, parameter :: INTERP    = 8 	! Interpolated value
integer, parameter :: MISSING   = 9 	! Missing value
logical, parameter :: flag_missing_values = .true.

integer, parameter :: MAXPOS = 100     ! max nbr of positions
integer, parameter :: MAX_MEASUREMENT = 200000;
integer, parameter :: MAXDRIFTTIMES = 3*MAX_MEASUREMENT; ! DRIFT MEASUREMENT TIME, TIME DRIFT START,TIME DRIFT END

integer, parameter :: IO_OK  =  0     ! IO Status = 0, read succeeded

integer :: i
integer :: nlev = 0;
integer :: ilev, iflg, icol

! are ALL of these old?
! noqc variables

integer :: ihed, nhed ! number of heder lines NOT including end separator ==============liner
character(len=SZC) :: hedline(MAXHED);                ! Stash the header lines for later
character(len=SZC) :: poslabs                         ! position lables (LAT/LON/DATE/etc.
integer :: npos        ! # of position lines NOT including col names, beg/end separator ============== or FORTRAJ line
integer :: npos2,ipos  ! # of position lnies EXCLUDING 999 lines, position counter\

integer :: npos_traj, ipos_traj, pos_index_first703;
character(len=SZC) :: posline_traj(MAXPOS), posadnl_traj(MAXPOS);
character(len=1)   :: pos_posnqc_traj(MAXPOS);	! Overall posn qc (combo of pis_posnqc & pos_posiqc)
character(len=1)   :: pos_posiqc_traj(MAXPOS);	! Impossible position (QC03)
character(len=1)   :: pos_dateqc_traj(MAXPOS);	! Impossible date (QC02)
character(len=1)   :: pos_onland_traj(MAXPOS);	! Position on land (QC04)
character(len=1)   :: pos_speedc_traj(MAXPOS);	! Speed check (QC05)

character(len=SZC) :: pqcline(MAXHED);                ! Stash the header lines from the prev. qc file
integer npqc, ipqc                                    ! Number of pqc lines, counter for them

logical :: this_class_present = .false. ; ! CLASS is in the position block. the most common type
logical :: this_posqc_present = .false.; ! POSITION_QC (or-QC) in the position block instead of CLASS
logical :: this_mcode_present = .false.; ! POSN-QC UNCER CYC and the MC codes are in the position block soloii_new

integer :: this_posqc(MAXPOS), prev_posqc(MAXPOS),index_999_1,index_999_2,count_999;
integer :: valid_pos1 ! first valid position
character(len=SZC) :: posline(MAXPOS);                ! Stash the position lines for later
character(len=24)  :: posadnl(MAXPOS);                ! Stash the extra/additional parts of soloii

character(len=1)   :: sat_posn(MAXPOS);		! Which sat for this position (I or G for iridium or GPS)
character(len=1)   :: pos_posnqc(MAXPOS);	! Overall posn qc (combo of pis_posnqc & pos_posiqc)
character(len=1)   :: pis_posnqc(MAXPOS);	! PI-supplied position qc OR interpolated (8) by us or PI
character(len=1)   :: pos_posiqc(MAXPOS);	! Impossible position (QC03)
character(len=1)   :: pos_dateqc(MAXPOS);	! Impossible date (QC02)
character(len=1)   :: pos_onland(MAXPOS);	! Position on land (QC04)
character(len=1)   :: pos_speedc(MAXPOS);	! Speed check (QC05)

integer :: ipos_posnqc(MAXPOS);       ! Overall posn qc (combo of pis_posnqc & pos_posiqc)
integer :: ipis_posnqc(MAXPOS);       ! PI-supplied position qc OR interpolated (8) by us or PI
integer :: ipos_posiqc(MAXPOS);       ! Impossible position (QC03)
integer :: ipos_dateqc(MAXPOS);       ! Impossible date (QC02)
integer :: ipos_onland(MAXPOS);       ! Position on land (QC04)
integer :: ipos_speedc(MAXPOS);       ! Speed check (QC05)
integer :: ipos_allqc(MAXPOS);        ! Overall qc used to in pick_profile_pos and fill_poslocs

! PI-supplied position from PHY file (SOLOOF_TS10 phy from old nc files)
integer :: pis_posnqc_phy; logical :: found_pis_posnqc_in_phy = .false.;

character(len=SZC) :: posnqc_comment(MAXPOS);		! Comment on how position_qc was determined
logical            :: prev_posn_found, prev_posn_good	! Did we find a previous position? Did we find a GOOD previous position
character(len=256) :: prev_posn_qcfile			! Which qc file was used for the previous position
real               :: prev_good_lat,prev_good_lon;	! read in read_pqc, used in speed check
character(len=14)  :: prev_good_posn_date		! read in read_pqc, used in speed check
logical :: TIME_DIFF_SMALL=.false.

character(len=SZC) :: metline(MAXMET)			! max nbr of lines in the meta file
integer  :: nmet, imet;				! counters for meta lines

character(len=SZC) :: datline(MAXLEV);                ! Stash the data lines for later
real :: datvals(MAXCOLS);
character(len=16)::cdatvals;

! The major arrays that are to be qc'd
integer :: ipres, npres, inprescol(MAXPRES), inpresadjcol
integer :: itemp, ntemp, intempcol(MAXTEMP)
integer :: ipsal, npsal, inpsalcol(MAXPSAL), inpsaladjcol
integer :: idoxy,  ndoxy,  indoxycol(MAXDOXY);

! for theose individual parameters with has<param> logicals (hasPRES,hasTEMP,etc.), element zero is 1 if that parameter
! is present in ANY OF THE PROFILES (eg, CNDC in profile 3 & 4 for some floats)
logical :: hasDATA(MAXDAT)
integer,parameter :: DATPAR(MAXDAT) = (/ 1, 1, 1, 1, 1, 1, 1, 1,  2, 2, 2, 2,  3, 3, 3, 3, 3, 3, 3, 3,  4, 4, 4, 4, 4 /)
! pres             temp      psal      psal adj  doxy
integer,parameter :: DATPRF(MAXDAT) = (/ 1, 1, 2, 2, 3, 3, 4, 4,  1, 2, 3, 4,  1, 1, 2, 2, 3, 3, 4, 4,  1, 2, 2, 2, 2 /)
! pres                temp      psal                doxy
logical :: hasPRES(0:MAXPRF);
integer :: regPRES(MAXPRF) = (/p1r,p2r,p3r,p4r/);	! index of regular pres
integer :: adjPRES(MAXPRF) = (/p1a,p2a,p3a,p4a/);	! index of adjusted pres
logical :: hasTEMP(0:MAXPRF);
integer :: regTEMP(MAXPRF) = (/t1r,t2r,t3r,t4r/);	! index of regular temp
logical :: hasPSAL(0:MAXPRF);
integer :: regPSAL(MAXPRF) = (/s1r,s2r,s3r,s4r/);	! index of regular psal
integer :: adjPSAL(MAXPRF) = (/s1a,s2a,s3a,s4a/);	! index of adjusted psal
! for the DOXY logicals, the index does NOT correspond the the nprf - DOXYA,DOXYA,DOXYC are ALL nprf=2
! create specific logicals for each, case
logical :: DOXY_float, hasDOXY(0:MAXDOXY), hasDOXYA, hasDOXYB, hasDOXYC
logical :: outRDOXY; ! DOXY will go into the R-file - not B-file is made
integer :: regDOXY(MAXDOXY) = (/ do1, do2, doa, dob, doc /);
logical :: hasCNDC(0:MAXPRF); integer :: ncndc, icndc, incndccol(MAXPRF); real :: profcndc(MAXPRF,MAXLEV)
logical :: cnvrt_cndc(0:MAXPRF)=.false.; 

! MV TO COMM. ! these are the parameters in the nc file, in the order in which the appear in the nc file
! MV TO COMM. logical :: hasNCPAR(0:MAXPRF,MAXNCPAR);
! MV TO COMM. character(len=4) ::   ncparname(MAXPAR) = (/"PRES","TEMP","CNDC","PSAL","DOXY"/); ! note the order is different from parname

! moved up integer :: nparam;

! Flag columns for profiles (Note that Sensor drive/Jump check and Frozen profile report only a single flag)/
integer, parameter :: gross_ncol     = 1	! P,T,S,O Which column is used for the gross test (global range)
integer, parameter :: pres_incr_ncol = 2	! P       Which column is used for the pressure increasing test
integer, parameter :: deep_pres_ncol = 3        ! P       Which column is used for the deepest pressure test
integer, parameter :: spike_ncol     = 2        ! T,S,O   Which column is used for the spike test
integer, parameter :: vertgrad_ncol  = 3	! T,S,O   Which column is used for the vertical gradient test
integer, parameter :: analysis_ncol  = 4	! T,S Which column is used for the analysis test
integer, parameter :: clim_ncol      = 5	! T,S Which column is used for the climatology test
integer, parameter :: dengrad_ncol   = 6	! T&S Which column is used for the density gradient test
integer, parameter :: regional_ncol  = 7        ! T,S     Which column is used for the regional range test
integer, parameter :: MAXFLAGS       = 7        ! max nbr of flags for each param
real,    parameter :: multplr_CM2M   = 0.1      ! Multiplier to convert cm to m for cndc. 
integer, parameter :: NFLAG(MAXDAT) = (/ &						! Dimensioned using MAXPRF=4
NFLAGPRES, NFLAGPRES, NFLAGPRES, NFLAGPRES, NFLAGPRES, NFLAGPRES, NFLAGPRES, NFLAGPRES, &
NFLAGTEMP, NFLAGTEMP, NFLAGTEMP, NFLAGTEMP,                     			&
NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, NFLAGPSAL, &
NFLAGDOXY, NFLAGDOXY, NFLAGDOXY, NFLAGDOXY, NFLAGDOXY  /);
!============================================================================

!============================================================================
character(len=1) :: prof_qc(MAXPAR,MAXPRF)			! profile_<param>_qc grade (A-F)
character(len=1) :: prof_pres_qc(MAXPRF);
character(len=1) :: prof_temp_qc(MAXPRF);
character(len=1) :: prof_psal_qc(MAXPRF);
character(len=1) :: prof_doxy_qc(MAXPRF); ! also 2a
character(len=1) :: prof_doxy2_qc(MAXPRF); ! 2b
character(len=1) :: prof_doxy3_qc(MAXPRF); ! 2c
character(len=1) :: prof_cndc_qc(MAXPRF);

integer          :: max_flag(MAXDAT,MAXFLAGS)		! max value for each para
integer          :: overflag(MAXDAT)			! overall flag for each parameter
real ::     profdata(MAXDAT,MAXLEV);
integer ::  profflag(MAXDAT,MAXLEV);			! flag for each level in profile
integer ::  testflag(MAXDAT,MAXLEV,0:MAXFLAGS); 	! 7 is max number of tests
character(len=16) :: histflag(N_HIST,MAXTRAJ)

integer :: maxdataflag(MAXDAT)        ! Max flag for all levels for a single data column
integer :: maxpriprofflag             ! Max flag for all data from PRIMARY profile
integer :: maxoverflag;               ! Max flag for ALL data (basis of JOB_STATUS)
integer :: maxlevflag(MAXPRF,MAXDAT); ! Max flag for each profile for each parameter over all levels
integer :: maxprfflag(MAXPRF);        ! Max flag for each profile over all parameters

logical :: near_surface(0:MAXPRF);			! Is this profile a near_surface profile THAT NEEDS TESTING?
logical :: STS_prof(0:MAXPRF), NST_prof(0:MAXPRF);	! If so, is this profile an STS or or an NST profile?
integer :: nsflag(MAXLEV);			! flag for near-survace profles (1 ONLY per float, so far)
integer :: near_surface_result = 0		! Were any levels marked as PROB_BAD - if so, this will be a 3
integer :: near_surface_profile = 0		! Which profile is the tested near_surface profile
integer :: nspdat, nstdat, nssdat		! profflag cols for pres,temp,psal
logical :: nsprofile_is_nst = .false.; logical :: nsprofile_is_sts = .false.;
character(len=80) :: nsqcline

!============================================================================
! This is the order of the output columns in the qcprof file wrt the HASDAT array
integer :: nqcout, iqcout, qcout(MAXDAT)
!============================================================================
logical :: profile_reversed = .false.
real :: rdum(MAXLEV);			! in case the profile nedds to be reversed
integer idum(MAXLEV);
!============================================================================
! Special-purpose flags (may not always be present
integer :: proftop(MAXDAT),profbot(MAXDAT),proftot(MAXDAT); real :: total(MAXDAT);
logical :: profiles_overlap = .false.			! For ididium floats
logical :: interp_status = .false.
integer :: interp_t2_col, interp_t2(MAXLEV), interp_t3_col, interp_t3(MAXLEV), interp_t4_col, interp_t4(MAXLEV)
integer :: interp_s2_col, interp_s2(MAXLEV), interp_s3_col, interp_s3(MAXLEV), interp_s4_col, interp_s4(MAXLEV)
!============================================================================
! Other arrays that may need to be transferred
logical :: hasCKSM(MAXPRF); integer ncksmcol(MAXPRF); real cksm(MAXPRF,MAXLEV) ! This is almost always present
logical :: cksm1_in_prof1, cksm1_in_prof2	! Some of the inst_types have CHECKSUM after nprf2, but not CKSM2

integer :: ncolnoc, ncoldat;	! from "NUMBER OF COLUMNS" and from first data line
integer :: ncolinp_elim = 0;	! number of inp cols to ELIMINATE as inp cols pres_adj/psal_adj
integer :: icolinp, ncolinp;	! from the column lines
character (len=SZC-40) :: inpcols(MAXCOLS)
character (len=SZC-40) :: outcols(MAXCOLS)
double precision    :: launch_lat, launch_lon
integer :: launch_mon,launch_day,launch_year,launch_hh,launch_mm,launch_sc
character(len=10) :: yearmody; character(len=8) :: hrmnsc;
integer :: ppos,ppos2,ppos2_good; ! ppos number of previous positions from pqc file,ppos2-excudes +99.999 and +999.999
real    :: this_lat(MAXPOS), this_lon(MAXPOS); double precision :: dble_lat(MAXPOS), dble_lon(MAXPOS)
double precision   :: prev_lat(MAXPOS), prev_lon(MAXPOS)
real    :: this_lat2(MAXPOS), this_lon2(MAXPOS),prev_lat2(MAXPOS), prev_lon2(MAXPOS)
integer :: this_year(MAXPOS),this_mon(MAXPOS),this_day(MAXPOS),this_hh(MAXPOS),this_mm(MAXPOS),this_sc(MAXPOS)
integer :: this_year2(MAXPOS),this_mon2(MAXPOS),this_day2(MAXPOS),this_hh2(MAXPOS),this_mm2(MAXPOS),this_sc2(MAXPOS)
character(len=14) :: this_date_time(MAXPOS);
double precision :: juld_date_time(MAXPOS);
double precision :: juld_diff;
integer :: prev_year(MAXPOS),prev_mon(MAXPOS),prev_day(MAXPOS),prev_hh(MAXPOS),prev_mm(MAXPOS),prev_sc(MAXPOS)
integer :: prev_year2(MAXPOS),prev_mon2(MAXPOS),prev_day2(MAXPOS),prev_hh2(MAXPOS),prev_mm2(MAXPOS),prev_sc2(MAXPOS)
integer :: this_pos_flag(MAXPOS),prev_pos_flag(MAXPOS)
character(len=2) :: this_classqc(MAXPOS),prev_classqc(MAXPOS),prev_sign(MAXPOS),prev_sign2(MAXPOS)
logical :: FOUND_P_PROF, FOUND_P_POS;
real ::     prevdata(MAXDAT,MAXLEV);
integer ::  prevflag(MAXDAT,MAXLEV);          ! flag for each level in profile
integer :: nprevlev
real    :: prevpresdata(MAXLEV),prevtempdata(MAXLEV),prevpsaldata(MAXLEV)
integer :: prevpresflag(MAXLEV),prevtempflag(MAXLEV),prevpsalflag(MAXLEV)
real :: this_jump_bin_temp(MAXDAT,MAXLEV), this_jump_bin_psal(MAXDAT,MAXLEV) ! 50 dbar bins for drive/jump check
real :: prev_jump_bin_temp(MAXDAT,MAXLEV), prev_jump_bin_psal(MAXDAT,MAXLEV)
real :: this_froz_bin_temp(MAXDAT),        this_froz_bin_psal(MAXDAT)        ! 100 dbar bottom bin for frozen profile
real :: prev_froz_bin_temp(MAXDAT,MAXLEV), prev_froz_bin_psal(MAXDAT,MAXLEV)

! ======================================================================================
! This controls how a qc18_frozen profile is handled.  if T, handle as a Gilson float
logical :: frozen_profile_fail, frozen_profile_rule1
logical :: temp_is_frozen, psal_is_frozen
integer :: frozen_profile_rule
! ======================================================================================
! A couple things for TESAC/GTS
! do we need to create a TESAC file for GTS?
logical :: send_to_GTS;			! do we need to create a TESAC file for GTS?
integer, parameter :: MAXLEV_GTS = 790;	! max nbr of levels in one profile1

! grey list
logical :: add_to_grey_list
integer, parameter :: MAX_GL = 5	 ! currently, have only 1 & 2
integer, parameter :: MAX_GL_PARAM = 4 ! PRES,TEMP,PSAL,DOZY (DOXY no yet present)
integer :: ngl, igl, tgl	! number of grey list entries grey list counter, which grey list entry applies (def=0)
integer :: ngp, igp   ! number of parameters in current entry and counter
character(len=10) :: grey_list_beg(MAX_GL),grey_list_end(MAX_GL)
integer :: igl_beg_yyyy, igl_beg_mm, igl_beg_dd, igl_end_yyyy, igl_end_mm, igl_end_dd
logical :: param_on_grey_list(MAX_GL,MAX_GL_PARAM)
logical :: pres_is_greylisted = .false.;
logical :: temp_is_greylisted = .false.;
logical :: psal_is_greylisted = .false.;
logical :: doxy_is_greylisted = .false.; ! Note there is not grey list for DOXY

integer :: grey_flag(MAX_GL,MAX_GL_PARAM)	! holds flags for multiple grey list entried
integer :: greyflag(MAX_GL_PARAM);		! holds final flags (for the time period)

! are either pres or psla adjusted?

logical :: use_spres_for_offset = .false.; ! Also self explanatory
logical :: use_spres_for_mrvsp  = .false.; ! set this also, in cases nlev=0, and no offset is used
logical :: use_mrvsp_for_offset = .false.; ! Also self explanatory

logical :: found_surf_pres = .false.; ! SURFACE PRESSURE from phy file
real :: surf_pres = FILL;             ! SURFACE PRESSURE from phy file
character(len=80) :: surf_pres_comment = "INIT SURFACE PRESSURE NOT AVAILABLE"; ! SURFACE PRESSURE from phy file

logical :: adjustments
! pressure adjustment
logical :: try_pres_adj = .false.; ! Is this float inst_type indicate pres_adj should be tested/done
logical :: do_pres_adj  = .false.; ! The pres_adj should be done (ie, nlev/=0, good pres_olffset found)
character(len=50) :: pres_adj_comment = "PRES ADJUSTMENT NOT REQUIRED"; ! NOT printed in qc file if try_pres_adj=F
logical :: found_pres_offset = .false.; ! Did we find a good pres_offset?
   real :: pres_offset = FILL;
character(len=8) :: cpres_offset
logical :: found_pres_mrvsp; real :: pres_mrvsp;	! most recent valid surface pres

logical :: found_pqc_mrvsp = .false.;
   real :: pqc_mrvsp = FILL;
logical :: found_pqc_mrcyc = .false.;
integer :: pqc_mrcyc = FILLINT;
   real :: out_mrvsp = FILL;
integer :: out_mrcyc = FILLINT;
character(len=40) :: pqc_mrvsp_comment = "INIT INP MRVSP NOT FOUND IN PQC FILE";
character(len=40) :: out_mrvsp_comment = "INIT OUT MRVSP NOT AVAILABLE";

! salinity adjustment
integer, parameter :: MAX_PA = 9 ! max nbr of psal adjust lines in meta file
logical :: do_psal_adj
logical :: do_psal_prof(MAXPRF);	! Include/exclude profiles for adjustment (STS is NOT adjusted; see STS_PROF above)

integer :: ipa, npa;
character(len=SZC) :: psal_adj_line(MAX_PA) ! The psal adjust lines - store all, use the last
character(len=SZC) :: psal_adj_used         ! The psal_adjust line that is actually used
integer :: psal_adj_dd, psal_adj_mm,psal_adj_yr,psal_adj_cyc
real :: psal_offset,psal_stddev;

! SPEED CHECK - max days to go back to find the previous qc file
integer, parameter :: DT_MAX_DAYS = 365

! regional test
logical :: do_regional_test = .false.; ! do we do the regional test
character(len=3) :: sea = 'N/A'				  ! Which regional sea is used

! Deepest pressure test needs the nominal

real    :: maximum_prof_pres  = FILL			! maximum expected profile pres (used in qc19)
character(len=80) :: maximum_prof_pres_from;		! where did the max pres come from?

real    :: nominal_prof_pres  = FILL			! Pres to use for QC19, initialize to fill value
real    :: nominal_prof_pres1 = FILL			! initialize to fill value
real    :: nominal_prof_pres2 = FILL			! initialize to fill value
integer :: nominal_prof_ncyc1 = 1			! initialize to "every cycle"
integer :: nominal_prof_ncyc2 = 0			! initialize to "no cycle"

integer :: icnt, nextra

! Keep track of any ZERO_LINE levels
logical :: zerolines1 = .false.; logical :: zerolines2 = .false.; logical :: zerolines = .false.; integer :: nzerolines = 0;
logical :: zeroline(0:2,MAXLEV);						! Is this particular line a zeroline?
logical :: zeropres1,zerotemp1,zeropsal1
logical :: zeropres2,zerotemp2,zeropsal2
logical :: zeropres3,zerotemp3,zeropsal3
logical :: zerodoxy1, zerodoxya, zerodoxyb, zerodoxyc

! Also take not of CHECKSUM=4 condition
logical :: cksm_bad = .false.;

! create arrays with only FILL & fill_qc values for the nc files:
real :: FILLDATA1(1,MAXLEV); character(len=1) :: FILLCHAR1(1,MAXLEV) ! for use in the nc files
real :: FILLDATA2(2,MAXLEV); character(len=1) :: FILLCHAR2(2,MAXLEV) ! for use in the nc files
real :: FILLNCD(MAXLEV); character(len=1) :: FILLNCF(MAXLEV) ! for use in the nc files

! see the "pri_prof" subdir in the tests directory (ie, I forgotnwhat this is suppoled to do) FIT THIS
logical :: check_primary; integer :: pres_start_cont

! moved from communications
logical           :: sampling_found;                       ! Did we find the sampling scheme?
integer           :: sampling_entries                      ! How many sampling records are there?
character(len=216) :: chr_VERT_SAMPLING_SCHEME(MAXPRF);    ! set in read_meta or read_phy
character(len=9)  :: VERT_SAMPLING_SCHEME_FROM             ! where is it set?

!============================================================================
! logical :: TESTMODE = .false.
character (len=SZC-40) :: prev_inpcols(MAXCOLS)
character(len=SZC) :: prev_datline(MAXLEV);                ! Stash the data lines for later

! The major arrays that are to be qc'd
logical :: hasPrvPRES;   integer :: prev_inprescol(MAXPRES)
logical :: hasPrvTEMP;   integer :: prev_intempcol(MAXTEMP)
logical :: hasPrvPSAL;   integer :: prev_inpsalcol(MAXPSAL)
logical :: hasLevFLAGS;  integer :: lev_inprescol, lev_intempcol, lev_inpsalcol;
character(len=1) :: presflag(MAXLEV), tempflag(MAXLEV), psalflag(MAXLEV);
integer :: prev_icolinp, prev_ncolinp;! from the column lines
real :: pdum(MAXLEV);                 ! in case the prev profile nedds to be reversed
real :: tdum(MAXLEV);                 ! in case the prev profile nedds to be reversed
real :: sdum(MAXLEV);                 ! in case the prev profile nedds to be reversed
integer pdumflag(MAXLEV);
integer tdumflag(MAXLEV);
integer sdumflag(MAXLEV);

logical :: DEEP_ARGO_FLOAT = .false.; ! Is this a deep argo float?

!============================================================================
! profile resolution columns 0 do we need tem?
! logical, parameter :: omitRSLN = .true.
! integer, parameter :: MAXRSLN = 9;
! logical :: hasRSLN(MAXRSLN)
! integer :: rslncol(MAXRSLN), rsln(MAXRSLN,MAXLEV)

!============================================================================

integer,parameter :: NTESTS = 22 ! The number of tests performed
integer :: QCTEST_FLAG(MAXPRF,NTESTS);
integer :: qcresult(MAXDAT,NTESTS)	! results of qc test for each test/column
integer :: qcdataresult(MAXDAT);		! results of qc test for each data column
integer :: qcprofresult(MAXPRF);		! results of qc test for each profile
integer :: maxresult
integer :: maxhistflag(MAXPRF)

!============================================================================
! TRAJECTORY arrays
logical :: hasMC , hasMC_qc, prev_hasMC_qc;     ! indicates if the phy/qc file has Measurement codes or not.
integer, parameter :: MAXMCL = 10000		! max number of mc lines (shuld be same as MAXHED)
character(len=SZC)	:: mcline(MAXMCL);	! what are the mc lines?
integer :: nmc, imc				! number of mc lines we have * counter
!============================================================================

integer :: maxflag
!============================================================================

! These arrays are used to quickly point to the CONFIG_MISSION_NUMBER for thos
! floats that altername nominal_profile_pres/cycles_for_profile_pres
integer :: mission12(MAXCYC3) =  (/ ( (/ 1, 2, 2 /), i = 1, MAXCYC3/3) /)
integer :: mission21(MAXCYC3) =  (/ ( (/ 1, 1, 2 /), i = 1, MAXCYC3/3) /)
integer :: mission13(MAXCYC4) =  (/ ( (/ 1, 2, 2, 2 /), i = 1, MAXCYC4/4) /)
integer :: mission31(MAXCYC4) =  (/ ( (/ 1, 1, 1, 2 /), i = 1, MAXCYC4/4) /)

! The following are needed by either nctraj or ncmeta, but also needed by read_pqc
logical :: PREV_TRANS_START_FOUND=.false.; ! needed by traj, but need to be visisble to read_pqc also.
logical :: has_irid_positioning = .false.; ! To indicate if the float has "Iridium" as positioning system anywhere in the profiles
logical :: fst_srf_actvtd; ! To indicate the value from "FSTSRF ACTIVATED STATUS" in John Gilson's files. Eg. float : 6807

! This would be in ncmeta_arrays, but is used by qcplus itslef
integer :: n_sensor_model_table
character(len=9), parameter :: CONFIG_incoming_fillval = '-999.0000'
integer :: nparam_meta  ! parameter count as per meta file
integer ::  configuration_mission_numbers(MAXCYC), cnt_missions = 0
character(len=4) prev_XMITS(MAXPOS),prev_SAT(MAXPOS),prev_posOrClass(MAXPOS)
character(len=4) prev_XMITS2(MAXPOS),prev_SAT2(MAXPOS),prev_posOrClass2(MAXPOS)
character(len=1)   :: prev_RTPQC(MAXPOS),prev_RTTQC(MAXPOS),prev_RTLQC(MAXPOS),prev_RTSQC(MAXPOS)
character(len=1)   :: prev_RTPQC2(MAXPOS),prev_RTTQC2(MAXPOS),prev_RTLQC2(MAXPOS),prev_RTSQC2(MAXPOS)
double precision   :: launch_time_juld=FILLDBLE;
integer :: pre_deploy_last_idx; !indicates the index of the last pre-deployment mc line in qc file.
integer :: nparam_traj = 0;
logical :: hasPRES_traj,hasTEMP_traj,hasPSAL_traj,hasCNDC_traj,hasCNDC_meta;
logical :: hasNCPAR_traj(MAXNCPAR), hasQCPAR_traj(MAXPAR);
logical :: cnvrt_cndc_traj=.false. !indicates if the  DRIFT CONDUCTIVITY in qc file needs conversion before inserting it

double precision  :: PREV_JULD=FILLDBLE;
character(len=1) :: PREV_JULD_QC=FILLCHAR;
character(len=25) :: prev_start_of_trans
double precision  :: prev_start_of_trans_juld
character(len=2) :: prev_classqc2(MAXPOS)

integer :: n_tech_param=0

logical :: bounce_float = .false.
logical :: orbcom_float = .false.

character(len=4) :: tcyc = "  -1";     ! test cycle number for get_all_qc, etc.

logical 	  :: qc_descending = .false.    ! Is the qc file for this profile descending
logical           :: phy_descending = .false.	! Is the phy file for this profile descending
character(len=1) :: direction_filename	= 'U';	! Direction of profile fro the FILENAME (if cyc is _nnnD, then DSC
character(len=1)  ::direction_configdir	= 'U';	! Direction of profile from the CONFIG DIRECTION in the phy file header
character(len=1) :: direction_DOPs 	= 'U';	! Direction of profile from DIRECTION OF PROFILES in the phy file header
character(len=1) :: direction_WMO_rec	= 'U';	! Direction of profile from the WMO recorder type -NOT RELIABLE
character(len=4) :: direction_summary;		! Summary of the types of direction clues


! ===========================================================================================================
! FORMAT_VERSION variables
! The version here is that of the FORMAT_VERSION, not the USER MANUAL for cycnum_conventions, UG3.1=UG3.2

logical :: profver3p0 = .false., profver3p1 = .false.;

character(len=4) :: FORMAT_VERSION;
character(len=4),parameter :: FORMAT_VERSION_3p0 = '3.0 ';
character(len=4),parameter :: FORMAT_VERSION_3p1 = '3.1 ';
character(len=4) :: prof_FORMAT_VERSION = '3.1'; ! defaulted, changable from the environment or the DOXY choice (core/biop
character(len=4) :: meta_FORMAT_VERSION = '3.1'; ! defaulted, changable from the environment
character(len=4) :: tech_FORMAT_VERSION = '3.1'; ! defaulted, changable from the environment
character(len=4) :: traj_FORMAT_VERSION = '3.1'; ! defaulted, changable from the environment or the DOXY choice (core/biop


character(len=4) :: USER_MANUAL;
character(len=4),parameter :: USER_MANUAL_3p0 = '3.03';
character(len=4),parameter :: USER_MANUAL_3p1 = '3.1 ';
character(len=4),parameter :: USER_MANUAL_3p2 = '3.2 ';
character(len=4) :: prof_USER_MANUAL; integer :: lumv; ! To be determined in qcprof
character(len=4) :: meta_USER_MANUAL = USER_MANUAL_3p2;
character(len=4) :: tech_USER_MANUAL = USER_MANUAL_3p2;
character(len=4) :: traj_USER_MANUAL = USER_MANUAL_3p2;

character(len=15) :: global_conventions
character(len=15),parameter :: global_conventions_3p0 = "Argo-3.0 CF-1.6";
character(len=15),parameter :: global_conventions_3p1 = "Argo-3.1 CF-1.6";
character(len=15),parameter :: global_conventions_3p2 = "Argo-3.2 CF-1.6";
character(len=15) :: prof_global_conventions; ! To be determined in qcprof
character(len=15) :: meta_global_conventions = global_conventions_3p2;
character(len=15) :: tech_global_conventions = global_conventions_3p2;
character(len=15) :: traj_global_conventions = global_conventions_3p2;
character(len=15) :: genl_global_conventions = global_conventions_3p2; ! Currently used for all except prof

! Note also that in the raj file , this is also used for CYCLE_NUMBER_ADJUSTED,CYCLE_NUMBER_INDEX,CYCLE_NUMBER_INDEX_ADJUSTED
character(len=65) :: cycnum_conventions
character(len=65),parameter :: cycnum_conventions_3p0      = "0..N, 0 : launch cycle (if exists), 1 : first complete cycle";
character(len=65),parameter :: prof_cycnum_conventions_3p1 = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";
character(len=65),parameter :: traj_cycnum_conventions_3p1 = "0...N, 0 : launch cycle, 1 : first complete cycle";
character(len=65),parameter :: prof_cycnum_conventions_3p2 = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";
character(len=65),parameter :: traj_cycnum_conventions_3p2 = "0...N, 0 : launch cycle, 1 : first complete cycle";

! prof UM3.1: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";	! p21
! prof UM3.2: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";	! p22
! traj UM3.1: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle, 1 : first complete cycle";			! p36
! traj UM3.2: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle, 1 : first complete cycle";			! p38
! tech same as prof
! tech UM3.1: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";	! p65
! tech UM3.2: CYCLE_NUMBER:conventions = "0...N, 0 : launch cycle (if exists), 1 : first complete cycle";	! p66

character(len=65) :: prof_cycnum_conventions; ! To be determined in qcprof
character(len=65) :: meta_cycnum_conventions = prof_cycnum_conventions_3p2; ! Actually, meta not used
character(len=65) :: tech_cycnum_conventions = prof_cycnum_conventions_3p2;
character(len=65) :: traj_cycnum_conventions = traj_cycnum_conventions_3p2;

! ===========================================================================================================
! Global variables used by ncbiop

integer, parameter :: MAXBPAR = 5; ! Note that this does NOT includ PRES or DOXY
integer, parameter :: MAXADNL = MAXBPAR*MAXPRF ! Theoretical nax number of addiitonal doxy values

logical :: hasBDOXY(MAXBPAR); ! Which BIOP parameters do we have (NOT including PRES or DOXY)

! This counter "locates" the correct elements of the information parameter arrays. 
! For example, if your FIRST output paramerer is FREQUENCY_DOXY, the BPAR_loc is 2
! For two ouput parameters, say TEMP_DOXY & RPHASE_DOXY, THE BPAR_locS ARE 1 & 6

! BR special parameters DOXY2A/B &  DOXY2A/B/C There is only one pres profile these are handled differently -
! Or at least, later, after the "normal" BR files are being produce.
!
! inst_type APEXIR_TSO7 has DOXY2A/B, and there are 6 floats, all dead 
! the columns are:
! DISSOLVED OXYGEN2A (MICROMOL/KG)  DISSOLVED OXYGEN2B (MICROMOL/KG)
! TEMP_DOXY2A (DEG C)               TEMP_DOXY2B (DEG C)
! TPHASE_DOXY2A                     TPHASE_DOXY2B
! RPHASE2A (no _DOXY)               RPHASE2B (no _DOXY)
!
! inst_type APEXIR_TSO9 has DOXY2A/B/c, and there are 2 floats, both alive (as of mid-Dec 2017)
! The columns are:
! DISSOLVED OXYGEN2A (MICROMOL/KG)  DISSOLVED OXYGEN2B (MICROMOL/KG)  DISSOLVED OXYGEN2C (MICROMOL/KG)
! TEMP_DOXY2A (DEG C)               TEMP_DOXY2B (DEG C)               TEMP_DOXY2C (DEG C)
! TPHASE2A (mp +DOXY)               PHASE_DOXY2B                      TPHASE_DOXY2C

integer :: BPAR_loc(MAXBPAR); ! what is the location in the input infomation arrays

! If the parameters are changed/added to, change also read_phy
character(len=64) :: BPARNAME(MAXBPAR) = (/ &
"TEMP_DOXY                                                       ", &
"FREQUENCY_DOXY                                                  ", &
"BPHASE_DOXY                                                     ", &
"TPHASE_DOXY                                                     ", &
"RPHASE_DOXY                                                     " /);
!1234567890123456789012345678901234567890123456789012345678901234 - 64 chars

character(len=64) :: bparpres = "PRES                                                            ";
character(len=64) :: bpardoxy = "DOXY                                                            ";

character(len=64) :: bpartemp; equivalence (BPARNAME(1),bpartemp);
character(len=64) :: bparfreq; equivalence (BPARNAME(2),bparfreq);
character(len=64) :: bparbpha; equivalence (BPARNAME(3),bparbpha);
character(len=64) :: bpartpha; equivalence (BPARNAME(4),bpartpha);
character(len=64) :: bparrpha; equivalence (BPARNAME(5),bparrpha);

! Does a given paramet5er exist (the "0"), and in which profile(s) (1 to MAXPRF)
logical :: hasTEMP_DOXY(0:MAXPRF) = (/.false.,.false.,.false.,.false.,.false./);
logical :: hasFREQ_DOXY(0:MAXPRF) = (/.false.,.false.,.false.,.false.,.false./);
logical :: hasBPHA_DOXY(0:MAXPRF) = (/.false.,.false.,.false.,.false.,.false./);
logical :: hasTPHA_DOXY(0:MAXPRF) = (/.false.,.false.,.false.,.false.,.false./);
logical :: hasRPHA_DOXY(0:MAXPRF) = (/.false.,.false.,.false.,.false.,.false./);
character(len=SZC) :: bnamadnl(MAXHED);    ! Stash the addiitonal DOXY header lines (NOT including PRES/DOXY)
! Defined elsewhere: profbdata(MAXBPAR,MAXPRF,MAXLEV);
integer :: nadnlcols; ! How many colums do we add besides PRES & DOXY?
integer :: bcoladnl(MAXADNL), bparadnl(MAXADNL), bprfadnl(MAXADNL); 

character(len=64) :: BPAR_lname(MAXBPAR) = (/ &
"Sea temperature from oxygen sensor ITS-90 scale                 ", & ! TEMP_DOXY
"Frequency reported by oxygen sensor                             ", & ! FREQUENCY_DOXY
"Uncalibrated phase shift reported by oxygen sensor              ", & ! BPHASE_DOXY
"Uncalibrated phase shift reported by oxygen sensor              ", & ! TPHASE_DOXY
"Uncalibrated red phase shift reported by oxygen sensor          " /) ! RPHASE_DOXY
logical :: BPAR_use_snam(MAXBPAR) = (/ .true., .false., .false., .false., .false. /);
character(len=64) :: BPAR_sname(MAXBPAR) = (/ &
"temperature_of_sensor_for_oxygen_in_sea_water                   ", &
"n/a                                                             ", &
"n/a                                                             ", &
"n/a                                                             ", &
"n/a                                                             " /);
!................................................................", &

character(len=64) :: BPAR_units(MAXBPAR) = (/ &
"degree_Celsius                                                  ", &
"hertz                                                           ", &
"degree                                                          ", &
"degree                                                          ", &
"degree                                                          " /);
!................................................................", &
!                                  TEMP_D  .REQ_D   BPH_D   TPH_D   RPH_D
logical :: BPAR_mnmx(MAXBPAR) = (/.true.,  .true., .true., .true., .true. /);
real    :: BPAR_vmin(MAXBPAR) = (/  -2.0,     0.0,   10.0,   10.0,   10.0 /);
real    :: BPAR_vmax(MAXBPAR) = (/  40.0, 25000.0,   70.0,   70.0,   70.0 /);

character(len=64) :: BPAR_prof_qc_lname(MAXBPAR) = (/ &
"Global quality flag of TEMP_DOXY profile                       ", & ! TEMP_DOXY
"Global quality flag of FREQUENCY_DOXY profile                  ", & ! FREQUENCY_DOXY
"Global quality flag of BPHASE_DOXY profile                     ", & ! BPHASE_DOXY
"Global quality flag of TPHASE_DOXY profile                     ", & ! TPHASE_DOXY
"Global quality flag of RPHASE_DOXY profile                     " /) ! RPHASE_DOXY
!................................................................", &

integer :: ibparinp(MAXHED);              ! Which parameter number goes wih which doxyline

! These are the non-Qc'd parameters
real    profbdata(MAXBPAR,MAXPRF,MAXLEV); ! The bio data itself (MAXBPAR=5, MAXPRF=4

logical :: hasBPRFinp(MAXBPAR,MAXPRF);   ! If so, which profile actually has data in it?
logical :: hasBPRFout(MAXBPAR+2,MAXPRF); ! Include PRES & DOXY
integer :: BPARCOL(MAXBPAR,MAXPRF)

character(len=64) :: bparout(MAXBPAR+2)	! What are the names of the OUTPUT doxy parameters?

! ===========================================================================================================
logical :: hasBPAR(MAXBPAR)  ! Is this parameter present in the input file?

integer :: nbdx, ibdx, nbout, ibout, ibinp;
integer :: ibpar, nbpar, ibprf, nbprf;
integer :: nbparadnl                      ! This number does NOT include PRES or DOXY
integer :: nbparout;                      ! This number INCLUDES the PRES & DOXY parmetersi
character(len=64) :: BPARNAMEout(MAXBPAR)

integer :: posloc=0,posloc_iri=0,posloc_noniri=0; ! which position was selected in pick_profile_pos?
character(len=70) :: posn_cmnt = "N/A";          ! Why was that position selelcted?

! Ice Detect Information
! We cannot depend on the "values" given by any Ice Detect status-type line.
! The best we can do is only to detect when the Ice Detect MAY be operational.
!
! Each different calss of floats (APEX,APEXIR,NAVISIR,SOLO,other) uses a different
! set of header lines to indicate Ice Detection, howere, all of htem have at least
! one line containing "ICE DETECT"
!
! ICE INT - variables conneted to interpolated positions/ice float
! ICE INT status is to be written to qc file only if ice lines present OR there is an interpolation problem
! ICE INT STATUS =  0 posiiton is NOT interpolated, all ok
! ICE INT STATUS =  1 means ice lines present, position interpolated, npos=1, posloc=1, all ok
! ICE INT STATUS =  2 means ice lines present, position interpolated, npos=3, posloc=2, all ok
! ICE INT STATUS = -1 means the position is interpolated, but NO ice lines are present in the phy file
! ICE INT STATUS = -2 means ice lines present, position interpolated, npos=3, but psns 1 and/or 3 are NOT 999 lines
! ICE INT STATUS = -3 means ice lines present, position interpolated, but the number of positions is NOT 1 or 3
integer           :: ice_int_status = 0;         ! ICE INT status - add to qc file if ice lines are present OR interpolation problem
character(len=85) :: ice_int_cmnt;               ! ICE INT comment - what the status means
character(len=85) :: ice_int_cmnts(-3:2) = (/ &  ! ICE INT coomNTnts - all the possibilities
'-3 = ice lines & interp posn present, number of posns not 1 or 3, set posloc=0, error', &
'-2 = ice lines & interp posn present, posns 1 and/or 3 NOT 999, set posloc=0, error  ', &
'-1 = ice lines NOT present, interp posn present, set posloc=0, error                 ', &
'0 = posiiton is NOT interpolated, all ok, all ok                                     ', &
'1 = ice lines & interp posn present, npos=1, posloc=1, all ok                        ', &
'2 = ice lines & interp posn present, npos=3, posloc=2, all ok                        '/);
!1234567890123456789012345678901234567890123456789012345678901234567890123456789012345'
logical :: ice_lines_present = .false.    ; ! ICE INT Are there ice lines in the py file?
logical :: position_interpolated = .false.; ! ICE INT We did find an interpolated position?
logical :: DBGINTICE = .false.;             ! ICE INT do we want the special debugs for iceint?
logical :: bad_position_found(MAXPOS) = .false.;

character(len=30) :: dblstr;

integer :: gmtdate(6) , gmtyr , gmtmo , gmtdy , gmthh , gmtmm , gmtss
equivalence (gmtdate(1),gmtyr), (gmtdate(2),gmtmo),(gmtdate(3),gmtdy),(gmtdate(4),gmthh),(gmtdate(5),gmtmm),(gmtdate(6),gmtss)

! BADGPS - do we have a bad gps fix of ANY kind?
logical :: badgpsfix_found = .false.; ! If needed, it is set if check_cdate_qc finds a 'G' returned from test_for_badgps

! BADS found in a "regular" time in the header
logical :: badgpsfix_ascnt = .false.;  ! TIME ASCENT END
logical :: badgpsfix_strns = .false.;  ! START OF TRANSMISSION
logical :: badgpsfix_epoch = .false.;  ! START OF TRANSMISSIONN FROM EPOCH
logical :: badgpsfix_upload = .false.; !PROFILE_UPLOAD_TIME

logical :: badgpsfix_MC100 = .false.;
logical :: badgpsfix_MC600 = .false.;
logical :: badgpsfix_MC700 = .false.;
logical :: badgpsfix_MC090 = .false.;
logical :: badgpsfix_MC099 = .false.;
logical :: badgpsfix_MC589 = .false.;
logical :: badgpsfix_MC590 = .false.;

character(len=14) :: TIME_TEST; character(len=1) :: TIME_TEST_QC, this_date_time_qc(MAXPOS); ! Used to check for badgpsfix condition

logical :: badgpsfix_pos_found = .false.;
logical :: badgpsfix_pos(MAXPOS)  = .false.;

! ===========================================================================================================
! General stuff ysed by any module
character(len=1) :: SQ = "'"; ! Single quote
character(len=1) :: DQ = '"'; ! Double quote
character(len=1) :: DS = "$"; ! Dollar sign
! ===========================================================================================================

contains

! ===========================================================================================================
! module functions/subroutines
! ===========================================================================================================

subroutine parse_dataline(caller,lev) ! This will load the data values into the arrays
    character(*), intent( in )     :: caller
    integer, intent( in )          :: lev ! Level to load
    integer :: ipar
    character(len=14), parameter   :: sub = 'parse_dataline';

    ! before maybe jumping the data trans, sotre any checksum value
    do iprf=1,MAXPRF
	if(hasCKSM(iprf))then
	    cksm(iprf,lev) = ifix(datvals(ncksmcol(iprf)));
	endif
    enddo

    ! 15Oct2014 complete revision of this - now, only declare ZEROLIN if ALL avlues in a given profile are zero.

    ! Check primary profile for all zeroes
    if(hasDATA(p1r))then; if(datvals(inprescol(1))/= 0.0)then; go to 10; endif; endif;
    if(hasDATA(t1r))then; if(datvals(intempcol(1))/= 0.0)then; go to 10; endif; endif;
    if(hasDATA(s1r))then; if(datvals(inpsalcol(1))/= 0.0)then; go to 10; endif; endif;
    if(hasDATA(do1))then; if(datvals(indoxycol(1))/= 0.0)then; go to 10; endif; endif;

    ! All values in primary profile are exactly zero, so mark as zeroline1
    zerolines1 = .true.; zeroline(1,lev) = .true.;

    if(DEBUG) write(DBG,'("PPQC_ZM profile 1 is all zeros for lev: ",i4)')lev

    if(hasDATA(p1r))then; datvals(inprescol(1)) = FILL; profflag(p1r,lev) = MISSING; endif
    if(hasDATA(t1r))then; datvals(intempcol(1)) = FILL; profflag(t1r,lev) = MISSING; endif
    if(hasDATA(s1r))then; datvals(inpsalcol(1)) = FILL; profflag(s1r,lev) = MISSING; endif
    if(hasDATA(do1))then; datvals(indoxycol(1)) = FILL; profflag(do1,lev) = MISSING; endif

    10 continue

    ! Now, check the secondary profile IF THERE IS ONE - note that we DO need to check, as ~5 inst types
    ! are APEX, npt APEXIR, so check vor zerolines - note, however, that DOXY is NOT checked - these only
    ! floats with DOXY2are all in <>IR floats

    if(.not. hasDATA(p2r)) go to 20; ! There is no profile 2
    if(hasDATA(p2r))then; if(datvals(inprescol(2))/= 0.0)then; go to 20; endif; endif;
    if(hasDATA(t2r))then; if(datvals(intempcol(2))/= 0.0)then; go to 20; endif; endif;
    if(hasDATA(s2r))then; if(datvals(inpsalcol(2))/= 0.0)then; go to 20; endif; endif;

    ! All values in secondary profile are exactly zero, so mark as zeroline2
    zerolines2 = .true.; zeroline(2,lev) = .true.;

    if(DEBUG) write(DBG,'("PPQC_ZM profile 2 is all zeros for lev: ",i4)')lev

    if(hasDATA(p2r))then; datvals(inprescol(2)) = FILL; profflag(p2r,lev) = MISSING; endif
    if(hasDATA(t2r))then; datvals(intempcol(2)) = FILL; profflag(t2r,lev) = MISSING; endif
    if(hasDATA(s2r))then; datvals(inpsalcol(2)) = FILL; profflag(s2r,lev) = MISSING; endif

    20 continue

    ! Now, evaluate if zerolines are present
    if(zerolines1 .or. zerolines2)then
	zerolines = .true.; zeroline(0,lev) = .true.; nzerolines = nzerolines + 1;
	if(DEBUG) write(DBG,'("PPQC_ZM zerolines1 or 2 are true, we have zerolines for lev: ",i4)')lev;
    endif

    if(hasDATA(p1r))then
	if(datvals(inprescol(1)) /= FILL)then
	    profdata(p1r,lev) = datvals(inprescol(1))
	    !CA write(cdatvals,*)datvals(inprescol(1))
	    !CA res = LEN(cdatvals(index(TRIM(cdatvals),".")+1:LEN(TRIM(cdatvals))))
            res=get_resolution(datvals(inprescol(1)));
	    if(resolution(1) < res)resolution(1)=res;
	    profflag(p1r,lev) = NOQC
	endif
    endif
    if(hasDATA(p2r))then
	if(datvals(inprescol(2)) /= FILL)then
	    profdata(p2r,lev) = datvals(inprescol(2))
	    profflag(p2r,lev) = NOQC
	endif
    endif
    if(hasDATA(p3r))then
        if(datvals(inprescol(3)) /= FILL)then
            profdata(p3r,lev) = datvals(inprescol(3))
            profflag(p3r,lev) = NOQC
        endif
    endif
    if(hasDATA(p4r))then
	if(datvals(inprescol(4)) /= FILL)then
	    profdata(p4r,lev) = datvals(inprescol(4))
	    profflag(p4r,lev) = NOQC
	endif
    endif

    if(hasDATA(t1r))then
	if(datvals(intempcol(1)) > FILT)then ! some apex_tso use -9 for temp fill value
	    profdata(t1r,lev) = datvals(intempcol(1))
	    !CA write(cdatvals,*)datvals(intempcol(1))
	    !CA res = LEN(cdatvals(index(TRIM(cdatvals),".")+1:LEN(TRIM(cdatvals))))
            res=get_resolution(datvals(intempcol(1)));
	    if(resolution(2) < res)resolution(2)=res;
	    profflag(t1r,lev) = NOQC
	endif
    endif
    if(hasDATA(t2r))then
	if(datvals(intempcol(2)) > FILT)then ! some apex_tso use -9 for temp fill value
	    profdata(t2r,lev) = datvals(intempcol(2))
	    profflag(t2r,lev) = NOQC
	endif
	if(interp_status)then
	    interp_t2(lev) = ifix(datvals(interp_t2_col));
	    profflag(t2r,lev) = interp_t2(lev);
	endif
    endif
    if(hasDATA(t3r))then
        if(datvals(intempcol(3)) > FILT)then ! some apex_tso use -9 for temp fill value
            profdata(t3r,lev) = datvals(intempcol(3))
            profflag(t3r,lev) = NOQC
        endif
        if(interp_status)then
            interp_t3(lev) = ifix(datvals(interp_t3_col));
            profflag(t3r,lev) = interp_t3(lev);
        endif
    endif
    if(hasDATA(t4r))then
	if(datvals(intempcol(4)) > FILT)then ! some apex_tso use -9 for temp fill value
	    profdata(t4r,lev) = datvals(intempcol(4))
	    profflag(t4r,lev) = NOQC
	endif
	if(interp_status)then
	    interp_t4(lev) = ifix(datvals(interp_t4_col));
	    profflag(t4r,lev) = interp_t4(lev);
	endif
    endif


    if(hasCNDC(0))then
	do iprf=1,MAXPRF
	    if(hasCNDC(iprf))then
		if(datvals(incndccol(iprf)) /= FILL)then
		    profcndc(iprf,lev) = datvals(incndccol(iprf))
		endif
	    endif
	enddo
    endif

    if(hasDATA(s1r))then
	if(datvals(inpsalcol(1)) /= FILL)then
	    profdata(s1r,lev) = datvals(inpsalcol(1))
	    !CA write(cdatvals,*)datvals(inpsalcol(1))
	    !CA res = LEN(cdatvals(index(TRIM(cdatvals),".")+1:LEN(TRIM(cdatvals))))
            res=get_resolution(datvals(inpsalcol(1)));
	    if(resolution(3) < res)resolution(3)=res;
	    profflag(s1r,lev) = NOQC
	endif
    endif

    if(hasDATA(s2r))then
	if(datvals(inpsalcol(2)) /= FILL)then
	    profdata(s2r,lev) = datvals(inpsalcol(2))
	    profflag(s2r,lev) = NOQC
	endif
	if(interp_status)then
	    interp_s2(lev) = ifix(datvals(interp_s2_col));
	    profflag(s2r,lev) = interp_s2(lev);
	endif
    endif
    if(hasDATA(s3r))then
        if(datvals(inpsalcol(3)) /= FILL)then
            profdata(s3r,lev) = datvals(inpsalcol(3))
            profflag(s3r,lev) = NOQC
        endif
        if(interp_status)then
            interp_s3(lev) = ifix(datvals(interp_s3_col));
            profflag(s3r,lev) = interp_s3(lev);
        endif
    endif
    if(hasDATA(s4r))then
	if(datvals(inpsalcol(4)) /= FILL)then
	    profdata(s4r,lev) = datvals(inpsalcol(4))
	    profflag(s4r,lev) = NOQC
	endif
	if(interp_status)then
	    interp_s4(lev) = ifix(datvals(interp_s4_col));
	    profflag(s4r,lev) = interp_s4(lev);
	endif
    endif


    if(hasDOXY(1))then
	if(datvals(indoxycol(1)) /= FILL)then
	    profdata(do1,lev) = datvals(indoxycol(1))
	    profflag(do1,lev) = NOQC;
	endif
    endif
    if(hasDOXY(2))then
	if(datvals(indoxycol(2)) /= FILL)then
	    profdata(do2,lev) = datvals(indoxycol(2))
	    profflag(do2,lev) = NOQC;
	endif
    endif
    if(hasDOXY(3))then
	if(datvals(indoxycol(3)) /= FILL)then
	    profdata(doa,lev) = datvals(indoxycol(3))
	    profflag(doa,lev) = NOQC;
	endif
    endif
    if(hasDOXY(4))then
	if(datvals(indoxycol(4)) /= FILL)then
	    profdata(dob,lev) = datvals(indoxycol(4))
	    profflag(dob,lev) = NOQC;
	endif
    endif
    if(hasDOXY(5))then
	if(datvals(indoxycol(5)) /= FILL)then
	    profdata(doc,lev) = datvals(indoxycol(5))
	    profflag(doc,lev) = NOQC;
	endif	
    endif

    ! 20171016 Add the BioARGO data parse

    if(DOXY_float .and. (nbparadnl> 0))then

	do i=1,nadnlcols
	    profbdata(bparadnl(i),bprfadnl(i),lev) = datvals(bcoladnl(i));
	enddo
    endif; ! DOXY_float


end subroutine parse_dataline

  !===================================================================================!
logical function skip_test(flag)      ! If the flag is ok, run the test
    integer, intent( in )          :: flag ! flag value to be tested

    skip_test = .false.
    if(flag > PROB_GOOD) skip_test = .true.
    if(flag == 8) skip_test = .false.           ! 8 = INTERP

end function skip_test

  !===================================================================================!

subroutine dbg_levs(caller,beg,iend,lmsg)
    character(*), intent( in )     :: caller
    integer,      intent( in )     :: beg
    integer,      intent( in )     :: iend
    character(*), intent( in )     :: lmsg
    character(len=8), parameter   :: sub = 'dbg_levs';
    character(len=SZC) :: pfl,tfl,sfl
    integer :: end,j,ic     ! out - testflags for top or bottom of profile

    if((iend-beg) > 10)then
	write(msg,'("In dbg_levs, max of 10 levs exceeded, reset to end=beg_9")'); call wrtdbg(sub,msg);
	end = beg + 9;
    elseif(iend > nlev)then
	write(msg,'("In dbg_levs, end > nlev, reset to end=nlev")'); call wrtdbg(sub,msg);
	end = nlev;
    else
	end = iend
    endif
    write(msg,'("======================================================")'); call wrtdbg(sub,msg);
    call wrtdbg(sub,msg);
    write(msg,*)'    hasDATA: ',hasDATA; call wrtdbg(sub,msg);
    write(msg,'("======================================================")'); call wrtdbg(sub,msg);
    write(msg,'(11x,10i9)')(i,i=beg,end)
    write(msg,'(a)')trim(msg); call wrtdbg(sub,msg);
    do idat=1,13
        if(hasDATA(idat))write(msg,'(i2.2,x,a,2x,10F9.2)')idat,datname(idat),(profdata(idat,i),i=beg,end)
	call wrtdbg(sub,msg);
    enddo
	
    write(msg,'("-----------------------------------------------------------------------")'); call wrtdbg(sub,msg);
    write(pfl,'("overall,  Pf:",3i1,x,3i1)')profflag(1,beg),profflag(5,beg),profflag(7,beg),(testflag(1,beg,i),i=1,3)
    write(tfl,'("          Tf: ", 6i1)')(testflag(5,beg,i),i=1,6)
    write(sfl,'("          Sf:",7i1,2x)')(testflag(5,beg,i),i=1,7)
    do ilev=beg+1,end
	write(pfl,'(a,2x,3i1,x,3i1,3x)')trim(pfl),profflag(1,ilev),profflag(5,ilev),profflag(7,ilev),(testflag(1,ilev,i),i=1,3)
	write(tfl,'(a,3x,6i1)')trim(tfl),(testflag(5,ilev,i),i=1,6)
	write(sfl,'(a,2x,7i1)')trim(sfl),(testflag(7,ilev,i),i=1,7)
    enddo
    write(msg,'(a)')trim(pfl); call wrtdbg(sub,msg);
    write(msg,'(a)')trim(tfl); call wrtdbg(sub,msg);
    write(msg,'(a)')trim(sfl); call wrtdbg(sub,msg);
    write(msg,'("-----------------------------------------------------------------------")'); call wrtdbg(sub,msg);
    write(msg,'(a)')lmsg; call wrtdbg(sub,msg);
    write(msg,'("=======================================================================")'); call wrtdbg(sub,msg);
    11 format(11x,5i9,2x,5i2,2x,3i2,x,6i2,x,7i2)
    12 format(i2.2,x,a,2x,5F9.2,2x,5i2,2x,3i2,x,6i2,x,7i2)
end subroutine dbg_levs

  !===================================================================================!

subroutine reverse_profdata(caller) ! This will reverse the data values into the arrays
    character(*), intent( in )     :: caller
    character(len=16), parameter   :: sub = 'reverse_profdata';
    do idat=1,MAXDAT
        if(hasDATA(idat))then;
	    do ilev=1,nlev
		rdum(ilev) = profdata(idat,ilev)
		idum(ilev) = profflag(idat,ilev)
	    enddo
	    i=nlev+1
	    do ilev=1,nlev
		i=i-1;
		profdata(idat,ilev) = rdum(i);
		profflag(idat,ilev) = idum(i);
	    enddo
	endif
    enddo
    do iprf=1,MAXPRF
	if(hasCKSM(iprf))then;
	    do ilev=1,nlev
		idum(ilev) = cksm(iprf,ilev)
	    enddo
	    i=nlev+1
	    do ilev=1,nlev
		i=i-1;
		cksm(iprf,ilev) = idum(i);
	    enddo

	endif
    enddo
    if(hasCNDC(0))then
	do iprf=1,MAXPRF
	    if(hasCNDC(iprf))then
		if(DEBUG)then; write(msg,'("dd.reverse_profdata reverses CNDC for iprf: ",i2)')iprf; endif
		do ilev=1,nlev
		    rdum(ilev) = profcndc(iprf,ilev)
		enddo
		i=nlev+1
		do ilev=1,nlev
		    i=i-1;
		    profcndc(iprf,ilev) = rdum(i);
		enddo
	    endif
	enddo
    endif
    if(DOXY_float)then
	if(hasTEMP_DOXY(0))then 
	    do iprf=1,nprf 
		do ilev=1,nlev; rdum(ilev) = profbdata(1,iprf,ilev); enddo 
		i=nlev+1; do ilev=1,nlev; i=i-1; profbdata(1,iprf,ilev) = rdum(i); enddo 
	    enddo 
	endif
	if(hasFREQ_DOXY(0))then 
	    do iprf=1,nprf 
		do ilev=1,nlev; rdum(ilev) = profbdata(2,iprf,ilev); enddo 
		i=nlev+1; do ilev=1,nlev; i=i-1; profbdata(2,iprf,ilev) = rdum(i); enddo 
	    enddo 
	endif
	if(hasBPHA_DOXY(0))then 
	    do iprf=1,nprf 
		do ilev=1,nlev; rdum(ilev) = profbdata(3,iprf,ilev); enddo 
		i=nlev+1; do ilev=1,nlev; i=i-1; profbdata(3,iprf,ilev) = rdum(i); enddo 
	    enddo 
	endif
	if(hasTPHA_DOXY(0))then 
	    do iprf=1,nprf 
		do ilev=1,nlev; rdum(ilev) = profbdata(4,iprf,ilev); enddo 
		i=nlev+1; do ilev=1,nlev; i=i-1; profbdata(4,iprf,ilev) = rdum(i); enddo 
	    enddo 
	endif
	if(hasRPHA_DOXY(0))then 
	    do iprf=1,nprf 
		do ilev=1,nlev; rdum(ilev) = profbdata(5,iprf,ilev); enddo 
		i=nlev+1; do ilev=1,nlev; i=i-1; profbdata(5,iprf,ilev) = rdum(i); enddo 
	    enddo 
	endif
    endif

end subroutine reverse_profdata

  !===================================================================================!

subroutine set_proftopbottot(caller)
    character(*), intent( in )     :: caller
    character(len=16), parameter   :: sub = 'set_prftopbottot';

    do idat=1,MAXDAT
        if(hasDATA(idat))then;
            proftop(idat) = 1;
            do ilev=nlev+1,2,-1
		profbot(idat) = 1;
                if(profdata(idat,ilev) /= FILL)then
                    profbot(idat) = ilev;
                    go to 10;   ! Got it
                endif
            enddo
            10 continue
            proftot(idat) = profbot(idat) - proftop(idat) + 1;
        endif           ! if hasDATA
    enddo               ! do idat...MAXDAT
end subroutine set_proftopbottot

  !===================================================================================!

subroutine fill_profvect_tmp(proftot_tmp,profdata_tmp,ilev_vector_tmp,ifl)
    integer :: i;
    integer,intent(in) :: ifl
    integer, intent(out ) :: proftot_tmp
    real, intent(out) ::  profdata_tmp(MAXLEV);
    integer, intent(out)::  ilev_vector_tmp(MAXLEV);
    character(len=16) :: sub = 'fill_prfvect_tmp';

    proftot_tmp=0
    if(DBLEV== 8)then
	write(msg,*)"fl_pv_tmp idat ,proftop(idat),profbot(idat)",idat,proftop(idat),profbot(idat), &	! LEV08
	 	profdata(idat,proftop(idat)),profdata(idat,profbot(idat))				! LEV08
    endif
    do i=proftop(idat),profbot(idat)
	if(profflag(idat,i) >= PROB_BAD)then
	    if(DBLEV>999) write(msg,'("Dummy write statement")');
	else
	    proftot_tmp = proftot_tmp+1;
	    profdata_tmp(proftot_tmp)= profdata(idat,i);
	    ilev_vector_tmp(proftot_tmp)=i
	endif
    enddo
    if(DBLEV== 8)then												! LEV08
	write(msg,'("fill_profvect_tmp number of good lines: ",i4)')proftot_tmp; call wrtdbg(sub,msg);		! LEV08
	do i=1,proftot_tmp											! LEV08
	    write(msg,'("fill_profvect_tmp oldlev: ",i4," newlev: ",i4," flag: ",i3," data: ",f10.2)') &	! LEV08
		ilev_vector_tmp(i), i,testflag(idat,ilev_vector_tmp(i),ifl),profdata_tmp(i)			! LEV08
		call wrtdbg(sub,msg);										! LEV08
	enddo													! LEV08
    endif													! LEV08
end subroutine fill_profvect_tmp
  !===================================================================================!

subroutine median_check(proftot_tmp,profdata_tmp,ilev_vector_tmp,filter_width,filter_tolerance,nbad)

    integer, intent(in ) :: proftot_tmp,filter_width
    integer :: m,mw,nbad,ilev_tmp,tmp
    integer, intent(in ) ::  ilev_vector_tmp(MAXLEV)
    integer :: wing		! size of "sides: of filter: (filter_width-1)/2 (do NOT assume truncation of fw/2).
    integer :: imd		! counter for filling x()
    real :: x(MAXLEV),xmed;
    real, intent(in) ::  profdata_tmp(MAXLEV),filter_tolerance

    character(len=12), parameter :: sub = 'median_check'

    wing = (filter_width-1)/2;

    if(DBLEV== 8)then;
	write(msg,*)'LEV08 MEDIAN PARAMETERS BEG median_check filter_width,filter_tolerance: ',filter_width,filter_tolerance
	call wrtdbg(sub,msg);
	do ilev=1,proftot_tmp
	    call wrtdbgdataflags("LEV08 median_check in data_arrays BEG",idat,ilev,1,3) ! There are three PRES flags
	enddo
    endif
    do m = wing+1,proftot_tmp-wing
        do i=1,filter_width
	    imd = m-wing+i-1;		! Trust me, it works
            x(i)=profdata_tmp(imd);
        enddo
	if(DBLEV==30)then; write(msg,'("LEV30 median_check m,x: ",i5,2x,11f9.2)')m, &
			(x(i),i=1,filter_width); call wrtdbg(sub,msg); endif

        call median(x,filter_width,xmed)

        if(abs( xmed - profdata_tmp(m)) > filter_tolerance)then
	    write(msg,*)'CLAUDIA2 (med,p,difference) ',xmed,profdata_tmp(m),abs(xmed-profdata_tmp(m)); call wrtdbg(sub,msg);
            nbad = nbad+1;
            ilev_tmp = ilev_vector_tmp(m);
            testflag(idat,ilev_tmp,pres_incr_ncol)=BAD
	    if(DBLEV== 8)then
		write(msg,*)'LEV08 median_check in data_arrays idat,ilev_tmp,ifl: ',idat,ilev_tmp,pres_incr_ncol,' flag set to BAD'
		call wrtdbg(sub,msg);
	    endif
        endif
    enddo

    if(DBLEV== 8)then
	do ilev=1,proftot_tmp
	    call wrtdbgdataflags("LEV08 median_check in data_arrays END",idat,ilev,1,3) ! There are three PRES flags
	enddo
    endif

    if(DBLEV== 8)then;
	write(msg,*)'MEDIAN PARAMETERS END median_check filter_width,filter_tolerance: ',filter_width,filter_tolerance
	call wrtdbg(sub,msg);
    endif

end subroutine median_check

  !===================================================================================!
subroutine median(x, n, xmed)

  ! Find the median of X(1), ... , X(N), using as much of the quicksort
  ! algorithm as is needed to isolate it.
  ! N.B. On exit, the array X is partially ordered.
  !     Latest revision - 26 November 1996

    IMPLICIT NONE

    INTEGER n
    REAL    x(*),xmed

    ! Local variables

    REAL     temp, xhi, xlo, xmax, xmin
    LOGICAL  odd
    INTEGER  hi, lo, nby2, nby2p1, mid, i, j, k
    character prg_sub*6
    data     prg_sub /'median'/

    nby2 = n / 2
    nby2p1 = nby2 + 1
    odd = .true.

!   HI & LO are position limits encompassing the median.

    if(n .eq.(2* nby2)) odd = .false.

    lo = 1
    hi = n
    if(n .lt. 3)then;
        if(n .lt. 1)then;
	    xmed = 0.0
    	    return
        endif
        xmed = x(1)
        if(n .eq. 1) return
    	    xmed = 0.5*(xmed + x(2))
    	    return
        endif

!     Find median of 1st, middle & last values.

 10     mid = (lo + hi)/2
        xmed = x(mid)
        xlo = x(lo)
        xhi = x(hi)
        if(xhi .lt. xlo)then;          ! Swap xhi & xlo
        	    temp = xhi
             xhi = xlo
             xlo = temp
        endif
        if(xmed .gt. xhi)then;
            xmed = xhi
        elseif(xmed .lt. xlo)then;
            xmed = xlo
        endif

! The basic quicksort algorithm to move all values <= the sort key (XMED)
! to the left-hand end, and all higher values to the other end.

        i = lo
        j = hi

 50     continue

        if(x(i) .ge. xmed) go to 60
	i = i + 1
        go to 50
 60     continue

        if(x(j) .le. xmed) go to 70
        j = j - 1
        go to 60
 70     continue
        if(i .lt. j)then;
            temp = x(i)
            x(i) = x(j)
            x(j) = temp
            i = i + 1
            j = j - 1
!           Decide which half the median is in.
            if(i .le. j) GO TO 50
        endif

        if(.NOT. odd)then;
            if(j .eq. nby2 .AND. i .eq. nby2p1) GO TO 130
            if(j .lt. nby2) lo = i
            if(i .gt. nby2p1) hi = j
            if(i /= j) GO TO 100
            if(i .eq. nby2) lo = nby2
            if(j .eq. nby2p1) hi = nby2p1
        else
            if(j .lt. nby2p1) lo = i
            if(i .gt. nby2p1) hi = j
            if(i /= j) GO TO 100

!       Test whether median has been isolated.

            if(i .eq. nby2p1) return
        endif
 100    continue
        if(lo .lt. hi - 1) GO TO 10

        if(.NOT. odd)then;
            xmed = 0.5*(x(nby2) + x(nby2p1))
            return
        endif
        temp = x(lo)
        if(temp .gt. x(hi))then;
            x(lo) = x(hi)
            x(hi) = temp
        endif
        xmed = x(nby2p1)
        return

! Special case, N even, J = N/2 & I = J + 1, so the median is
! between the two halves of the series.   Find max. of the first
! half & min. of the second half, then average.

 130    continue
        xmax = x(1)
        do k = lo, j
             xmax = MAX(xmax, x(k))
        enddo
        xmin = x(n)
        do k = i, hi
            xmin = MIN(xmin, x(k))
        enddo
        xmed = 0.5*(xmin + xmax)

        return
end subroutine median

  !===================================================================================!
subroutine report_ngnbns(caller,dat,lev,flg,ng,nb,ns,cmnt)

    character(*), intent( in )     :: caller
    integer,      intent( in )     :: dat	! idat input
    integer,      intent( in )     :: lev	! ilev inupt
    integer,      intent( in )     :: flg	! ifl  input (flag position in testflag)
    integer,      intent( in )     :: ng	! ngood
    integer,      intent( in )     :: nb	! nbad
    integer,      intent( in )     :: ns	! nskip
    character(*), intent( in )     :: cmnt	! comment (8ch max)

    character(len=8) :: ch8

    ch8 = cmnt;

    write(msg,'("in ",a4,x,a8," idt,lev,ng,nb,ns: ",5i5," flgs/data: ",i3,f12.2)') &
	caller,ch8,dat,lev,ng,nb,ns,testflag(dat,lev,flg),profdata(dat,lev)
    call wrtdbg(caller,msg);

end subroutine report_ngnbns

  !===================================================================================!
  ! Mark profflag for ALL columns of a ive parameter
subroutine mark_profflag_param(caller,tno,greyflag,info)

  character(*), intent( in ) :: caller
  integer,      intent( in ) :: tno		! Test number (ex: grey=15)
  integer,      intent( in ) :: greyflag	! The flag to use
  integer,      intent( in ) :: info(3)		! info = P(1,p1r,p4a), T(2,...), S(3,...)
  integer :: itype, beg, end;
  character(len=16), parameter :: sub = 'mark_prfflg_parm'

  itype=info(1); beg=info(2); end=info(3); ! run characteristics

  if(nlev==0)then; write(msg,'("W nlev=0, no test performed")');call wrttrc(sub,msg); return; endif

  do idat=beg,end                                                                                 ! idat=beg,end
      if(hasDATA(idat))then                                                                       ! if hasDATA
	do ilev=1,nlev
	    if(greyflag > profflag(idat,ilev)) profflag(idat,ilev) = greyflag;	! Grey list - change all levs to a single flag
	enddo
	! Stash the results for this test
	qcdataresult(idat) = greyflag;
      endif
  enddo

  return
end subroutine mark_profflag_param

!============================================================================
!    land_or_water.f
!    ===============
!    This subroutine will examine the position (X) with all 8 surrounding
!    (lat,lon) positions from ETOPO5 grids and decide the float is on land
!    or water. If more positions are on water than land then the X (float)
!    is on water.
!
!    1     2     3
!    4     X     6
!    7     8     9
!
!    (where X is a position to check)
!
!    YHD, 12-Jan-2007
!
! Converted to a function Feb 2014
!
! Returns int status:
!
! -1 = FAILURE, TOPO_FILE database not found
!  0 = SUCCESS, The positin is NOT on land
!  1 = SUCCESS, The position is on land
!

integer function land_or_water(caller,alat, alon)

      character(*), intent( in ) :: caller        ! who is calling me?

      integer depth,iland,iwater,i,drt
      integer idummy
      real rlat,rlon,alat,alon,xadd,yadd,delta,rlat1,rlon1
      character(len=128) :: filename

      logical :: file_exist

      integer :: status

      character(len=13) :: sub = 'land_or_water';

	! Do we need to check for the topo file?
	if(.not. topo_file_checked)then
	    topo_file_checked = .true.;
	    if(DEBUG)then; write(msg,'("RTB TEST Check for the TOPO file")'); call wrtdbg(sub,msg); endif
	    topo_file="/a1/ARGO/qcplus/clima_mask/ETOPO5.DAT";	! set the default value
	    if(DEBUG)then; write(msg,'("RTB TEST default TOPO_FILE: ",a)')trim(topo_file); call wrtdbg(sub,msg); endif
	    call get_default_envvar(sub,'TOPO_FILE',filename);
	    if(len(trim(filename)) == 128)then
		write(wrnmsg,'("WARNING 111 land_or_water ENV VAR TOPO is not defined")'); ! WRN
		call warning(sub,111,wrnmsg); !  WARNING 111 in data_arrays function land_or_water ! WRN
		topo_file_exists = .false.;
		land_or_water = -1; ! land/water status for positions cannot be checked, set return value then return
		if(DEBUG)then; write(msg,'("RTB TEST TOPO_FILE at 2: ENV VAR not defined")'); call wrtdbg(sub,msg); endif
		return;
	    else
		topo_file = trim(filename);
		if(DEBUG)then; write(msg,'("RTB TEST retrieve TOPO_FILE ENVVAR: ",a)')trim(topo_file); call wrtdbg(sub,msg); endif
	    endif
	    if(DEBUG)then; write(msg,'("RTB TEST inquire about TOPO_FILE at 3: ",a)')trim(topo_file); call wrtdbg(sub,msg); endif
	    inquire(file=topo_file,exist=file_exist)
	    if(.not. file_exist)then;
		write(wrnmsg,'("WARNING 112 land_or_water called by ",a," TOPO file missing @1",a)')trim(caller),trim(topo_file); ! WRN
		call warning(sub,112,wrnmsg); ! WARNING 112 in data_arrays function land_or_water ! WRN
		topo_file_exists = .false.;
		land_or_water = -1; ! land/water status for positions cannot be checked, set return value then return
		if(DEBUG)then; write(msg,'("RTB TEST TOPO_FILE: ",a," DOES NOT EXIST")')trim(topo_file); call wrtdbg(sub,msg); endif
		return;
	    endif
	    topo_file_exists = .true.;
	    if(DEBUG)then; write(msg,'("RTBT TOPO topo_file at 3: ",a," exists")')trim(topo_file); call wrtdbg(sub,msg); endif
	endif
	! Did we check, but not find the file"
	if(topo_file_checked .and. .not. topo_file_exists)then
	    land_or_water = -1; ! land/water status for positions cannot be checked, set return value then return
	    if(DEBUG)then; write(msg,'("RTB TEST TOPO_FILE checked, but does not exist")'); call wrtdbg(sub,msg); endif
	    return;
	endif
    inquire(file=topo_file,exist=file_exist)
    if(.not. file_exist)then;
	write(wrnmsg,'("WARNING 113 land_or_water called by ",a," TOPO file missing @2")')trim(caller),trim(topo_file); ! WRN
	call warning(sub,113,wrnmsg); ! WARNING 113  in data_arrays function land_or_water ! WRN
	land_or_water = -1; ! land/water status for positions cannot be checked, set return value then return
	return
    endif

    if(DBLEV== 4 .and. DOGRID)then
	grxlon = grxlonlo - grinc;
	do while (grxlon < grxlonhi)
	    grxlon = grxlon + grinc;
	    grylat = grylatlo - grinc;
	    do while (grylat < grylathi)
		grylat = grylat + grinc;
		depth=frdtbase(sub,grylat,grxlon)
		write(DBG,'(2f12.5,i12," : ",a,"_lnd 777777 ",a)')grxlon, grylat,depth,trim(gridloc)  ! DBG
	    enddo
	enddo
	if(DBLEVEXIT) call dblev_exit("after doing grid in qc04")
    endif

      depth=frdtbase(sub,alat,alon)
      if(DEBUG)then; write(msg,*)'RTB return from frdtbase @1, depth: ',depth; call wrtdbg(sub,msg); endif
      if(DBLEV== 4)then; write(msg,'("LEV04 RTB @1 depth/lat/lon: ",i12,2f12.3)')depth,alat,alon; call wrtdbg(sub,msg); endif


      ! the following if statement does not make much sense anymore because                             CS comment
      ! with the changed rdtbase.c it prevents testing positions near alat,alon in the loop below       CS comment
      ! Also: 1) are we not testing the existence of TOPO elsewhere?                                    CS comment
      !       2) I deleted the second one of the two lines setting land_or_water to -1                  CS comment
      if(depth > 99990)then
	land_or_water = -1; ! mark as bad but DO NOT EXIT, jusr report the error, then return           CS comment
	write(errmsg(1),'("NON-FATAL ERROR 224 d.land_or_water called by ",a)')trim(caller);
	write(errmsg(2),'("NON-FATAL ERROR 224 land_or_water in data_arrays reports depth > 99990")');
	write(errmsg(3),'("NON-FATAL ERROR 224 probable cause: a required file is missing - No TOPO file exists ")');
	call error_report(sub,224,3,errmsg); ! land_or_water in data_arrays depth > 99990
	if(DEBUG)then; write(msg,'("The land/water status cannobe determined, bad rdtbase")'); call wrtstd(sub,msg); endif
	return
      endif
      if(depth.lt.0)then
	land_or_water = 0;	! The float is in WATER
        ! LET THE CALLER SEND THE MESSAGE write(msg,*)'The float is in WATER!!!'; call wrttrc(sub,msg)
	if(DEBUG)then; write(msg,'("This float is in the WATER (at 1)")'); call wrtstd(sub,msg); endif
        return
      endif

      delta=0.0833
      iland=0
      iwater=0
      idummy=0
      xadd=0.0
      yadd=0.0
      rlat=0.0
      rlon=0.0
      drt=0

      do i=1,9
        if(i.eq.1)then
          xadd=-1.0
          yadd=-1.0
        elseif(i.eq.2)then
          xadd=0.0
          yadd=-1.0
        elseif(i.eq.3)then
          xadd=+1.0
          yadd=-1.0
        elseif(i.eq.4)then
          xadd=-1.0
          yadd=0.0
        elseif(i.eq.5)then
          xadd=0.0
          yadd=0.0
        elseif(i.eq.6)then
          xadd=1.0
          yadd=0.0
        elseif(i.eq.7)then
          xadd=-1.0
          yadd=1.0
        elseif(i.eq.8)then
          xadd=0.0
          yadd=1.0
        elseif(i.eq.9)then
          xadd=1.0
          yadd=1.0
        endif
        rlat1=alat+(xadd*delta)
        rlon1=alon+(yadd*delta)

	drt=frdtbase(sub,rlat1,rlon1)
        if(DEBUG)then; write(msg,*)'RTB return from frdtbase @2, depth: ',drt; call wrtdbg(sub,msg); endif
	if(DBLEV== 4)then; write(msg,'("LEV04 RTB @2 depth/lat/lon: ",i12,2f12.3)')drt,alat,alon; call wrtdbg(sub,msg); endif

      if(drt > 99990)then
        land_or_water = -1; ! mark as bad but DO NOT EXIT, jusr report the error, then return           CS comment
        write(errmsg(1),'("NON-FATAL ERROR 225 d.land_or_water called by ",a)')trim(caller);
        write(errmsg(2),'("NON-FATAL ERROR 225 land_or_water in data_arrays reports drt > 99990")');
        write(errmsg(3),'("NON-FATAL ERROR 225 probable cause: a required file is missing - No TOPO file exists ")');
        call error_report(sub,225,3,errmsg); ! land_or_water in data_arrays drt > 99990
        if(DEBUG)then; write(msg,'("The land/water status cannobe determined, bad rdtbase")'); call wrtstd(sub,msg); endif
        return
      endif

        if(drt.gt.0)then
            iland=iland+1
	elseif(drt.le.-9999)then
	    idummy=idummy+1
        else
          iwater=iwater+1
        endif
      enddo
      if(idummy.gt.0)then
	write(errmsg(1),'("NON-FATAL ERROR 226 in dd.land_or_water")');
	write(errmsg(2),'("NON-FATAL ERROR 226 bad vlues returned from frtdbase (idummy>0)")')
	call error_report(sub,226,2,errmsg); ! land_or_water in data_arrays drt > 99990 bad vlues returned from frtdbase
      endif

      if(iwater.gt.iland)then
	 ! LET THE CALLER SEND THE MESSAGE write(msg,*)'The float is in WATER!!!'; call wrttrc(sub,msg)
	 land_or_water = 0;	! The float is in WATER
	 if(DEBUG)then; write(msg,'("This float is in the WATER (at 2)")'); call wrtstd(sub,msg); endif
      else
	! LET THE CALLER SEND THE MESSAGE write(msg,*)'The float is on LAND!!!'; call wrttrc(sub,msg)
	land_or_water = 1;	! The float is on LAND
	if(DEBUG)then; write(msg,'("This float is on LAND")'); call wrtstd(sub,msg); endif
      endif

      return
end function land_or_water

  !===================================================================================!
  !
  ! Note that this reports back the FLAG VALUE (0=notd,1=GOOD,4=FAIL
  !

subroutine evaluate_flags(caller,QCTEST,dataname,idat,ifl,beg,result)
    character(*), intent( in )     :: caller
    character(*), intent( in )     :: QCTEST
    character(*), intent( in )     :: dataname
    integer, intent( in ) :: idat
    integer, intent( in ) :: ifl
    integer, intent( in ) :: beg
    integer, intent(out ) :: result

    integer :: nerr, ngood, nbad, nskip,ntotal, status;
    character(len=28) :: outstr                 ! hold the first part of the output message
    real :: pcntg, pcntb,pcnts        ! percent of good,bad,skipped points
    logical, parameter :: ANALYZE_ALL_EVAL = .false. ! .true. ! list the testflags in evaluate_flags
    logical, parameter :: ANALYZE_END_EVAL = .false. ! .true. ! list the testflags in evaluate_flags

    character(len=4),parameter :: sub = 'EVAL'

    write(outstr,'(a,1x,a)') trim(QCTEST),trim(dataname);

    ntotal = proftot(idat)
    nerr = 0; ngood = 0; nbad = 0; nskip = 0;
    do ilev=1,nlev
	if(testflag(idat,ilev,ifl) == 1)then
	    ngood = ngood + 1;
	elseif(testflag(idat,ilev,ifl) >= 4)then	! includes BAD and all above
	    nbad = nbad + 1;
	elseif(testflag(idat,ilev,ifl) == 3)then	! PROB_BAD
	    nbad = nbad + 1;
	elseif(	testflag(idat,ilev,ifl) == 0)then
	    nskip = nskip + 1;
	else
	    write(msg,*)'evaluate_flags ERR at idat,ilev,ifl: ',idat,ilev,ifl,' on flag: ',testflag(idat,ilev,ifl);
	    call wrtstd(sub,msg);
	    nerr = nerr + 1;
	endif
	if(ANALYZE_ALL_EVAL)then;
	    write(msg,1001)caller,ilev,nerr,ngood,nbad,nskip,testflag(idat,ilev,ifl),profdata(idat,ilev);
	    call wrtdbg(sub,msg);
	endif
    enddo
    if(ANALYZE_ALL_EVAL)then
	write(msg,'("----  ---------------------  ---- -----  ---- ---------  ---- ---- ----")')
	call wrtdbg(sub,msg);
    endif
    if(ANALYZE_END_EVAL)then; write(msg,1002)  trim(caller),nlev,nerr,ngood,nbad,nskip; call wrtdbg(sub,msg); endif
    1001 format(a," EVAL nlev: ",i5," nerr: ",i5," ng,nb,ns: ",3i5," flag: ",i2,"data: ",f12.2)
    1002 format(a," EVAL nlev: ",i5," nerr: ",i5," ng,nb,ns: ",3i5)
    if(nerr > 0)then
	write(errmsg(1),'("FATAL ERROR 911 incomplete qctest v1 flag evaluation")');
	write(errmsg(2),'("FATAL ERROR 911 in ",a,x,a)')trim(caller),trim(QCTEST)
	write(errmsg(3),'("FATAL ERROR 911 Number of error levels: ",i5," of ",i5," total")')nerr,ntotal;
	call error_exit(sub,911,3,errmsg); ! evaluate_flags in data_arrays - incomplete qctest flag evaluationin
    endif


    if(ntotal == 0)then
        result = NOTD;							! All points (from proftot)
        write(msg,'(a27," can not be determined.")') outstr
    elseif(ngood == ntotal)then
        result = GOOD;
        write(msg,'(a27," passed.")') outstr
    elseif(nbad == 0)then
	result = GOOD;
	pcntg = (real(ngood)/ntotal)*100.0
	pcnts = (real(nskip)/ntotal)*100.0
	write(msg,'("EVAL_FLAGS",a27," passed. (",f4.1," 0ood,",f4.1,"kipped)")') outstr,pcntg,pcnts
    else
        result = BAD;
	pcntg = (real(ngood)/ntotal)*100.0
        pcntb = (real(nbad)/ntotal)*100.0
	if(nskip > 0)then
	pcnts = (real(nskip)/ntotal)*100.0
	    write(msg,'(a27," failed. ",f5.1,"0ad, ",f4.1," 0ood,",f4.1,"kipped)")') outstr,pcntb,pcntg,pcnts
	else
            write(msg,'(a27," failed. ",f5.1,"0ad, ",f4.1," 0ood")') outstr,pcntb,pcntg
	endif
    endif
    ! Stash the results for this test
    qcdataresult(idat) = result;
    call wrttrc(caller,msg);

  return
end subroutine evaluate_flags

  !===================================================================================!
  !
! This subroutine uses qcdataresult to evaluate the profile status, and store in
  ! qcprofresult, to be used later when evaluating if  profiles are good or bad
  !
subroutine evaluate_qcresult(stage)

    character(*), intent( in )     :: stage
    integer :: result
    character(len=SZC) :: tmsg;
    character(len=80) :: rmsg(MAXPRF)

    character(len=16),parameter :: sub = "evaluate_qcrsult"

    result = max(qcdataresult(p1r),qcdataresult(p1a),qcdataresult(t1r),qcdataresult(s1r),qcdataresult(s1a));
    qcprofresult(1) = max(qcprofresult(1),result);
    result = max(qcdataresult(p2r),qcdataresult(p2a),qcdataresult(t2r),qcdataresult(s2r),qcdataresult(s2a));
    qcprofresult(2) = max(qcprofresult(2),result);
    result = max(qcdataresult(p3r),qcdataresult(p3a),qcdataresult(t3r),qcdataresult(s3r),qcdataresult(s3a));
    qcprofresult(3) = max(qcprofresult(3),result);
    result = max(qcdataresult(p4r),qcdataresult(p4a),qcdataresult(t4r),qcdataresult(s4r),qcdataresult(s4a));
    qcprofresult(4) = max(qcprofresult(4),result);

    ! reset qcdataresult for the next test
    qcdataresult = -1; ! DBG TEST 0;

    return
end subroutine evaluate_qcresult



  !===================================================================================!
  !
  !
  ! This works like evaluate_flags, but works on TWO variables (usually TEMP & PSAL)
  !
  ! The FIRST entry is the column that actually CONTAINS the testflags. (For ecample,
  ! when called to evaluate qc14, density_inversin, the flags are stored with the salinity
  ! flags, but must also be applied to the associated TEMP column
  !
  ! Note that this reports back the FLAG VALUE (0=notd,1=GOOD,4=FAIL
  !

subroutine evaluate_2v_flags(caller,QCTEST,dataname,pdat,ifl,sdat,beg,end,tno,result)
    character(*), intent( in )     :: caller
    character(*), intent( in )     :: QCTEST
    character(*), intent( in )     :: dataname
    integer, intent( in ) :: pdat       ! primary data column
    integer, intent( in ) :: ifl        ! flag number in the testflag array
    integer, intent( in ) :: sdat       ! secondary column (no flag in testflags array)
    integer, intent( in ) :: beg        ! beginning data column
    integer, intent( in ) :: end        ! ending dta column
    integer, intent( in ) :: tno        ! TEST number (01-22), not the index in the testflag array
    integer, intent(out ) :: result

    integer :: nerr, ngood, nbad, nskip,ntotal, status;
    character(len=28) :: outstr                 ! hold the first part of the output message
    real :: pcntg, pcntb,pcnts        ! percent of good,bad,skipped points
    logical, parameter :: ANALYZE_ALL_EVAL = .false. ! .true. ! list the testflags in evaluate_2v_flags
    logical, parameter :: ANALYZE_END_EVAL = .false. ! .true. ! list the testflags in evaluate_2v_flags

    character(len=4),parameter :: sub = 'EVAL'

    write(outstr,'(a,1x,a)') trim(QCTEST),trim(dataname);

    ntotal = proftot(pdat)
    nerr = 0; ngood = 0; nbad = 0; nskip = 0;
    do ilev=1,nlev
        if(testflag(pdat,ilev,ifl) == 1)then
            ngood = ngood + 1;
        elseif(testflag(pdat,ilev,ifl) >= 4)then        ! includes BAD and all above
            nbad = nbad + 1;
        elseif(testflag(pdat,ilev,ifl) == 3)then        ! PROB_BAD
            nbad = nbad + 1;
        elseif( testflag(pdat,ilev,ifl) == 0)then
            nskip = nskip + 1;
        else
            write(msg,*)'evaluate_2v_flags ERR at pdat,ilev,ifl: ',pdat,ilev,ifl,' on flag: ',testflag(pdat,ilev,ifl);
            call wrtstd(sub,msg);
            nerr = nerr + 1;
        endif
        if(ANALYZE_ALL_EVAL)then;
            write(msg,1001)caller,ilev,nerr,ngood,nbad,nskip,testflag(pdat,ilev,ifl),profdata(pdat,ilev);
            call wrtdbg(sub,msg);
        endif
    enddo
    if(ANALYZE_ALL_EVAL)then
        write(msg,'("----  ---------------------  ---- -----  ---- ---------  ---- ---- ----")')
        call wrtdbg(sub,msg);
    endif
    if(ANALYZE_END_EVAL)then; write(msg,1002)  trim(caller),nlev,nerr,ngood,nbad,nskip; call wrtdbg(sub,msg); endif
    1001 format(a," EVAL nlev: ",i5," nerr: ",i5," ng,nb,ns: ",3i5," flag: ",i2,"data: ",f12.2)
    1002 format(a," EVAL nlev: ",i5," nerr: ",i5," ng,nb,ns: ",3i5)
    if(nerr > 0)then
        write(errmsg(1),'("FATAL ERROR 912 incomplete qctest v2 flag evaluation")');
        write(errmsg(2),'("FATAL ERROR 912 in ",a,x,a)')trim(caller),trim(QCTEST)
        write(errmsg(3),'("FATAL ERROR 912 Number of error levels: ",i5," of ",i5," total")')nerr,ntotal;
        call error_exit(sub,912,3,errmsg); !  evaluate_2v_flags in data_arrays - incomplete qctest flag evaluation
    endif

    if(ntotal /= (ngood+nbad+nskip))then
        write(msg,'("================================================================")')
        call wrttrc(sub,msg)
        write(msg,'("From evaluate_2v_flags - NOTIFY VICKI ntot ne ngood+nbad+nskip - nt: ",i5," ng/nb/ns: ",3i5)') &
                        ntotal,ngood,nbad,nskip
        call wrttrc(sub,msg);
    endif

    if(ntotal == 0)then
        result = NOTD;                                                  ! All points (from proftot)
        write(msg,'(a27," can not be determined.")') outstr
    elseif(ngood == ntotal)then
        result = GOOD;
        write(msg,'(a27," passed.")') outstr
    elseif(nbad == 0)then
        result = GOOD;
        pcntg = (real(ngood)/ntotal)*100.0
        pcnts = (real(nskip)/ntotal)*100.0
        write(msg,'("EVAL_2V_fl",a27," passed. (",f4.1," 0ood,",f4.1,"kipped)")') outstr,pcntg,pcnts
    else
        result = BAD;
        pcntg = (real(ngood)/ntotal)*100.0
        pcntb = (real(nbad)/ntotal)*100.0
        if(nskip > 0)then
        pcnts = (real(nskip)/ntotal)*100.0
            write(msg,'(a27," failed. ",f5.1,"0ad, ",f4.1," 0ood,",f4.1,"kipped)")') outstr,pcntb,pcntg,pcnts
        else
            write(msg,'(a27," failed. ",f5.1,"0ad, ",f4.1," 0ood")') outstr,pcntb,pcntg
        endif
    endif
    ! Stash the results for this test
    qcdataresult(pdat) = result;
    qcdataresult(sdat) = result;
    call wrttrc(caller,msg);    
    result = qcdataresult(beg);                 ! Report back result for PRIMAY profile only

end subroutine evaluate_2v_flags

  !===================================================================================!
subroutine mark_testflag_levall(caller,info,flag,ifl)

    character(*), intent( in ) :: caller
    integer,      intent( in ) :: ifl		! which test to flag
    integer,      intent( in ) :: flag		! flag to use (0,1,4)
    integer,      intent( in ) :: info(3)		! which variable to flag
    integer :: itype, beg, end;
    character(len=4), parameter :: sub = 'MTFA'	! shorthand for Mark TestFlag levAll

    itype=info(1); beg=info(2); end=info(3); ! run characteristics

    if(nlev==0)then; write(msg,'("W nlev=0, nothing to do")');call wrttrc(sub,msg); return; endif

    do idat=beg,end											! idat=beg,end
        if(hasDATA(idat))then									! if hasDATA
	    do ilev=1,nlev
		testflag(idat,ilev,ifl) = flag;
	    enddo ! ilev
        endif	! if hasDATA
    enddo		! idat

    return
end subroutine mark_testflag_levall
!==========================================================================
subroutine evaluate_job_flags()

    integer :: iret=NOTD
    integer :: dmax, omax
    integer :: lmax
    integer :: idat, ilev, iprf
    integer :: pritmp(6);

    character(len=12), parameter   :: sub = 'evl_job_flgs';

    if(DBGIO) call dbg_enter(sub);

    maxoverflag = 0; do idat = 1,MAXDAT; maxdataflag(idat) = 0; enddo;
    do idat =1,MAXDAT
        dmax = -1;
        if(hasDATA(idat))then
            dmax = 0;
            do ilev = 1,nlev
                if(profflag(idat,ilev) < 5)then
                    if(profflag(idat,ilev) > dmax) dmax = profflag(idat,ilev)
                endif
            enddo
        endif
        maxdataflag(idat) = dmax;
    enddo
    ! Next find the max flag for each PROFILE
    ! This next part is UGLY!!!
    pritmp(1)=maxdataflag(p1r); pritmp(2)=maxdataflag(p1a); pritmp(3)=maxdataflag(t1r);
    pritmp(4)=maxdataflag(s1r); pritmp(5)=maxdataflag(s1a); pritmp(6)=maxdataflag(do1);
    maxpriprofflag = 0; do i=1,6; if(pritmp(i) > maxpriprofflag) maxpriprofflag = pritmp(i); enddo
    maxoverflag = 0; do i=1,MAXDAT; if(maxdataflag(i) > maxoverflag) maxoverflag = maxdataflag(i); enddo

end subroutine evaluate_job_flags
!==========================================================================
subroutine evaluate_final_flags(comment)

    character(*), intent( in )     :: comment

    integer :: iret=NOTD
    integer :: max
    integer :: idat,ipar,iprf
    integer :: np,nt,ns
    integer :: iflag
    integer :: p1d, p2d,p3d, p4d ! pres idat col to grade (depends on adjustment)
    integer :: s1d, s2d,s3d, s4d ! psal idat col to grade (depends on adjustment)
    character(len=1) :: grade, evluate_grade

    character(len=12), parameter   :: sub = 'evl_fin_flgs';

    if(DBGIO) call dbg_enter(sub);

    do iprf =1,MAXPRF
	do ipar=1,MAXPAR
	    prof_qc(ipar,iprf) = 'x'; ! Initialize with M for missing
	enddo
    enddo

    do iprf=1,MAXPRF
	prof_pres_qc(iprf) = 'M';
	prof_temp_qc(iprf) = 'M';
	prof_psal_qc(iprf) = 'M';
	prof_doxy_qc(iprf) = 'M';
	prof_doxy2_qc(iprf) = 'M';
	prof_doxy3_qc(iprf) = 'M';
    enddo

    ! evaluate_grade(jdat,jpar,jprf)
    if(hasPRES(0))then;
	if(do_pres_adj)then;
	    p1d = p1a; p2d = p2a; p3d = p3a; p4d = p4a; ! adj values
	else
	    p1d = p1r; p2d = p2r; p3d = p3r; p4d = p4r; ! reg values
	endif
	if(hasPRES(1)) prof_pres_qc(1) = evaluate_grade(p1d,1,1);
	if(hasPRES(2)) prof_pres_qc(2) = evaluate_grade(p2d,1,2);
	if(hasPRES(3)) prof_pres_qc(3) = evaluate_grade(p3d,1,3);
	if(hasPRES(4)) prof_pres_qc(4) = evaluate_grade(p4d,1,4);
    endif
    if(hasTEMP(0))then;
        if(hasTEMP(1)) prof_temp_qc(1) = evaluate_grade(t1r,2,1);
        if(hasTEMP(2)) prof_temp_qc(2) = evaluate_grade(t2r,2,2);
        if(hasTEMP(3)) prof_temp_qc(3) = evaluate_grade(t3r,2,3);
        if(hasTEMP(4)) prof_temp_qc(4) = evaluate_grade(t4r,2,4);
    endif
    if(DEBUG) write(DBG,'("PPQC_ZM dd. temp grade: ",4a2)')prof_temp_qc;
    if(hasPSAL(0))then;
        if(do_psal_adj)then;
            s1d = s1a; s2d = s2a; s3d = s3a; s4d = s4a; ! adj values
        else
            s1d = s1r; s2d = s2r; s3d = s3r; s4d = s4r; ! reg values
        endif
	if(hasPSAL(1)) prof_psal_qc(1) = evaluate_grade(s1d,3,1);
	if(hasPSAL(2)) prof_psal_qc(2) = evaluate_grade(s2d,3,2);
	if(hasPSAL(3)) prof_psal_qc(3) = evaluate_grade(s3d,3,3);
	if(hasPSAL(4)) prof_psal_qc(4) = evaluate_grade(s4d,3,4);
    endif
    if(hasDOXY(0))then;
	if(hasDOXY(1)) prof_doxy_qc(1) = evaluate_grade(do1,4,1);
	if(hasDOXY(2))then;
	    prof_doxy_qc(2) = evaluate_grade(do2,4,2); ! doxy2
	elseif(hasDOXY(3))then;
	    prof_doxy_qc(2) = evaluate_grade(doa,4,2); ! doxy2a
	endif
	if(hasDOXY(4)) prof_doxy2_qc(2) = evaluate_grade(dob,4,2); ! doxy2b
	if(hasDOXY(5)) prof_doxy3_qc(2) = evaluate_grade(doc,4,2); ! doxy2c
    endif

    if(DEBUG) write(DBG,'("MAXFLAG maxprfflag: ",4i2,"  After prof_qc evaluation")')maxprfflag; 

    ! Now do the overall profile evaluation for every column with data
    do idat=1,MAXDAT
	ipar = DATPAR(idat); iprf = DATPRF(idat);
	if(hasDATA(idat))then;
	    call evaluate_profile(idat,NFLAG(idat));
	    if(DEBUG)then;
		write(msg,'(a," call evaluate_profile for idat,nflag: ",2i3)')sub,idat,NFLAG(idat);
		call wrtdbg(sub,msg);
	    endif
	else
	    if(STS_prof(iprf))then;
		call evaluate_profile(idat,NFLAG(idat));
	    endif
	endif
    enddo

    ! Don't forget the DOXYs:

    if(hasDOXY(0))then;
        if(hasDOXY(1))then; call evaluate_profile(do1,NFLAGDOXY); call evaluate_prof_qc(do1,4,1); endif
        if(hasDOXY(2))then; call evaluate_profile(do2,NFLAGDOXY); call evaluate_prof_qc(do2,4,1); endif
        if(hasDOXY(3))then; call evaluate_profile(doa,NFLAGDOXY); call evaluate_prof_qc(doa,4,1); endif
        if(hasDOXY(4))then; call evaluate_profile(dob,NFLAGDOXY); call evaluate_prof_qc(dob,4,1); endif
        if(hasDOXY(5))then; call evaluate_profile(doc,NFLAGDOXY); call evaluate_prof_qc(doc,4,1); endif
    endif

    ! Finally, fin the the overall flags for each data column
    do idat=1,MAXDAT
	if(hasDATA(idat))then
	    max = 0;
	    do ilev=1,nlev
		if(profflag(idat,ilev) < 5)then; if(profflag(idat,ilev) > max) max = profflag(idat,ilev); endif
	    enddo
	    if(max == 0)then
		! rerun, but do not restrict max to 0-4
		do ilev=1,nlev
		    if(profflag(idat,ilev) > max) max = profflag(idat,ilev);
		enddo
	    endif
	    overflag(idat) = max
	endif
    enddo


    if(DBGIO) call dbg_leave(sub,iret);

    return

end subroutine evaluate_final_flags

!=====================================================================
subroutine evaluate_profile(idat,nflg)
    ! get overall flag for a test
    integer, intent( in )     :: idat		! index of data column
    integer, intent( in )     :: nflg		! number of flags for tis parameter

    integer :: max

    do iflg=1,nflg
        max = 0
        do ilev=1,nlev
            if(testflag(idat,ilev,iflg) > max) max = testflag(idat,ilev,iflg);
        enddo
        max_flag(idat,iflg) = max
	! add part for profile_qc(id) ch*1 determination
    enddo
    if(do_regional_test)then;
	max = 0
	do ilev=1,nlev
	    if(testflag(idat,ilev,regional_ncol) >  max) max = testflag(idat,ilev,regional_ncol)
	enddo
	max_flag(idat,regional_ncol) = max
    endif

    return
end subroutine evaluate_profile

!=====================================================================
subroutine evaluate_prof_qc(idat,ipar,iprf)

    integer, intent( in )     :: idat		! index of data column
    integer, intent( in )     :: ipar		! index of profile_<param>_qc (by PRES/TEMP/PSAL/DOXY)
    integer, intent( in )     :: iprf		! index of n_prof

    integer :: ng,np				! number of points, ngood,
    integer :: lmax                             ! local max flag value
    real :: pcnt
    character(len=16) :: sub = 'evaluate_prof_qc'

    lmax = maxprfflag(iprf); ! Store what we have so far for this profile
    np = 0; ng = 0;
    do ilev=1,nlev
	if(profflag(idat,ilev) == 1)then; 		! good
	    np = np + 1; ng = ng + 1; if(lmax < 1) lmax = 1; 
	elseif(profflag(idat,ilev) == 2)then;		! probably good
	    np = np + 1; ng = ng + 1; if(lmax < 2) lmax = 2;
	elseif(profflag(idat,ilev) == 5)then;		! value changed
	    np = np + 1; ng = ng + 1;
	elseif(profflag(idat,ilev) == 8)then;		! interpolated data
	    np = np + 1; ng = ng + 1;
	elseif(profflag(idat,ilev) == 9)then;		! Omit missing data levels
	   if(DBLEV>999) write(msg,'("Dummy write statement")');
	elseif(profflag(idat,ilev) == 3)then;
	    np = np + 1; if(lmax < 3) lmax = 3;
	elseif(profflag(idat,ilev) == 4)then;
	    np = np + 1; if(lmax < 4) lmax = 4;
	else
	    np = np + 1
	endif
    enddo
    maxprfflag(iprf) = lmax; ! Store the new max value for this profile
    if(np == 0)then
	write(msg,'("W after filtering out missing levels, NP=ZERO")'); call wrttrc(sub,msg)
	write(msg,'("W return a space as the grade")');			call wrttrc(sub,msg)
	prof_qc(ipar,iprf) = " ";
	return
    endif
    pcnt = (real(ng)/real(np))*100.0			! percent good
        if(ng == np)     then; prof_qc(ipar,iprf) = "A";		! all good points
    elseif(ng ==  0)     then; prof_qc(ipar,iprf) = "F";		!  no good points
    elseif(pcnt >= 75.0)then; prof_qc(ipar,iprf) = "B";		! >75 to 100 percent (except all good)
    elseif(pcnt >= 50.0)then; prof_qc(ipar,iprf) = "C";		! >50 to  75 percent
    elseif(pcnt >= 25.0)then; prof_qc(ipar,iprf) = "D";		! >25 to  50 percent
    else;                      prof_qc(ipar,iprf) = "E";		! > 0 to 25 percent
    endif

    return
end subroutine evaluate_prof_qc

  !===================================================================================!

subroutine reverse_prevqcdata(caller) ! This will reverse the data values into the arrays
    character(*), intent( in )     :: caller
    character(len=16), parameter   :: sub = 'reverse_prevqcdata';

   if(DBLEV==18)then;
	write(msg,*)" prevpresdata(1) in test :: ",prevpresdata(1),prevpresdata(nprevlev)
	call wrtdbg(sub,msg);
    endif
   if(prevpresdata(1) > prevpresdata(nprevlev))then
       do ilev=1,nprevlev
            pdum(ilev) = prevpresdata(ilev)
            tdum(ilev) = prevtempdata(ilev)
            sdum(ilev) = prevpsaldata(ilev)
            pdumflag(ilev)= prevpresflag(ilev)
            tdumflag(ilev)=prevtempflag(ilev)
            sdumflag(ilev)=prevpsalflag(ilev)
       enddo
       i = nprevlev +1
       do ilev = 1,nprevlev
           i = i - 1;
           prevpresdata(ilev)=pdum(i);
           prevtempdata(ilev)=tdum(i);
           prevpsaldata(ilev)=sdum(i);
           prevpresflag(ilev)=pdumflag(ilev);
           prevtempflag(ilev)=tdumflag(ilev);
           prevpsalflag(ilev)=sdumflag(ilev);
       enddo
   endif

end subroutine reverse_prevqcdata

  !===================================================================================!

subroutine mark_profdata_bad(caller,idat) ! This will mark the data values into the arrays
    character(*), intent( in )     :: caller
    integer, intent( in ) :: idat
    character(len=16), parameter   :: sub = 'mark_profdata_baa';

   if(DBLEV==18)then; write(msg,*)" IN mark_profdata_bad"; call wrtlog(sub,msg); endif
       do ilev=proftop(idat),profbot(idat)
           profflag(idat,ilev)=BAD;
       enddo
end subroutine mark_profdata_bad

  !===================================================================================!

subroutine time_elapsed_in_days(caller) ! This will mark the data values into the arrays
    character(*), intent( in )     :: caller
    real, parameter :: daysecs = 24.*60.*60. ! seconds in a day
    real :: time1,time2
    double precision t1,t2,dt
    integer ::ihr;
    character(len=16), parameter   :: sub = 'tim_elpsd_in_dys';

    ! Calculate t1 -- Time of previous profile or launch
    ihr=prev_hh2(ppos2_good)*100+prev_mm2(ppos2_good);
    call julday(prev_year2(ppos2_good),prev_mon2(ppos2_good),prev_day2(ppos2_good),ihr,time1)!launch or previous tim
    t1=time1+float(prev_sc2(ppos2_good))/daysecs

    !! Time of position i
    if(DBLEV== 5)then;
	write(msg,*)"LEV05      prev_hh/ss: ",prev_hh(ppos2_good),   ", ",prev_mm(ppos2_good);   call wrtdbg(sub,msg);
	write(msg,*)"LEV05    prev_hh2/mm2: ",prev_hh2(ppos2_good),  ", ",prev_mm2(ppos2_good);  call wrtdbg(sub,msg);
	write(msg,*)"LEV05 prev_year2/mon2: ",prev_year2(ppos2_good),", ",prev_mon2(ppos2_good); call wrtdbg(sub,msg);
    endif
    ihr = this_hh2(1)*100+this_mm2(1)
    
    call julday(this_year2(1),this_mon2(1),this_day2(1),ihr,time2)
    t2=time2+float(this_sc2(1))/daysecs

    !! Time elapsed betw previous profile or launch and pos i
    dt=(t2-t1)
    if(dt .le. DT_MAX_DAYS)then;
	TIME_DIFF_SMALL=.true.;
    else;
    endif
    if(DBLEV== 5)then;
	write(msg,*)" Time Diff dt,dt = ",dt; call wrtdbg(sub,msg);
	write(msg,*)" Time Diff t1,t1 = ",t1; call wrtdbg(sub,msg);
	write(msg,*)" Time Diff t2,t2 = ",t2; call wrtdbg(sub,msg);
	write(msg,*)"     DT_MAX_DAYS = ",DT_MAX_DAYS; call wrtdbg(sub,msg);
	write(msg,*)" TIME_DIFF_SMALL = ",TIME_DIFF_SMALL; call wrtdbg(sub,msg);
    endif;


end subroutine time_elapsed_in_days

!==========================================================================
! Temporarily ported over from data_arrays...
!
! To test this use the deepest test, profile 4881_159663_241 - there is
! a transition from 1 to for in deepest pres flag: lev 285=1, lev 206=4
! The debugs are removed, but a typical debug line is:
! write(msg,'("lev: ",i4," if: ",i2," max: ",i2," flgs: ",8i2)')ilev,if,maxflag, &
! 	(testflag(idat,ilev,i),i=1,NFLAG(idat))

!

subroutine evaluate_level_flags(comment,info)

    character(*), intent( in )     :: comment
    integer,      intent( in ) :: info(3)
    integer :: if, itype, beg, end;
    ! MOVED UP INTO MOD DECLARATOINS integer :: maxflag
    integer :: sgr				! the idat locaton of salniyt wrt temperature

    character(len=20), parameter   :: sub = 'evaluate_level_flags'

    itype=info(1); beg=info(2); end=info(3); ! run characteristics

    if(nlev==0)then; write(msg,'("W nlev=0, no test performed")');call wrttrc(sub,msg); return; endif

    if(DEBUG)then; write(msg,1000)sub,parmname(itype),itype,beg,end,trim(comment); call wrtdbg(sub,msg); endif
    1000 format(a," for: ",a," itype,beg,end: ",3i5," comment: ",a)

    do idat=beg,end
	if(hasDATA(idat))then
	    if(itype==2) sgr = tbud(idat-beg+1);
	    do ilev=1,nlev
		maxflag = profflag(idat,ilev);
		do if=1,NFLAG(idat)
		    if(testflag(idat,ilev,if) > maxflag) maxflag = testflag(idat,ilev,if);
		enddo
		if(itype==2)then;
		    if(testflag(sgr,ilev,dengrad_ncol) > maxflag) maxflag = testflag(sgr,ilev,dengrad_ncol);
		endif
		if( (sea=="MED" .or. sea=="RED") .and. (itype==2 .or. itype==3) )then
		    ! itype 2 -> TEMP, itype 3 -> PSAL
		    if(testflag(idat,ilev,regional_ncol) > maxflag) maxflag = testflag(idat,ilev,regional_ncol);
		endif
		profflag(idat,ilev) = maxflag;
	    enddo;
	endif
    enddo

    if(DEBUG)then; write(msg,'(" ====== Leave evaluate_level_flags ======")'); call wrtdbg(sub,msg); endif

end subroutine evaluate_level_flags

subroutine evaluate_pres_flags(comment)

    character(*), intent( in )     :: comment

    character(len=16), parameter   :: sub = 'evaluate_pres_flags';

    if(DEBUG)then; write(msg,1000)sub,"PRES"; call wrtdbg(sub,msg); endif
    1000 format(a,"evaluate_pre_flags calls evaluate_level_flags for ",a)

    call evaluate_level_flags(comment,P);

end subroutine evaluate_pres_flags

subroutine evaluate_data_flags(itest,comment)

    integer,      intent( in )     :: itest
    character(*), intent( in )     :: comment

    character(len=16), parameter   :: sub = 'evaluate_data_flags';

    if(hasDATA(t1r))then;
	if(DEBUG)then; write(msg,1000)sub,"TEMP"; call wrtdbg(sub,msg); endif
	call evaluate_level_flags(comment,T)
    endif
    if(hasDATA(s1r))then;
	if(DEBUG)then; write(msg,1000)sub,"PSAL"; call wrtdbg(sub,msg); endif
	call evaluate_level_flags(comment,S)
    endif
    if(hasDOXY(0))then;
	if(DEBUG)then; write(msg,1000)sub,"DOXY"; call wrtdbg(sub,msg); endif
	call evaluate_level_flags(comment,O)
    endif
    1000 format(a," Call evaluate_level_flags for ",a)

    if(itest >= 14)then;
	! Special case - use utterly brute force for the density_gradient flag - transfer s test flag to t
	do ilev=1,nlev
	    if(testflag(s1r,ilev,dengrad_ncol) > profflag(t1r,ilev)) profflag(t1r,ilev) = testflag(s1r,ilev,dengrad_ncol)
	    if(testflag(s2r,ilev,dengrad_ncol) > profflag(t2r,ilev)) profflag(t2r,ilev) = testflag(s2r,ilev,dengrad_ncol)
	    if(testflag(s3r,ilev,dengrad_ncol) > profflag(t3r,ilev)) profflag(t3r,ilev) = testflag(s3r,ilev,dengrad_ncol)
	    if(testflag(s4r,ilev,dengrad_ncol) > profflag(t4r,ilev)) profflag(t4r,ilev) = testflag(s4r,ilev,dengrad_ncol)
	enddo
    endif


end subroutine evaluate_data_flags

!==========================================================================
subroutine evaluate_qc19_flags(caller)
! If deepest pres is bad anywhere, mark TEMP & PSAL also as bad (flag=4)

    character(*), intent( in )     :: caller

    ! A bit of brute force here, I know, I know

    do ilev=1,nlev;
	if(testflag(p1r,ilev,deep_pres_ncol)==4)then;
            if(profflag(t1r,ilev) < 4) profflag(t1r,ilev)=4;
            if(profflag(s1r,ilev) < 4) profflag(s1r,ilev)=4;
        endif
	if(testflag(p1a,ilev,deep_pres_ncol)==4)then;
            if(profflag(s1a,ilev) < 4) profflag(s1a,ilev)=4;
        endif
    enddo
    if(hasPRES(2))then
	do ilev=1,nlev;
	    if(testflag(p2r,ilev,deep_pres_ncol)==4)then; profflag(t2r,ilev)=4; profflag(s2r,ilev)=4; endif
	    if(testflag(p2a,ilev,deep_pres_ncol)==4)then;                       profflag(s2a,ilev)=4; endif
	enddo
    endif
    if(hasPRES(2))then
        do ilev=1,nlev;
            if(testflag(p2r,ilev,deep_pres_ncol)==4)then;
                if(profflag(t2r,ilev) < 4) profflag(t2r,ilev)=4;
                if(profflag(s2r,ilev) < 4) profflag(s2r,ilev)=4;
            endif
            if(testflag(p2a,ilev,deep_pres_ncol)==4)then;
                if(profflag(s2a,ilev) < 4) profflag(s2a,ilev)=4; 
            endif
        enddo
    endif

    if(hasPRES(3))then
        do ilev=1,nlev;
            if(testflag(p3r,ilev,deep_pres_ncol)==4)then;
                if(profflag(t3r,ilev) < 4) profflag(t3r,ilev)=4;
                if(profflag(s3r,ilev) < 4) profflag(s3r,ilev)=4;
            endif
            if(testflag(p3a,ilev,deep_pres_ncol)==4)then;
                if(profflag(s3a,ilev) < 4) profflag(s3a,ilev)=4;
            endif
        enddo
    endif

    if(hasPRES(4))then
        do ilev=1,nlev;
            if(testflag(p4r,ilev,deep_pres_ncol)==4)then;
                if(profflag(t4r,ilev) < 4) profflag(t4r,ilev)=4;
                if(profflag(s4r,ilev) < 4) profflag(s4r,ilev)=4;
            endif
            if(testflag(p4a,ilev,deep_pres_ncol)==4)then;
                if(profflag(s4a,ilev) < 4) profflag(s4a,ilev)=4;
            endif
        enddo
    endif

end subroutine evaluate_qc19_flags
!==========================================================================
subroutine evaluate_qc22_flags(caller)

    character(*), intent( in )     :: caller

    integer :: nsp			! local shorthand for near surface profile
    integer :: nsf			! local shorthand for near surface flag - level

    nsp = near_surface_profile
    do ilev=1,nlev
	if(nsflag(ilev) > 0)then
	    nsf = nsflag(ilev);
	    if(nsf > profflag(nspdat,ilev)) profflag(nspdat,ilev) = nsf;
	    if(do_pres_adj)then; if(nsf > profflag(nspdat+1,ilev)) profflag(nspdat+1,ilev) = nsf; endif
	    if(hasTEMP(nsp))then; if(nsf > profflag(nstdat,ilev)) profflag(nstdat,ilev) = nsf; endif
	    if(hasPSAL(nsp))then
		if(nsf > profflag(nssdat,ilev)) profflag(nssdat,ilev) = nsf;
		if(do_psal_adj)then; if(nsf > profflag(nssdat+1,ilev)) profflag(nssdat+1,ilev) = nsf; endif
	    endif
	endif
    enddo

    return
end subroutine evaluate_qc22_flags

!==========================================================================

subroutine evaluate_qctest(iret,itest)

    integer, intent( in ) :: iret               ! what is the return value
    integer, intent( in ) :: itest              ! which test do I evaluate

    character(len=SZC) :: tmp

    character(len=15), parameter :: sub = 'evaluate_qctest'

end subroutine evaluate_qctest

!==========================================================================

subroutine get_platform_type(float_mfrer,platform_type)

    character(*), intent(in out) :: platform_type
    character(*), intent(in) :: float_mfrer

    character(len=14), parameter ::  sub = 'get_platform_type'

    platform_type = 'UNKNOWN'; ! In case it is not found

    if( (inst_type(1:5) == 'APEX_' .or. inst_type(1:7) == 'APEXIR_' )&
  .and. (float_mfrer(1:4) == 'WEBB' .or. float_mfrer(1:8) == 'TELEDYNE') )then
	platform_type = 'APEX';
    elseif(inst_type(1:8) == 'APEXDIR_' .and. float_mfrer(1:8) == 'TELEDYNE')then
	platform_type = 'APEX_D';
    elseif(inst_type(1:6) == 'PROVOR' .and. float_mfrer(1:8) == 'METOCEAN' )then
	platform_type = 'PROVOR_MT';
    elseif( (inst_type(1:5) == 'SOLO_' .or. inst_type(1:7) == 'SOLOIR_') .and. float_mfrer(1:3) == 'SIO' )then
	platform_type = 'SOLO';
    elseif( (inst_type(1:5) == 'SOLO_' .or. inst_type(1:7) == 'SOLOIR_' .or. inst_type(1:7) == 'SOLOOF_') &
  .and. float_mfrer(1:4) == 'WHOI' )then
	platform_type = 'SOLO_W';
    elseif( inst_type(1:8) == 'SOLO2IR_' .and. float_mfrer(1:3) == 'SIO')then
	platform_type = 'SOLO_II';
    elseif( inst_type(1:8) == 'SOLO2IR_' .and. float_mfrer(1:3) == 'MRV')then
	platform_type = 'S2A';
    elseif( inst_type(1:8) == 'SOLODIR_' .and. float_mfrer(1:3) == 'SIO')then
	platform_type = 'SOLO_D';
    elseif( inst_type(1:8) == 'SOLODIR_' .and. float_mfrer(1:3) == 'MRV')then
	platform_type = 'SOLO_D_MRV';
    elseif( inst_type(1:8) == 'NAVISIR_' .and.  float_mfrer(1:3) == 'SEA')then
	platform_type = 'NAVIS_A';
    elseif( inst_type(1:6) == 'PALACE' .and.  float_mfrer(1:4) == 'WEBB')then
	platform_type = 'PALACE';
    elseif(inst_type(1:7) == 'ALAMOIR' .and. float_mfrer(1:3) == 'MRV')then
	platform_type = 'ALAMO';
    elseif(inst_type(1:6) == 'ALTOIR' .and. float_mfrer(1:3) == 'MRV')then
	platform_type = 'ALTO';
    endif

    if(DEBUG .and. DBLEV>90)then							
	write(msg,*)"    inst_type: ", trim(inst_type);    call wrtdbg(sub,msg);	! LEV.90
	write(msg,*)"  float_mfrer: ",trim(float_mfrer);   call wrtdbg(sub,msg);	! LEV.90
	write(msg,*)"platform_type: ",trim(platform_type); call wrtdbg(sub,msg);	! LEV.90
    endif

end subroutine get_platform_type

  !===================================================================================!

subroutine fill_profvect_tmp_3param(proftot_tmp,profdata_ptmp,profdata_ttmp,profdata_stmp,pdat,tdat,sdat,ilev_vector_tmp)
    integer :: i;
    integer, intent(out ) :: proftot_tmp
    integer, intent(in) :: pdat,tdat,sdat;
    real, intent(out) ::  profdata_ptmp(MAXLEV),profdata_ttmp(MAXLEV),profdata_stmp(MAXLEV);
    integer, intent(out)::  ilev_vector_tmp(MAXLEV);
    character(len=16) :: sub = 'fill_prfvct_3prm';
    proftot_tmp=0
    do i=1,nlev
	if(skip_test(profflag(pdat,i))) go to 10;        ! Skip level
	if(skip_test(profflag(tdat,i))) go to 10;	! Skip level
	if(skip_test(profflag(sdat,i))) go to 10;	! Skip level
	! pres/temp/psal not bad, add to temporary array
        proftot_tmp = proftot_tmp+1;
        profdata_ptmp(proftot_tmp)= profdata(pdat,i);
        profdata_ttmp(proftot_tmp)= profdata(tdat,i);
        profdata_stmp(proftot_tmp)= profdata(sdat,i);
        ilev_vector_tmp(proftot_tmp)=i
	10 continue;					! Skip level
    enddo

end subroutine fill_profvect_tmp_3param

  !===================================================================================!

subroutine fill_profvect_tmp_2param(proftot_tmp,profdata_tmp1,profdata_tmp2,dat1,pdat,ilev_vector_tmp)
    integer :: i,found_pres;
    integer, intent(out ) :: proftot_tmp
    integer, intent(in) :: dat1,pdat
    real, intent(out) ::  profdata_tmp1(MAXLEV),profdata_tmp2(MAXLEV)
    integer, intent(out)::  ilev_vector_tmp(MAXLEV);
    character(len=16) :: sub = 'fill_prfvct_2prm';
    proftot_tmp=0
    do i=1,nlev
!CS        if(profflag(dat1,i) /= BAD .AND. profflag(dat1,i) /= MISSING .and. profflag(pdat,i) /= BAD .AND.&
!CS            profflag(pdat,i) /= MISSING )then
        found_pres=0;
        if(profflag(dat1,i) /= BAD .AND. profflag(dat1,i) /= MISSING)then
            proftot_tmp = proftot_tmp+1;
            profdata_tmp1(proftot_tmp)= profdata(dat1,i);
            if(profflag(pdat,i) /= BAD .AND. profflag(pdat,i) /= MISSING )then ! CS - found good pressure at i
               profdata_tmp2(proftot_tmp)= profdata(pdat,i);found_pres=1;
            endif
            if(found_pres.eq.0 .and. i>1)then ! CS - pressure at i is bad -> look for good pressure i-1
               if(profflag(pdat,i-1) /= BAD .AND. profflag(pdat,i-1) /= MISSING )then
                  profdata_tmp2(proftot_tmp)=profdata(pdat,i-1);found_pres=1;
               endif
            endif
            if(found_pres.eq.0 .and. i<nlev)then ! CS - pressure at i, i-1 are bad -> look for good pressure at i+1
               if(profflag(pdat,i+1) /= BAD .AND. profflag(pdat,i+1) /= MISSING )then
                  profdata_tmp2(proftot_tmp)=profdata(pdat,i+1);found_pres=1;
               endif
            endif
            if(found_pres.eq.0)then ! CS - pressure at i+1, i, i-1 are bad -> too bad, take pressure at i
               profdata_tmp2(proftot_tmp)= profdata(pdat,i);found_pres=1;
            endif
            ilev_vector_tmp(proftot_tmp)=i
        endif
    enddo
end subroutine fill_profvect_tmp_2param

!==========================================================================
subroutine adjust_pres(caller)

    character(*), intent( in ) :: caller

    integer :: ireg, iadj;
    character(len=SZC) :: adjmsg; real :: diffs;
    real :: ptest2 = 5.0;
    logical :: good_surf_pres, good_pqc_mrvsp;

    character(len=11), parameter :: sub = 'adjust_pres'
    character(len=20), parameter :: mbg = 'DBGPRESADJ adj_pres '

    ! First, we need to evaluate the surf_pres/most_recent... before applying the offset

    ! See if the surf pres passes the +/- 20db test
 
    if(found_surf_pres)then
	if(abs(surf_pres) > 20.0)then
	    good_surf_pres = .false.; ! FAILURE Test 1 proves it's bad
	    surf_pres_comment = "SURFACE PRESSURE IN PHY FILE - TEST 1 FAILURE +/-20 DB";
	else
	    good_surf_pres = .true.; ! SUCCESS far Test 1, so far it is good
	    surf_pres_comment = "SURFACE PRESSURE IN PHY FILE - TEST 1 SUCCESS +/-20 DB";
	endif
    else
	good_surf_pres = .false.; surf_pres_comment = "SURFACE PRESSURE NOT AVAILABLE";
    endif
 
    ! See if we have a good value for MOST RECENT VALID ...

    if(found_pqc_mrvsp)then
	if(abs(pqc_mrvsp) <= 20.0)then
	    good_pqc_mrvsp = .true.; ! pqc_mrvsp passes first pres validity test
	    pqc_mrvsp_comment = "MRVSP GOOD SUCCESS +/0 20DB TEST";
	else
	    good_pqc_mrvsp = .false.;
	    pqc_mrvsp_comment = "MRVSP BAD FAILURE +/0 20DB TEST";
	endif
    else
	good_pqc_mrvsp = .false.;
	pqc_mrvsp_comment = "MRVSP NOT AVAILABLE";
    endif

    if(good_surf_pres .and. good_pqc_mrvsp)then
	if(float_name == '2987_85111')then
	    ! The surf press bounces around a lot - this interferes with the MRSVP that is used, so increate
	    ! the difference tesr test from 5 to 15
	    ptest2 = 15.0;
	    write(wrnmsg,'("WARNING 114 float ",a," use a test threshold of 15.0,not 5.0. surf_pres:",f9.2," pqc_mrvsp:",f9.2)')& ! WRN
            trim(float_name), surf_pres, pqc_mrvsp;
	    call warning(sub,114,wrnmsg); ! WARNING 114 in data_arrays subroutine adjust_pres ! WRN
	    write(wrnmsg,1004)trim(float_name), surf_pres, pqc_mrvsp; ! WRN
	    1004 format("WARNING 114 float ",a," use a test threshold of 15.0,not 5.0. surf_pres:",f9.2," pqc_mrvsp:",f9.2)
	    call warning(sub,114,wrnmsg); ! WARNING 114 in data_arrays subroutine adjust_pres ! WRN
	endif
	if(abs(surf_pres-pqc_mrvsp) > ptest2)then
	    use_mrvsp_for_offset = .true.;
	else
	    use_spres_for_offset = .true.;
	    use_spres_for_mrvsp  = .true.;
	endif
    elseif(good_surf_pres)then
	use_spres_for_offset = .true.;
	use_spres_for_mrvsp  = .true.;
    elseif(good_pqc_mrvsp)then
	use_mrvsp_for_offset = .true.;
    else
	do_pres_adj = .false.; pres_adj_comment = "PRES ADJUSTMENT NOT AVAILABLE - NO/BAD SURF PRES & MRVSP";
	out_mrvsp = pqc_mrvsp; out_mrcyc = pqc_mrcyc; out_mrvsp_comment = "MRVSP BAD OR NOT AVAILABLE";
	return;
    endif

    if(use_spres_for_mrvsp)then
	out_mrvsp = surf_pres; out_mrcyc = thiscyc;   out_mrvsp_comment = "MRVSP FROM CURRENT SURFACE PRES";
    else
	out_mrvsp = pqc_mrvsp; out_mrcyc = pqc_mrcyc; out_mrvsp_comment = "MRVSP RETAINED";
    endif

    if(DEBUG)then;
	write(DBG,'(" ")');                                                                                                 ! DBG
	write(DBG,'("ADJUST PRES surf_pres: ",f9.2," spcyc: ",i6," cmnt: ",a)')surf_pres,thiscyc,  trim(pqc_mrvsp_comment); ! DBG
	write(DBG,'("ADJUST PRES pqc mrvsp: ",f9.2," mrcyc: ",i6," cmnt: ",a)')pqc_mrvsp,pqc_mrcyc,trim(pqc_mrvsp_comment); ! DBG
	write(DBG,'("ADJUST PRES out mrvsp: ",f9.2," mrcyc: ",i6," cmnt: ",a)')out_mrvsp,out_mrcyc,trim(out_mrvsp_comment); ! DBG
	write(DBG,'(" ")');                                                                                                 ! DBG
    endif

    ! Test for nlev=0, no offset needed
    if(nlev == 0)then
	do_pres_adj = .false.;
	pres_adj_comment = "PRES ADJUSTMENT NOT REQUIRED - NLEV==0";
	write(msg,'("ADJUST PRES - No adjustment will be applied - nlev == 0.")'); call wrttrc(sub,msg);
	return
    endif

    if(use_spres_for_offset)then
	pres_offset = surf_pres;
	pres_adj_comment = "PRESSURE OFFSET FROM CURRENT SURFACE PRESSURE";
    elseif(use_mrvsp_for_offset)then
	pres_offset = out_mrvsp;
	pres_adj_comment = "PRESSURE OFFSET FROM PRIOR MRVSP";
    else
	write(errmsg(1),'("FATAL ERROR 716 - failure in adjust_pres")')
	write(errmsg(2),'("FATAL ERROR 716 - unable to determine pres_offset")');
	call error_exit(sub,715,2,errmsg); ! adjust_pres in data_arrays - unable to determine pres_offset
    endif

    if(DEBUG)then;
	write(msg,'("ADJUST PRES hasDATA:  P p P p P p P p T T T T S s S s S s S s D D D D D")'); call wrtdbg(sub,msg);
	write(msg,'("ADJUST PRES hasDATA: ",25l2)')hasDATA; call wrtdbg(sub,msg); ! MAXDAT=25 for N_PROF=4
    endif
    do iprf=1,nprf
	if(hasPRES(iprf))then
	    ireg = regPRES(iprf); iadj = adjPRES(iprf)
	    hasDATA(iadj) = .true.; proftot(iadj) = proftot(ireg);
	    do ilev=1,nlev
		if(profdata(ireg,ilev) /= FILL)then;
		    profdata(iadj,ilev) = profdata(ireg,ilev) - pres_offset
		    profflag(iadj,ilev) = NOQC
		endif
	    enddo
	endif
    enddo
    if(DEBUG)then; write(msg,*)hasDATA; write(msg,'("da.adjust_pres end hasDATA: ",a)')trim(msg); call wrtdbg(sub,msg); endif
end subroutine adjust_pres

!==========================================================================
subroutine adjust_psal(caller)

    character(*), intent( in ) :: caller

    integer :: ireg, iadj

    character(len=11), parameter :: sub = 'adjust_psal'

    ! do the adjustment
    do iprf=1,nprf
	if(hasPSAL(iprf))then;
	    if(do_psal_prof(iprf))then;				! The STS auxiliary CTD data are NOT adjusted
		ireg = regPSAL(iprf); iadj = adjPSAL(iprf)
		hasDATA(iadj) = .true.; proftot(iadj) = proftot(ireg);
		do ilev=1,nlev
		    if(profdata(ireg,ilev) /= FILL)then;
			profdata(iadj,ilev) = profdata(ireg,ilev) + psal_offset;
			profflag(iadj,ilev) = NOQC
		    endif
		enddo
	    elseif(STS_prof(iprf))then;			! STS profile from AUXILIARY CTD (primary CTD dne above)
		ireg = regPSAL(iprf); iadj = adjPSAL(iprf)
		hasDATA(iadj) = .true.; proftot(iadj) = proftot(ireg);
		do ilev=1,nlev
		    if(profdata(ireg,ilev) /= FILL)then;
			profdata(iadj,ilev) = profdata(ireg,ilev);	! STS_prof - do NOT apply adjustment
			profflag(iadj,ilev) = NOQC			! Just copy the unadjusted psal to the adjusted col
		    endif
		enddo
	    else
		ireg = regPSAL(iprf); iadj = adjPSAL(iprf)
		do ilev=1,nlev
		    if(profdata(ireg,ilev) /= FILL)then;
			profdata(iadj,ilev) = profdata(ireg,ilev);	! psal adj. not requested, but pres is adjusted
			profflag(iadj,ilev) = NOQC			! Just copy the unadjusted psal to the adjusted col
		    endif
		enddo
	    endif	! if(do_psal_prof(iprf))
	endif		! if(hasPSAL(iprf)
    enddo		! do iprf=1,nprf

end subroutine adjust_psal

  !==========================================================================
subroutine wrtdbgdataflags(label,i1,i2,i3b,i3e)
  !
  ! Note if MAXFLAGS is changed, change last "int" field to be MAXFLAGS
  !
    character(*), intent( in ) :: label
    integer, intent( in ) :: i1         ! i1 is the "data colun" index
    integer, intent( in ) :: i2         ! i2 is the level index
    integer, intent( in ) :: i3b        ! i3b is thenbeginng of test flags
    integer, intent( in ) :: i3e        ! i3e is the ending  of test flags

    write(msg,'(a,i3,i5,i2,":",i1," data: ",f9.2," flags: ",i3," = ",7i3)')trim(label),i1,i2,i3b,i3e,profdata(i1,i2), &
        profflag(i1,i2),(testflag(i1,i2,i),i=i3b,i3e);
    call wrtdbg("wrtdbgdataflags",msg);

end subroutine wrtdbgdataflags

  !==========================================================================
subroutine test_juldays1950()
    integer          :: idate(6);
    double precision :: juld
    integer iyr,imo,idy,ihh,imn,isc

    character(len=16), parameter ::  sub = 'test_juldays1950'

    equivalence (idate(1),iyr), (idate(2),imo),(idate(3),idy),(idate(4),ihh),(idate(5),imn),(idate(6),isc)

    imo = 1; idy = 1; ihh = 0; imn = 0; isc = 0;
    do iyr = 1950,2020
        juld  = juldays1950(idate)
        if(DEBUG)then; write(msg,'("DAY ONE: ",i4,5i3,f8.1)')idate,juld; call wrtdbg(sub,msg); endif
     enddo
    iyr = 1950; imo = 1; idy = 1; ihh = 0; imn = 0;
    do isc = 0,59
        juld  = juldays1950(idate)
        write(msg,'(i4,5i3)')idate
        if(DEBUG)then; write(msg,*)'TEST JULD seconds: ',trim(msg),' juld: ',juld; call wrtdbg(sub,msg); endif
    enddo
end subroutine test_juldays1950

  !==========================================================================
double precision function juldays1950(JDATEP)

  !     Float reference date : 1950,01-01 00:00 --- 2433283
  !     jref=2433283    ! 01/01/1950
  !..   Finds elapsed time in minutes from Jan. 1, 1950
  !     Mike Minton, AOML, 30-Apr-1991
  !
  !     JDATEP AND JDATE (1)= YEAR
  !     JDATEP AND JDATE (2)= MONTH
  !     JDATEP AND JDATE (3)= DAY
  !     JDATEP AND JDATE (4)= HOUR
  !     JDATEP AND JDATE (5)= MIN
  !     JDATEP AND JDATE (6)= SEC
  !
  !     inputs: integer  --- year,month,day,hour,min,sec
  !     output: double precision  --- juldays1950
  !
  !  check value 2001,7,25,19,14,0   ->   18833.801388919353
  !
  integer JDATE(6),JDATEP(6)
  integer i
  double precision dble_hrs, dble_mns, dble_scs
  double precision trns_hrs, trns_mns, trns_scs
  double precision tmpday, base_day
  trns_hrs = 1.0/24.0; trns_mns = 1.0/1440.0; trns_scs = 1.0/86400.0
  base_day = 2433283.0
  !
  do i=1,6
    JDATE(i)=JDATEP(i)
  enddo
  CALL JULDAT(JDATE(3),JDATE(2),JDATE(1),i) ! This give the non-decimal julian day
  ! ! 2433283 = JULDAT for Jan.1,1950, 1440 = min/day, 60 = min/hr
  juldays1950=real(jdate(4))/24.+real(jdate(5))/1440.+ real(jdate(6))/86400.
  dble_hrs = dble(jdate(4))*trns_hrs; dble_mns = dble(jdate(5))*trns_mns; dble_scs = dble(jdate(6))*trns_scs;
  juldays1950 = (dble(i)-base_day) + dble_hrs + dble_mns + dble_scs
  return
end function juldays1950

  !======================================================================================
double precision function get_utc_juld()
    integer :: ivalues(8)			! year,mon,day,hr,mn.sc,ms
    integer :: idate (6)			! array to stuff into juldays1950
    integer :: in				! counter
    integer          :: utc_diff_mins		! difference between local and utc in INTEGER MINUTES
    double precision :: utc_diff_juld		! difference between local and utc in DBLE JULDs

	    character(len=12), parameter ::  sub = 'get_utc_juld'

    call date_and_time(VALUES=ivalues)
    do in=1,3; idate(in)=ivalues(in); enddo
    do in=4,6; idate(in)=ivalues(in+1); enddo

    utc_diff_mins = ivalues(4);
    utc_diff_juld = 0.0;
    if(utc_diff_mins /= 0)then
	utc_diff_juld = DBLE(utc_diff_mins)/1440.0;
    endif

    get_utc_juld = juldays1950(idate);		! Convert the int date into JULD

end function get_utc_juld


!-------------------------------------------------------------------
subroutine julday(iyr,imon,iday,ihour,rtime)
integer :: iyr,imon,iday,ihour
real    :: rtime
integer :: jref, m, j, jday, ih, imin, rmin
jref=2439126    ! 12/31/65
call juldat(iday,imon,iyr,j)
jday=j-jref
ih=ihour/100
imin=ihour-ih*100
rmin=ih*60+imin
rtime=float(jday)+rmin/(24.0*60.0)
return
end subroutine julday

!-------------------------------------------------------------------
!     JULDAT.FOR
!     JULDAT... CONVERTS GREGORIAN CALENDAR DATES TO CORRESPONDING
!               JULIAN DAY NUMBERS.
!     PROGRAMMER... J. MALTAIS
!     DATE      ... NOVEMBER 6,1967
!*
!     Y... MAY BE EXPRESSED AS CENTURY PLUS YEAR WITHIN CENTURY
!          OR MAY BE EXPRESSED AS TWO PLACE INTEGER ('73') FOR THE
!          TWENTIETH CENTURY
!
SUBROUTINE JULDAT(DP,MP,YP,J)
INTEGER J,D,M,Y,C,MO,YR,DP,MP,YP
D=DP
M=MP
Y=YP
    1 IF(Y.LT.100)Y=Y+1900
YR=Y
IF(M.GT.2)GO TO 5
MO=M+9
YR=Y-1
GO TO 10
    5 MO=M-3
   10 C=YR/100
YR=YR-100*C
J=(146097*C)/4+(1461*YR)/4+(153*MO+2)/5+D+1721119
return
end subroutine juldat
!-------------------------------------------------------------------
subroutine j2arr3(julian,ldate)
!
! Input:  julian - an absolute Julain day (not day_of_year )
!                  ( use j2toi3 sub for day_of_year )
! Output: ldate - array of year,month,day
implicit NONE
integer*4 julian,ldate(3)
!..   pl1900 entry,  adds ipaddt to 1900,1,1,0,0,0,0.
!     entry pl1900(ldate,ipaddt)
!     ldate(1)= year    iaddt(1)=days
!     ldate(2)= month   iaddt(2)=hours
!     ldate(3)= day     iaddt(3)=minutes
!
integer m(12),l(3),l2,i
equivalence (l(2),l2)
data m /31,28,31,30,31,30,31,31,30,31,30,31/
!
l(1)=1900
l(2)=1
!      l(3)=julian-2415020	! Subtract 12/31/1899
l(3)=julian+1	        ! Dave Behringer's analysis time
  120 if(l(2) /= 2)goto 110
m(2)=28
if(mod(l(1),4) /= 0)goto 110
if(mod(l(1),100) /= 0)goto 170
if(mod(l(1),400) /= 0)goto 110
  170 m(2)=29
  110 if(l(3).le.m(l2))go to 888
l(3)=l(3)-m(l2)
if(l(2).lt.12)go to 150
l(2)=l(2)-12
l(1)=l(1)+1
  150 l(2)=l(2)+1
goto 120
!
  888 continue
  do i=1,3
   ldate(i)=l(i)
 enddo
return
end subroutine j2arr3

!=========================================================================
! character posn_date_time
! Pass in the position line, get back the character*114 date_time
subroutine posn_date_time(caller,poslin,c14)

character(*), intent( in )       :: caller;		! who is calling me?
character(*), intent( in )       :: poslin;		! position line
character(len=14), intent( out ) :: c14;

integer :: iyr, imo, idy, ihr, imn, isc		! year/month/day/hour/min/sec
! NOT USED character(len=14) :: int2ch14			! convert ints to c*14


read(poslin,'(27x,i4,x,i2,x,i2,x,i2,x,i2,x,i2)')iyr, imo, idy, ihr, imn, isc
write(c14,'(i4,5i2.2)')iyr, imo, idy, ihr, imn, isc

end subroutine posn_date_time

!=========================================================================
! This adjust ONLY minutes in a big int date/time (into 14 ch)
subroutine adjust_minutes(int_inp,minutes,int_out)

integer*8 , intent( in )  :: int_inp		! input big integer - the date to modify
integer*4 , intent( in )  :: minutes		! input small integer - themnutes to add/sub
integer*8 , intent( out ) :: int_out		! output big integer - the date after adjustment

integer :: days(0:13) = (/ 31, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31 /)
                         ! d   j   f   m   a   m   j   j   a   s   o   n   d   j
integer :: iyr, imo, idy, ihr, imn, isc         ! year/month/day/hour/min/sec
character(len=14) :: c14;

character(len=14), parameter ::  sub = 'adjust_minutes'

write(c14,'(i14)')int_inp
read(c14,'(i4,5i2)')iyr, imo, idy, ihr, imn, isc
imn = imn + minutes;
if(imn > 59)then	!  overflow mn, increment ihr
    do while(imn > 59)
	ihr = ihr + 1; imn = imn - 60
	if(ihr > 23)then
	    idy = idy + 1; ihr = ihr - 24;
	    if(idy > days(imo))then
		imo = imo + 1; idy = 1;
		if(imo > 12)then;
		    imo = imo - 12; iyr = iyr+ 1
		endif
	    endif
	endif
    enddo
endif
if(imn < 0)then	! underflow mn, decrement ihr
    do while(imn < 0)
	ihr = ihr -1; imn = imn + 60
	if(ihr < 0)then
	    idy = idy - 1; ihr = ihr + 24;
	    if(idy < 1)then
		idy = idy + days(imo); if(imo==2 .and. leap(iyr)) idy = 29; imo = imo - 1;
		if(imo < 1)then; imo = 12; iyr = iyr - 1; endif
	    endif
	endif
    enddo
endif
write(c14,'(i4.4,5i2.2)')iyr, imo, idy, ihr, imn, isc
read(c14,'(i14)')int_out

end subroutine adjust_minutes

!=========================================================================
!
! Detected lat/lon off the globe
!
! Note that this fuction does NOT do any tests for fill value - it returns TRUE
!
logical function impossible_location(lat,lon)

    real, intent( in ) :: lat   ! latitude to test
    real, intent( in ) :: lon   ! longitude to test

    impossible_location = .true.        ! Assume position is NOT valid


    if(lat >  90.0)  return
    if(lat < -90.0)  return
    if(lon >  180.0) return
    if(lon < -180.0) return

    impossible_location = .false.

end function impossible_location

!============================================================================
integer function frdtbase(caller,lat,lon)
! lon -180 to 180, lat -90 to 90

  character(*), intent( in ) :: caller        ! who is calling me?
  real, intent( in ) :: lat
  real, intent( in ) :: lon
  real LATN,LONW,lon360
  integer XSIZE,YSIZE,ii,jj,kk,len,dummy,intopo,SEEK_SET,ierr,ios
  integer*2 depth;
  real*8 consta,consta2;
  character(len=8), parameter ::  sub = 'frdtbase'

  dummy = 99999;

! LATN and LONW were changed based on the documentation on
! http://www.ngdc.noaa.gov/mgg/global/etopo5.HTML
! in the original files (ETOPO5.DAT, ETOPO5.DOS, see
! /a1/ARGO/qc/clima_mask/topo_readme.txt):
! latitudes range from the North Pole 90 deg N to 89 deg 55' S
! longitudes range eastward from 0 deg 0' East longitude to 359 deg 55', but
! J. Trinanes used LATN=89.9583333; LONW=-179.958333 for topo.bin
!
  consta=1.0/12.0;  consta2=consta/2.0;
  LATN=90.-consta2;  LONW=0.+consta2;  XSIZE=4320;  YSIZE=2160 ! for ETOPO5.DAT, ETOPO5.DOS

  lon360=lon
  if(LONW .eq. 0.) lon360=lon360+180.
  if(DBLEV== 4)then; write(msg,*)'LEV04 RTB lat, lon, lon360: ',lat,lon,lon360; call wrtdbg(sub,msg); endif

  SEEK_SET=0
  ! estimate index for fseek form latitude and longitude
  ! write(6,*) '(LATN+consta2-lat)/consta',(LATN+consta2-lat)/consta
  ii=(LATN+consta2-lat)/consta;
  jj=((lon360-(LONW-consta2))/consta);
  kk=(ii*XSIZE+jj)*2;
  if(DBLEV== 4)then; write(msg,'("LEV04 RTB ii: ",i6," jj: ",i6," kk: ",i8)')ii,jj,kk; call wrtdbg(sub,msg); endif

  if(kk>18662400)then
    write(6,*) 'warning in frdtbase - index kk exceeds file size: ',kk
    frdtbase=dummy; return
  elseif(kk<0)then
    write(6,*) 'warning in frdtbase - index kk is less than 0: ',kk
    frdtbase=dummy; return
  endif
  if(DBLEV== 4)then; write(msg,'("RTB TOPO TEST open unit ",i2," topo_file: ",a)')intopo,trim(topo_file); call wrtdbg(sub,msg); 
endif
  intopo=231
  open(unit=intopo,file=topo_file,status='old',action='read',iostat=ios,access="stream",form="unformatted");
  ! TEST open(unit=in,file=topo_file,status='old',action='read',iostat=ios,access="sequential",form="binary");
  call fseek(intopo,kk,SEEK_SET,ierr)
  if(DBLEV== 4)then; write(msg,'("LEV04 RTB in,kk,SEEK_SET,ierr: ",4i10)')intopo,kk,SEEK_SET,ierr; call wrtdbg(sub,msg); endif
  read(intopo) depth
  if(DBLEV== 4)then; write(msg,'("LEV04 RTB depth: ",i12)')depth; call wrtdbg(sub,msg); endif
  close(unit=intopo);

  frdtbase=depth;
  if(DBLEV== 4)then
    write(msg,'(" Position_on_land - lat,lon,lon360,depth: ",3f10.3,i10)')lat,lon,lon360,frdtbase;
    call wrttrc(sub,msg);
  endif

  return
end function frdtbase

!========================================================================================
!
!     Revised date: 29-Oct-10 (ER  - Add treatment of new gross pressure test.
!                   11-Nov-09 (ER  - Cleaning of unnecessary grey list
!                                    variables )
!                   17-Feb-05 (YHD - Modified after R. Mccreadie's e-mail)
!                   14-Jan-05 (YHD - All tests are ready)
!                   29-Oct-04 (YHD - convert QCTEST id to hex)
!    qctest2hex.f
!    ============
!    Please refer to reference table 11: QC test ids
!
!    id                       test
!    ----------------------------------------------------------------
!    1) 2                     Platform test
!    2) 4                     Date test
!    3) 8                     Location test
!    4) 16                    Position on Land test
!    5) 32                    Speed test
!    6) 64                    Global range test
!    7) 128                   Regional Global Parameter test
!    8) 256                   Pressure Increasing test
!    9) 512                   Spike test
!    10) 1024                 Top/Bottom Spike test          --- obsolete
!    11) 2048                 Gradient test
!    12) 4096                 Digit Rollover test            --- for SOLO only
!    13) 8192                 Stuck value test
!    14) 16384                Density Inversion test
!    15) 32768                Grey List test               --- started 2003,12
!    16) 65536                Gross Salinity or Temperature Sensor Drift test
!    17) 131072               Visual test   ??????
!    18) 262144               Frozen profile
!    19) 524288               Deepest pressure
!    20) 1048576              Questionable Argos position test from JAMSTEC for TRAJECTORY data
!    21) 2097152              Near-surface unpumped CTD salinity test
!    22) 4194304              Near-surface mixed air/water test

!----------------------------------------------------------------------------
subroutine qctest2hex(caller, ipf, hqc_perf, hqc_fail) ! iprf cngd to ipf, to eliminate conflicts

character(*), intent( in )      :: caller	! who is calling me?
integer, intent( in )           :: ipf	        ! which profile (pri or sec?)
character(len=*), intent( out ) :: hqc_perf     ! the test performed hex varialbe
character(len=*), intent( out ) :: hqc_fail     ! the test failed hex variable

integer id(NTESTS)                    ! number of tests (22 currently)

integer :: dqc_perf, dqc_fail
integer :: imax
integer :: pflag, tflag, sflag		! IGNORE DOXY flags (vmh unilateral 02apr2015)

character(len=10), parameter :: sub = 'qctest2hex'

character(len=16) :: tmp
character(len=1) :: sq = "'"; ! single-quote (for matlab line)
character(len=1) :: pc = "%"; ! percent (for matlab line)
       ! 1 2 3  4  5  6   7   8   9   10   11   12   13    14    15    16
data id /2,4,8,16,32,64,128,256,512,1024,2048,4096,8192,16384,32768,65536, &
	 131072,262144,524288,1048576,2097152,4194304/
	!    17     18     19      20      21      22
!
!--- All tests performed on ref. table 11
!    except test 10.
if(DBLEV==31)then; write(msg,'("HEX Enter qctest2hex for profile ",i1)'); call wrtdbg(sub,msg); endif
hqc_perf='                ';		! clear the hex variables
hqc_fail='                ';		! clear the hex variables

dqc_perf=0;				! clear the decimal variables
dqc_fail=0;				! clear the decimal variables

do i=1,NTESTS
    if(DEBUG) write(DBG,'("QCTEST_FLAG ip: ",i2," test: ",i3," FLAG: ",i2)')ipf,i,QCTEST_FLAG(ipf,i)
    if(QCTEST_FLAG(ipf,i) > 0)then;		! This test WAS PERFORMED
	dqc_perf = dqc_perf + id(i);
	if(QCTEST_FLAG(ipf,i) > 1)then;		! This test, however, FAILED
	    dqc_fail = dqc_fail + id(i);
	    if(DEBUG) write(DBG,'(a," FAIL")')trim(msg);
	    if(DBLEV==31) write(msg,1001)'> 1 ',ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile
	else
	    if(DEBUG) write(DBG,'(a," PERF")')trim(msg);
	    if(DBLEV==31) write(msg,1001)'> 0 ',ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile
	endif
    else
	if(DEBUG) write(DBG,'(a," NOTP")')trim(msg);
	if(DBLEV==31) write(msg,1001)'= 0 ',ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile
    endif
    if(DEBUG) call wrtlog(sub,msg);
    1001 format("QCHIST flag ",a3,i3,i3,i3,2i12,2x,a)
    ! write(msg,'("QCHIST flag > 1 ",i3,i3,i3,2i12,2x,a)')ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile
    ! write(msg,'("QCHIST flag > 0 ",i3,i3,i3,2i12,2x,a)')ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile
    ! write(msg,'("QCHIST flag = 0 ",i3,i3,i3,2i12,2x,a)')ipf,i,QCTEST_FLAG(ipf,i),dqc_perf,dqc_fail,profile

enddo

call dec2hex(sub,dqc_perf,hqc_perf);
call dec2hex(sub,dqc_fail,hqc_fail);

write(msg,'("cl;check_val=",a1,a,a1,";check_val_f=",a1,a,a1,"; history4netcdf ",a1," QCHIST ",i2,2i9,3x,a)') &
	sq,trim(hqc_perf),sq,sq,trim(hqc_fail),sq,pc,ipf,dqc_perf,dqc_fail,trim(profile)
call wrtlin(sub,msg);

write(msg,'("QCHIST iprof: ",i2," perf: ",i9,x,a16," fail: ",i9,x,a16)')iprf,dqc_perf,hqc_perf,dqc_fail,hqc_fail
call wrtdbg(sub,msg);

return
end subroutine qctest2hex

!========================================================================================
!
! Adapted from deci2hex.f in qc2netcdf
!
subroutine dec2hex(caller,qcdec,qchex)

character(*),      intent( in )  :: caller
integer,           intent( in )  :: qcdec
character(len=16), intent( out ) :: qchex ! change later to a char function

integer dec,dec_num,hexb,quo,remain,i,ix,j,imax,status

character h(10),hexnum(16)*1,hex*16
parameter (hexb=16,imax=10)
integer r(10)

character(len=7),parameter :: sub = "dec2hex";

data hexnum/'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'/

dec = qcdec;	! we do not want to change the value of an intent=in variable

ix=0
hex='                '      ! clear it
dec_num=dec
!
! --- A decimal number is divided by 16 and keep the remainder
!     until quotient is 0.
!
 10   continue
 if(DBLEV==37)then; write(msg,*)'TESTLEV37 force error in dec2hex'; ix = 1; call wrtdbg(sub,msg); endif
if(ix.gt. 10)then
    write(errmsg(1),'("NON-FATAL ERROR 227 write_ncprof")');
    write(errmsg(2),'("NON-FATAL ERROR 227 converting dec to hex in ",a)')sub;
    write(errmsg(3),'("NON-FATAL ERROR 227 some thing called ix (",i2,") is greater than imax=10")');
    call error_report(sub,227,3,errmsg); ! dec2hex in data_arrays error converting dec to hex
    qchex = 'ffffffffffffffff'; return
endif
quo=dec/hexb
remain=dec-hexb*quo
if(quo.eq.0)then
  ix=ix+1
  r(ix)=remain
else
  ix=ix+1
  r(ix)=remain
  dec=quo
  go to 10
endif

do i=1,ix
  h(i)=hexnum(r(i)+1)
enddo
!
! --- Reverse Remainders to get hex number.
!
j=0                       ! initialize
do i=ix,1,-1
  j=j+1
  hex(j:j)=h(i)
enddo
if(DEBUG)then
    write(msg,*)'dec2hex dec: ',dec_num;  call wrtdbg(sub,msg);
    write(msg,*)'dec2hex hex: ',hex(1:j); call wrtdbg(sub,msg);
endif

qchex = hex;

return
end subroutine dec2hex

! ===============================================================================
  logical function check_cdate(cdate)

    use communication                       ! module

    implicit none

    character(len=*) , intent( in ) :: cdate    ! The date to be parsed and checked
    integer :: iyr,imo,idy,ihh,imn,isc
    character(len=16), parameter   :: sub = 'check_cdate';

    check_cdate = .false.               ! Assume the date is bad

    ! Which kind of cdate do we have?

    if(len(trim(cdate)) == 14)then
	read(cdate,'(i4,i2,i2,i2,i2,i2)')iyr,imo,idy,ihh,imn,isc
    elseif(len(trim(cdate)) == 19)then
	read(cdate,'(i4,i3,i3,i3,i3,i3)')iyr,imo,idy,ihh,imn,isc
    else
        write(msg,'("This data is neither 14 nor 19 characters: ",a)')trim(cdate); 
	call wrttrc(sub,msg); return; ! The functional value has already been set to false
    endif

    check_cdate = check_date(iyr,imo,idy,ihh,imn,isc);			! See if we have a good date
  end function check_cdate

! ===============================================================================
  logical function check_date(iyr,imo,idy,ihh,imn,isc)

    use communication                       ! module

    implicit none
    integer, intent( in ) :: iyr
    integer, intent( in ) :: imo
    integer, intent( in ) :: idy
    integer, intent( in ) :: ihh
    integer, intent( in ) :: imn
    integer, intent( in ) :: isc

    ! These are used to see if a date is in the future
    integer*8 :: loc_date_time, gmt_date_time		! local time and GMT
    integer*4 :: min_adj				! number of minutes to adjust convert local to gme
    character(len= 8) :: today_date
    character(len=10) :: today_time
    character(len= 5) :: today_zone
    integer :: today(8)
    character(len=8) :: yyyymmdd
    character(len=14) :: c14		! Used to adjust date to GMT
    integer itotst, itoday
    double precision  :: gmt_jul_date, tst_jul_date	! julds for "now" and for profile date
    double precision  :: ddiff; ! double-precision diff
    character(len=80) :: cjuld, cdiff  ! output for double-precesion values
    character(len=14) :: gmtc14,      tstc14
    character(len=20) :: gmtc20,      tstc20
    character(len=SZC) :: tmsg;
    integer :: tdate(6)					! test int array to find julds

    integer :: id, idays_mon(12) = (/31,28,31,30,31,30,31,31,30,31,30,31/)
    integer :: tyr;! Test uear = = g,t. exce[t last 2 days of the year, increment by 1

    character(len=10), parameter   :: sub = 'check_date';

    check_date = .false.               ! Assume the date is bad - early return if bad date found

    ! For the future test, convert iyr,imo,idy to char*8, then to integer
    write(yyyymmdd,'(i4,i2.2,i2.2)')iyr,imo,idy;
    read(yyyymmdd,*) itotst

    ! Check out the datd_and_time call

    call date_and_time( today_date, today_time, today_zone, today)

    write(c14,'(i4,5i2.2)')today(1),today(2),today(3),today(5),today(6),today(7); read(c14,*)loc_date_time;
    min_adj = -today(4); call adjust_minutes(loc_date_time,min_adj,gmt_date_time)

    ! Convert "right now" gmt to jul_day
    write(gmtc14,'(i14)')gmt_date_time;		! Convert the gmt_)date_tie integer to a string
    read(gmtc14,'(i4,5i2)')gmtdate;		! This is available program-wide
    read(gmtc14,'(i4,5i2)')tdate;		! Then read year/mon/day/hr/mn/sc from it
    gmt_jul_date = juldays1950(tdate);	! Convert to julian day
    ! Stuff the float date year/mon/day/hr/mn/sc tnto the tdate arrays, and convert ot float julian day
    tdate(1) = iyr; tdate(2) = imo; tdate(3) = idy; tdate(4) = ihh; tdate(5) = imn; tdate(6) = isc;
    tst_jul_date = juldays1950(tdate);
    write(tstc14,'(i4,5i2.2)')tdate;

    if(tst_jul_date > gmt_jul_date)then
	! We've got a problem here - but it MAY be ok, so issue WARNING, not NON-FATAL ERROR
        
	ddiff = gmt_jul_date - tst_jul_date;
	write(cdiff,*)ddiff;
	
	write(cjuld,'(f20.7)')gmt_jul_date; write(msg,1001)gmtc14,trim(cjuld); call wrtstd(sub,msg);
	write(cjuld,'(f20.7)')tst_jul_date; write(msg,1002)tstc14,trim(cjuld); call wrtstd(sub,msg);
	write(cdiff,'(f15.7)')ddiff; write(msg,1003)trim(cdiff);               call wrtstd(sub,msg);
        1001 format("GMT ERR - in check_date - may be BADGPS Date in the future: gmt c: ",a," < gmt jul: ",a);
        1002 format("GMT ERR - in check_date - may be BADGPS Date in the future: tst c: ",a," < tst jul: ",a)
        1003 format("GMT ERR - in check_date - may be BADGPS Date in the future:  diff: ",a," (in days)")
        if(badgpsfix_pos_found)then
            write(msg,'("GMT ERR - BADGPS have been identified in this position block")');
            call wrtstd(sub,msg);
        endif

	write(wrnmsg,'("WARNING 115 Date in the future, grep ",a,"TRAP GMT ERR",a," in the logs")')SQ,SQ; ! WRN
	call warning(sub,115,wrnmsg); ! WARNING 115 in data_arrays function check_date ! WRN
    endif
    ! find the test (today's) date
    ! THIS TEST PERFORMED ABOVE WITH JULDAYS read(today_date,*)itoday
    if(iyr< 1997)then;
	write(msg,'("check_cdate W BAD YEAR (<1997): ",i4)')iyr;
	call wrttrc(sub,msg); return;
    endif
    tyr = gmtyr;
    if(imo>=12 .and. idy > 29) tyr = tyr + 1;
    if(iyr > tyr)then;
	 write(msg,'("W BAD YEAR (> test year): iyr: ",i4," tyr: ",i4)')iyr,tyr;
	call wrttrc(sub,msg); return;
    endif
    if(imo< 1)then;
	write(msg,'("W BAD MON (< 1): ",i2)')imo;
	call wrttrc(sub,msg); return;  
    endif
    if(imo>12)then;
	write(msg,'("W BAD MON (>12): ",i2)')imo;
	call wrttrc(sub,msg); return;  
    endif
    if(idy< 1)then;
	write(msg,'("W BAD DAY (< 1): ",i2)')idy;
	call wrttrc(sub,msg); return;  
    endif
    id=idays_mon(imo); if(imo==2)then; if(mod(iyr,4).eq.0)then; id = id+1; endif; endif ! Leap year
    if(idy>id)then;
	write(msg,'("W BAD DAY (>",i2,"): ",i2)')id,idy;
	call wrttrc(sub,msg);return;
    endif
    if(ihh< 0)then;
	write(msg,'("W BAD HOUR (< 0): ",i2)')ihh;
	call wrttrc(sub,msg);return;  
    endif
    if(ihh>23)then;
	write(msg,'("W BAD HOUR (>23): ",i2)')ihh;
	call wrttrc(sub,msg);return;  
    endif
    if(imn< 0)then;
	write(msg,'("W BAD MIN (< 0): ",i2)')imn;
	call wrttrc(sub,msg);return;  
    endif
    if(imn>59)then;
	write(msg,'("W BAD MIN (>59): ",i2)')imn;
	call wrttrc(sub,msg);return;  
    endif
    if(isc< 0)then;
	write(msg,'("W BAD SEC (< 0): ",i2)')isc;
	call wrttrc(sub,msg);return;  
    endif
    if(isc>59)then;
	write(msg,'("W BAD SEC (>59): ",i2)')isc;
	call wrttrc(sub,msg);return;  
    endif 
    check_date = .true.                ! If it make it here, we have a good date
  end function check_date

! ===============================================================================
! subroutine cdate_check(cdate,io)
! ===============================================================================
! Notes
!
!  Revision date: 14-Aug-12 VMH - wrapper to parse char time variable and
!                                 use date_check to check for impossible dates
!
!    cdate_check.f
!    ============
!    Parses a date character string and checks validity
!
!    output: io=0  good; io=1 bad
!
subroutine cdate_check(cdate,io)

implicit none

character*19 cdate
integer iyr,imo,idy,ihh,imn,ndays,io
character(len=11), parameter :: sub = 'cdate_check'
     
character(len=1)   :: testchar

! First, parse the input character-date
testchar = cdate(5:5) 
if(testchar .eq. ' ')then       ! format: yyyy mm dd hh mm ss
    read(cdate(1:4),'(i4)') iyr
    read(cdate(6:7),'(i2)') imo
    read(cdate(9:10),'(i2)') idy
    read(cdate(12:13),'(i2)') ihh
    read(cdate(15:16),'(i2)') imn
else                             ! format: yyyymmddhhmmss 
    read(cdate(1:4),'(i4)') iyr
    read(cdate(5:6),'(i2)') imo
    read(cdate(7:8),'(i2)') idy
    read(cdate(9:10),'(i2)') ihh
    read(cdate(11:12),'(i2)') imn
endif

if(DEBUG) write(DBG,*)'cdate_check in data_arrays calls DATE_CHECK @2 for 19chr date: ',cdate

call date_check(iyr,imo,idy,ihh,imn,io)
return

end subroutine cdate_check

! ===========================================================================================================
! subroutine date_check(iyr,imo,idy,ihh,imn,io)
!
!  Revision date: 01-Oct-10 (ER  - fixed bug calculating current date
!                 31-Mar-10 (ER  - printing date from profile and now
!                                  for comparison when error future date
!                                  occurs )
!                 23-Mar-10 (ER  - using Zulu date for comparison)
!                 21-Oct-09 (ER  - add date in the future test)
!                 02-Oct-02 (YHD - impossible date exits with status=6)
!
!    Date_check.f
!    ============
!    Checks impossbile date and time.
!    yhd, 01-Jun-01
!
!    output: io=0  good; io=1 bad
!
subroutine date_check(iyr,imo,idy,ihh,imn,io)

use communication				! module

implicit none

integer iyr,imo,idy,ihh,imn,idays_mon(12),ndays,io
character(len=10), parameter :: sub = 'date_check'
data idays_mon/31,28,31,30,31,30,31,31,30,31,30,31/
     
! variables to be used on 'is in the future test' 
character(len=12)  :: today 
character(len=4)   :: c_year
character(len=2)   :: c_month, c_day, c_hour, c_min
character(len=1)   :: aux_1_char
character(len=12)  :: c_input_date ! Input date as a char.

integer  :: time  ! Function that returns # of sec since 1/1/1900 as an integer
integer, dimension(9) ::  tarray(9)

if (iyr<1997)then
    write(errmsg(1),'("FATAL ERROR 711 - qc02_imp_date fails")')
    write(errmsg(2),'("FATAL ERROR 711 - year less than 1997: ",i4)')iyr
    call error_exit(sub,711,2,errmsg); ! date_check in data_arrays - year less than 1997
endif
if (imo.lt.1.or.imo.gt.12)then
    write(errmsg(1),'("FATAL ERROR 712 - qc02_imp_date fails")')
    write(errmsg(2),'("FATAL ERROR 712 - Month error - imo=",i4)')imo
    call error_exit(sub,712,2,errmsg); ! date check in data_arrays -  Month error
endif
ndays=idays_mon(imo)
if (mod(iyr,4).eq.0.and.imo.eq.2) ndays=idays_mon(imo)+1     
if (idy.lt.1.or.idy.gt.ndays)then
    write(errmsg(1),'("FATAL ERROR 713 - qc02_imp_date fails")')
    write(errmsg(2),'("FATAL ERROR 713 - Day error - idy=",i4)')idy
    call error_exit(sub,713,2,errmsg); ! date check in data_arrays - Day error
endif
if (ihh.lt.0.or.ihh.gt.24)then
    write(errmsg(1),'("FATAL ERROR 714 - qc02_imp_date fails")')
    write(errmsg(2),'("FATAL ERROR 714 - Hour error - ihh=",i4)')ihh
    call error_exit(sub,714,2,errmsg); ! date check in data_arrays - Hour error
endif
if (imn.lt.0.or.imn.gt.60)then
    write(errmsg(1),'("FATAL ERROR 715 - qc02_imp_date fails")')
    write(errmsg(2),'("FATAL ERROR 715 - Min error - imn=",i4)')imn
    call error_exit(sub,715,2,errmsg); ! date check in data_arrays - Min error
endif
!
! Date, Time input is valid
! Now let's check whether it is in the future
!

write(c_input_date,'(i4,i2.2,i2.2,i2.2,i2.2)')iyr,imo,idy,ihh,imn

! Conform current date & time
!
!   -- Return system time in Zulu
call gmtime( time(), tarray )  
!   -- Get year
write(today(1:4),'(i4)') 1900+tarray(6)
!   -- Get month
if (tarray(5)<9)then
  write(today(5:6),'(A1,i1)')"0",tarray(5)+1 
else
  write(today(5:6),'(i2)')tarray(5)+1    
endif
!  -- Get day
if (tarray(4)<10)then
  write(today(7:8),'(A1,i1)')"0",tarray(4)  
else
  write(today(7:8),'(i2)')tarray(4) 
endif
!  -- Get hour
if (tarray(3)<10)then
  write(today(9:10),'(A1,i1)')"0",tarray(3) 
else
  write(today(9:10),'(i2)')tarray(3)
endif
! Get minute
if (tarray(2)<10)then
  write(today(11:12),'(A1,i1)')"0",tarray(2) 
else
  write(today(11:12),'(i2)')tarray(2) 
endif

if (c_input_date < today)then
    io = 0	! Signal a good date
else                     
    write(errmsg(1),'("NON-FATAL ERROR 228 - qc02_imp_date fails")')
    write(errmsg(2),'("NON-FATAL ERROR 228 - Date in the future=",i4)')imn
    write(errmsg(3),100)sub,": Date check in the future& ", c_input_date, " Now: ",today
    call error_report(sub,228,3,errmsg); ! date_check in data_arrays date in the future
    100 format("NON-FATAL ERROR 228 - ",A10,A27,1x,A12,A6,A12)
    io = 1	!Signal a bad date
endif

return
end subroutine date_check


! ======================================================================================
! subroutine analyze_positions
! Analyze the position block qc flags

subroutine analyze_positions(caller,ret) ! This will load the data values into the arrays
character(*), intent(in)     :: caller
integer, intent(out)         :: ret ! where there problems?

character(len=15), parameter   :: sub = 'analyze_positns';
integer :: num_pos_lines
integer :: it, in ! it = this i, in = i+1
integer :: iyr,imo,idy,ihh,imn,isc, idate(6)
equivalence (idate(1),iyr), (idate(2),imo),(idate(3),idy),(idate(4),ihh),(idate(5),imn),(idate(6),isc)

ret = SUCCESS; ! Assume SUCCESS
write(DBG,*)'DBG ANA_POS entry, caller is: ',trim(caller)
num_pos_lines=npos;
if(caller == 'read_pqc')then; num_pos_lines = ppos; endif ! set the number of pos based on who calls the subroutine.
               
if(DEBUG)then
    write(msg,'("DBG ANA_POS called by ",a,"for profile ",a24)')trim(caller),trim(profile)
    write(LOG,'(a)')trim(msg); write(DBG,'(a)')trim(msg); ! write(DBG,'("TEXT EXIT")'); call exit(0);
endif
do i=1,num_pos_lines    
    read(pis_posnqc(i),*)ipis_posnqc(i)      !pi supplied flag
    read(pos_dateqc(i),*)ipos_dateqc(i)      !datecheck flag
    read(pos_posiqc(i),*)ipos_posiqc(i)      !impossible location qc flag
    read(pos_onland(i),*)ipos_onland(i)      !onland check flag
    read(pos_speedc(i),*)ipos_speedc(i)      !speedcheck flag
            
    ipos_posnqc(i) = 0;			! The position has not yet been tested      
    if(ipis_posnqc(i) == 8)then;	! interpolated                                                           
	   position_interpolated = .true.; ! We found an interpolated position
	   ! NOTEST ipos_onland(i) = 4; ! TEST the error message
	   if(ipos_onland(i) == 4 .or. ipos_posiqc(i) == 4)then	! Something is wrong with the interpolation!!!
		bad_position_found(i) = .true.;
               ipos_posnqc(i) = 4;
	       write(errmsg(1),'("NON-FATAL ERROR 229 in analyze_positions called by ",a)')trim(caller)
	       write(errmsg(2),'("NON-FATAL ERROR 229 Interpolated position, but the position is ON LAND or OFFWORLD")')
	       write(errmsg(3),'("NON-FATAL ERROR 229 It occurres in position: ",i3)')i;
	       write(errmsg(4),'("NON-FATAL ERROR 229 This position will be marked as BAD, QC=4")')
	       call error_report(sub,229,4,errmsg); ! analyze_positions in data_arrays interpolated position bad or land
	       ret = FAILURE; ! A NON-FATAL error has occurred, inform the caller
 	   endif
	   ipos_posnqc(i) = ipis_posnqc(i);	! 8 - signal interpolated posn
    elseif(ipis_posnqc(i) == 9)then;            ! 9 - missing
	   ! For now, allow 9 to be valid, HOWEVER, write a lognote, and check every so often
	   write(logmsg,'(" LOGNOTE 198 PIQC=9 pos: ",i3,": ",x,a)')i,trim(posline(i));                                     ! LGNPI9
	   call lognote(sub,198,logmsg);                                                                                    ! LGNPI9
           ipos_posnqc(i) = ipos_posiqc(i)
    elseif(ipis_posnqc(i) >= 5)then;		! 5 - real-time PI changing IS NOT DONE - if 5, then is an error
	   write(errmsg(1),'("NON-FATAL ERROR 230 in analyze_positions called by ",a)')trim(caller)
	   write(errmsg(2),'("NON-FATAL ERROR 230 Ihe PI-supplied qc is INVALID: ",i2)')ipis_posnqc(i);
	   write(errmsg(3),'("NON-FATAL ERROR 230 It occurres in osition: ",i3)')i;
	   write(errmsg(4),'("NON-FATAL ERROR 230 This position will use the rreal-time POSQ only")');
	   call error_report(sub,230,4,errmsg); ! analyze_positions in data_arrays interpolated position bad or land
	   ret = FAILURE; ! A NON-FATAL error has occurred, inform the caller
	   ipos_posnqc(i) = ipos_posiqc(i)  
    else	      
	   if(ipis_posnqc(i) > ipos_posiqc(i))then    !PI supplied QC is worse than aoml's overall qc. 
	       if(DEBUG) write(DBG,'("DBG ANQCPOS pos: ",i3," pis_posnqc (",i1,") used, not pos_posiqc(",i1,")")')i, &
	   	ipis_posnqc(i),ipos_posiqc(i)
	       ipos_posnqc(i) = ipis_posnqc(i)
	   else
	       ipos_posnqc(i) = ipos_posiqc(i)                   
	   endif                                   
    endif 
    write(pos_posnqc(i),'(i1)')ipos_posnqc(i) 
enddo

if(make_qcprof)then

    do i=1,num_pos_lines		! Test for monotonically increasing time in posiiton block
	read(this_date_time(i),1000)idate
	1000 format(i4,5i2.2); ! read/write format for 14ch dates
	juld_date_time(i) = juldays1950(idate);
    enddo
    do i=1,(num_pos_lines-1)
	it = i; in = i+1;
	juld_diff = juld_date_time(in) - juld_date_time(it);
     if(juld_diff .lt. 0.0)then
	    write(errmsg(1),'("FATAL ERROR 701 An error occurred in analyze_positions - which was called by ",a)')trim(caller);
	    write(errmsg(2),'("FATAL ERROR 701 In analyze_positions, the times are NOT monotonically increasing")')
	    write(errmsg(3),'("FATAL ERROR 701  Analyze_positions  is called by ",a)')trim(sub);
	    write(errmsg(4),'("FATAL ERROR 701 Position number ",i3," juld: ",f20.12)')it,juld_date_time(it)
	    write(errmsg(5),'("FATAL ERROR 701 Position number ",i3," juld: ",f20.12)')in,juld_date_time(in)
	    write(errmsg(6),'("FATAL ERROR 701 Difference: ",f20.12)')juld_diff
	    call error_exit(sub,701,6,errmsg); ! FAT ERR
	endif
    enddo
endif ! end of make_qcprof
end subroutine analyze_positions

!=====================================================================
character*1 function evaluate_grade(jdat,jpar,jprf)

    integer, intent( in )     :: jdat		! index of data column
    integer, intent( in )     :: jpar		! index of profile_<param>_qc (by PRES/TEMP/PSAL/DOXY)
    integer, intent( in )     :: jprf		! index of n_prof
                                                                       
    integer :: np,ng,nm				! number of points, ngood, nmissing
    integer :: lmax                             ! local max flag value
    integer :: ipcnt
    real :: pcnt
    character(len=1) :: grade;
    character(len=80) :: hmsg;
    integer :: ib, ie, nc;
    character(len=15) :: sub = 'evaluate_grade'

    lmax = maxprfflag(jprf); ! Store what we have so far for this profile
    np = 0; ng = 0; nm = 0;
    do ilev=1,nlev
	if(profflag(jdat,ilev) == 1)then; 		! good
	    np = np + 1; ng = ng + 1; if(lmax < 1) lmax = 1; 
	elseif(profflag(jdat,ilev) == 2)then;		! probably good
	    np = np + 1; ng = ng + 1; if(lmax < 2) lmax = 2;
	elseif(profflag(jdat,ilev) == 5)then;		! value changed
	    np = np + 1; ng = ng + 1;
	elseif(profflag(jdat,ilev) == 8)then;		! interpolated data
	    ! 2018 late May/early June - do NOT include interpolated data in the nc file, or grade it
	    if(DBLEV > 1000) write(DBG,*)'evluate_grade jprf=',jprf,' lev=',ilev;
	elseif(profflag(jdat,ilev) == 9)then;		! Omit missing data levels
	   nm = nm + 1;                                 ! missing values are OMITTED for grades
	elseif(profflag(jdat,ilev) == 3)then;
	    np = np + 1; if(lmax < 3) lmax = 3;
	elseif(profflag(jdat,ilev) == 4)then;
	    np = np + 1; if(lmax < 4) lmax = 4;
	endif
    enddo
    maxprfflag(jprf) = lmax; ! Store the new max value for this profile
    write(msg,'("GRADE feval jdat,jpar,jprf: ",3i2," np,ng,nm: ",3i4)')jdat,jpar,jprf,np,ng,nm; ! DBG TESTGRADE
    if(np == 0)then
	evaluate_grade = 'Z'; ipcnt = -1;
	write(msg,'(a," pcnt: ",i3," grade: ",a," all missing")')trim(msg),ipcnt,evaluate_grade; call wrttrc(sub,msg);
	write(msg,'("W after filtering out missing levels, NP=ZERO")'); call wrttrc(sub,msg)
	write(msg,'("W return a Z as the grade")');                     call wrttrc(sub,msg)
	return
    endif
    pcnt = (real(ng)/real(np))*100.0                    ! percent good

        if(ng == np)    then; grade = "A";		! all good points
    elseif(ng ==  0)    then; grade = "F";		!  no good points
    elseif(pcnt >= 75.0)then; grade = "B";		! >75 to 100 percent (except all good)
    elseif(pcnt >= 50.0)then; grade = "C";		! >50 to  75 percent
    elseif(pcnt >= 25.0)then; grade = "D";		! >25 to  50 percent
    else;                     grade = "E";		! > 0 to 25 percent
    endif

     ipcnt= ifix(pcnt);

    evaluate_grade = grade;

    return
end function evaluate_grade

integer function get_resolution(datavalue)    

real, intent(in) ::  datavalue

real :: tmp;

if (nint(datavalue*10) == nint(datavalue)*10)then
   get_resolution = 0
elseif (nint(datavalue*100) == nint(datavalue*10)*10)then
   get_resolution = 1
elseif (nint(datavalue*1000) == nint(datavalue*100)*10)then
   get_resolution = 2
elseif (nint(datavalue*10000) == nint(datavalue*1000)*10)then
   get_resolution = 3
elseif (nint(datavalue*100000) == nint(datavalue*10000)*10)then
   get_resolution = 4
elseif (nint(datavalue*1000000) == nint(datavalue*100000)*10)then
   get_resolution = 5
elseif (nint(datavalue*10000000) == nint(datavalue*1000000)*10)then
   get_resolution = 6
endif


return  

end function get_resolution 

!=====================================================================
character*10 function test_for_badgps(date_time)

    use communication                       ! module

    character(*), intent(in) :: date_time
    character(len=4) :: cyear;
    integer          :: year

    cyear = date_time(1:4); ! Lets you just stuff the whole date and even date/time
    read(cyear,'(i4)')year;


    if(year > badgps_test_hi)then
	test_for_badgps = 'Year > max';
    elseif(year < badgps_test_lo)then
	test_for_badgps = 'Year < min';
    else
	test_for_badgps = 'Year valid';
    endif
end function test_for_badgps

character*1 function check_cdate_qc(cdate)

    use communication                       ! module
    implicit none

    character(len=*) , intent( in ) :: cdate    ! The date to be parsed and checked
    integer :: iyr,imo,idy,ihh,imn,isc
    character(len=10) :: reply = 'Not tested'
    logical :: gooddate
    integer :: lsd; ! length of the date
    character(len=19), parameter   :: sub = 'check_cdate_qc';

    check_cdate_qc = '0';	! The date has not yet been checked

    ! First, look for a badly-formed date )msg2phy problems)
    lsd = len(trim(cdate));
    if((lsd == 14) .or. (lsd == 19))then
	! All is ok, looks like the size it shuold be
    else
	! We've got a problem
	write(errmsg(1),'("FATAL ERROR 601 A badly-formed date/time value was found in profile",a)')profile
	write(errmsg(2),'("FATAL ERROR 601 The value of the date time is: ",a)')trim(cdate);
	write(errmsg(3),'("FATAL ERROR 601 This may be dure to a problem in msg2phy")')
	call error_exit(sub,601,3,errmsg); ! In check_cdate_qc, there is a badly formed time value - msg2phy problem?
    endif

    ! Next, look for fill value
    read(cdate,'(i4)')iyr
    if(iyr == 9999)then
	check_cdate_qc = '9';	! The date is missing, fill value
	return;
    endif
    ! Do we have a "badgpsfix" year?
    reply = test_for_badgps(cdate);
    if(reply /= 'Year valid')then
	check_cdate_qc = 'G';   ! The date is invalid/badgps
	badgpsfix_found = .true.
	return;
    endif
    ! Finally, is the date good or bad?
    gooddate = check_cdate(cdate)
    if(gooddate)then
	check_cdate_qc = '1';	! The date is good
    else
	check_cdate_qc = '4';	! The date is BAD
    endif
    return

end function check_cdate_qc

!===================================================================================!
end module data_arrays
! ===========================================================================================================
